Const TYPE_ANY     :Int = 0  + 1024
Const TYPE_BYTE    :Int = 1  + 1024
Const TYPE_SHORT   :Int = 2  + 1024
Const TYPE_INT     :Int = 3  + 1024
Const TYPE_LONG    :Int = 4  + 1024
Const TYPE_FLOAT   :Int = 5  + 1024
Const TYPE_DOUBLE  :Int = 6  + 1024
Const TYPE_STRING  :Int = 7  + 1024
Const TYPE_UBYTE   :Int = 8  + 1024
Const TYPE_USHORT  :Int = 9  + 1024
Const TYPE_UINT    :Int = 10 + 1024
Const TYPE_ULONG   :Int = 11 + 1024
Const TYPE_COMPOUND:Int = 12 + 1024
Const TYPE_FUNCTION:Int = 13 + 1024
Const TYPE_BOOL    :Int = 14 + 1024
Const TYPE_VOID    :Int = 15 + 1024
Const TYPE_POINTER :Int = 16 + 1024
Const TYPE_ARRAY   :Int = 17 + 1024

Global EmitCount:Int, EmitTable:Object[1024]

Function EmitObject:Int(Value:Object)
	If EmitCount >= EmitTable.Length Then EmitTable = EmitTable[.. EmitCount*2]
	
	EmitTable[EmitCount] = Value
	
	EmitCount :+ 1
	
	Return EmitCount - 1
End Function

Function RetrieveList:TList(Index:Int)
	If Index >= 0 Then Return TList(EmitTable[Index]) Else Return Null
End Function

Function RetrieveASTNode:TASTNode(Index:Int)
	If Index >= 0 Then Return TASTNode(EmitTable[Index]) Else Return Null
End Function

Function RetrieveInt:Long(Data:Long)
	Return Data
End Function

Function RetrieveFloat:Float(Data:Long)
	Return (Float Ptr (Varptr Data))[0]
End Function

Function RetrieveDouble:Double(Data:Long)
	Return (Double Ptr (Varptr Data))[0]
End Function

Function RetrieveString:String(Data:Long)
	Local CString:Byte Ptr = (Byte Ptr Ptr (Varptr Data))[0]
	
	Local Result:String = String.FromCString(CString)
	
	MemFree(CString)
	
	Return Result
End Function

Function MergeLists:TList(A:TList, B:TList)
	Local Out:TList = New TList
	
	For Local I:Object = EachIn A
		Out.AddLast(I)
	Next
	
	For Local I:Object = EachIn B
		Out.AddLast(I)
	Next
	
	Return Out
End Function

' Target
'TODO: Make this not global
Global CurrentTarget:TTarget

Type TCodeWriter Abstract
	Method EndScope() Abstract
	Method AddNewline() Abstract
	Method BeginScope() Abstract
	Method EndStatement() Abstract
	Method AddRaw(Raw:String) Abstract
	Method EndInvisibleScope() Abstract
	Method BeginInvisibleScope() Abstract
	Method AddInclude(File:String) Abstract
	Method SetCurrent(Node:TASTNode) Abstract
	Method AddIndentedRaw(Raw:String) Abstract
	Method AddStatement(Node:TASTNode) Abstract
	Method AddStatementRaw(Raw:String) Abstract
	Method BeginStatement(Node:TASTNode) Abstract
	Method AddBracketed(Source:TASTNode) Abstract
	Method AddLocal(Variable:TVariableDeclNode) Abstract
	Method AddList(Source:TList, Separator:String) Abstract
	Method AddArray(Source:TASTNode[], Separator:String) Abstract
	Method AddDeclarationList(Source:TList, Separator:String) Abstract
	Method AddDeclaration(Source:TASTNode, Prefix:String = "") Abstract
	Method AddImplementationList(Source:TList, Separator:String) Abstract
	Method AddImplementation(Source:TASTNode, Prefix:String = "") Abstract
End Type

Type TTarget

	Method CreateProgramNode:TProgramNode() Abstract
	Method CreateFunctionNode:TFunctionNode() Abstract
	Method CreateClassNode:TClassNode() Abstract
	Method CreateIfNode:TIfNode() Abstract
	Method CreateElseIfNode:TElseIfNode() Abstract
	Method CreateElseNode:TElseNode() Abstract
	Method CreateWhileNode:TWhileNode() Abstract
	Method CreateRepeatNode:TRepeatNode() Abstract
	Method CreateSelectNode:TSelectNode() Abstract
	Method CreateCaseNode:TCaseNode() Abstract
	Method CreateDefaultNode:TDefaultNode() Abstract
	Method CreateForWhileNode:TForWhileNode() Abstract
	Method CreateForNode:TForNode() Abstract
	Method CreateExitNode:TExitNode() Abstract
	Method CreateContinueNode:TContinueNode() Abstract
	Method CreateAssignmentNode:TAssignmentNode() Abstract
	Method CreateReturnNode:TReturnNode() Abstract
	Method CreateDeleteNode:TDeleteNode() Abstract  
	Method CreateAddressExpression:TAddressExpression() Abstract
	Method CreateAutoArrayExpression:TAutoArrayExpression() Abstract
	Method CreateNewArrayExpression:TNewArrayExpression() Abstract
	Method CreateIndexExpression:TIndexExpression() Abstract
	Method CreateFunctionCall:TFunctionCall() Abstract
	Method CreateTypeCastExpression:TTypeCastExpression() Abstract
	Method CreateBinaryOpExpression:TBinaryOpExpression() Abstract
	Method CreateUnaryOpExpression:TUnaryOpExpression() Abstract
	Method CreateGroupingExpression:TGroupingExpression() Abstract
	Method CreateMemberAccessExpression:TMemberAccessExpression() Abstract
	Method CreateLiteralNode:TLiteralNode() Abstract
	Method CreateVariableNode:TVariableNode() Abstract
	Method CreateNewExpression:TNewExpression() Abstract
	Method CreateVariableDeclNode:TVariableDeclNode() Abstract
	Method CreateDatatype:TDatatype() Abstract

	Method InitFunctionTable() Abstract

EndType

Type TScope
	Field Nodes:TMap
	
	Field Parent:TScope
	
	Method New()
		Nodes = New TMap
	End Method
	
	Method Find:TASTNode(Identifier:String)
		Local Result:TASTNode = TASTNode(Nodes.ValueForKey(Identifier.ToLower()))
		
		If Not Result And Parent Then Result = Parent.Find(Identifier)
		
		Return Result
	End Method
	
	Method Insert:Int(Identifier:String, Value:TASTNode)
		Return SafeInsert(Nodes, Identifier.ToLower(), Value)
	End Method
	
	Method Create:TScope(Parent:TScope)
		Self.Parent = Parent
		
		Return Self
	End Method
	
	Method Copy:TScope()
		Local Scope:TScope = New TScope
		
		Scope.Nodes  = Nodes.Copy()
		Scope.Parent = Parent
		
		Return Scope
	End Method
	
	Function SafeInsert:Int(Map:TMap, Key:String, Value:TASTNode)
		If Map.ValueForKey(Key) Then
			Return False
		Else
			Map.Insert(Key, Value)
			
			Return True
		EndIf
	End Function
End Type

Type TClassScope Extends TScope
	Field IsStatic:Int
	
	Field Compound:TClassNode
	
	Method Find:TASTNode(Identifier:String)
		Local Node:TASTNode = Compound.FindMember(Identifier, IsStatic)
		
		If Not Node Then Return Super.Find(Identifier) Else Return Node
	End Method
	
	Method CreateClassScope:TClassScope(Parent:TScope, Compound:TClassNode, IsStatic:Int)
		Self.Parent   = Parent
		Self.Compound = Compound
		Self.IsStatic = IsStatic
		
		Return Self
	End Method
End Type

Type TStringEmitter
	Function AppendIdentifier:Byte Ptr(A:Int, B:Int)
		Return (RetrieveString(A) + RetrieveString(B)).ToCString()
	End Function
	
	Function AppendConstChar:Byte Ptr(A:Int, B:Byte Ptr)
		Return (RetrieveString(A) + String.FromCString(B)).ToCString()
	End Function
End Type

Type TListEmitter
	Function PrependToList(List:Int, Node:Int)
		Local Tmp:TASTNode = RetrieveASTNode(Node)
		
		If RetrieveASTNode(Node) Then
			RetrieveList(List).AddFirst(RetrieveASTNode(Node))
		Else
			Local SourceList:TList = RetrieveList(Node)
			
			If Not SourceList.IsEmpty() Then
				Local Link:TLink = RetrieveList(List).AddFirst(SourceList.First())
				
				For Local A:TASTNode = EachIn SourceList
					If A <> SourceList.First() Then Link = RetrieveList(List).InsertAfterLink(A, Link)
				Next
			EndIf
		EndIf
	End Function
	
	Function AppendToList(List:Int, Node:Int)
		If RetrieveASTNode(Node) Then
			RetrieveList(List).AddLast(RetrieveASTNode(Node))
		Else
			Local SourceList:TList = RetrieveList(Node)
			
			For Local A:TASTNode = EachIn SourceList
				RetrieveList(List).AddLast(A)
			Next
		EndIf
	End Function
	
	Function AppendStringToList(List:Int, Data:Int)
		RetrieveList(List).AddLast(RetrieveString(Data))
	End Function
	
	Function EmitEmptyList:Int()
		Return EmitObject(New TList)
	End Function
	
	Function EmitSingleList:Int(Node:Int)
		Local Index:Int = EmitEmptyList()
		
		Local ASTNode:TASTNode = RetrieveASTNode(Node)
		
		If ASTNode Then
			RetrieveList(Index).AddLast(ASTNode)
		Else
			EmitTable[Index] = RetrieveList(Node).Copy()
		EndIf
		
		Return Index
	End Function
End Type

Type TASTNode Abstract
	Global CurrentRoot:TProgramNode
	
	Global CurrentFile:String
	Global Location:Int Ptr
	
	Global IgnoreSemanticErrors:Int
	
	Field TableID:Int
	
	Field Root:TProgramNode
	
	Field Parent:TASTNode
	Field Children:TList
	
	Field IsCompiled:Int
	Field IsExtern  :Int
	
	Field StartLine:Int, EndLine:Int
	Field StartCol:Int,  EndCol:Int
	Field SourceFile:String
	
	Method GetName:String() Abstract
	
	Method New()
		TableID = EmitObject(Self)
		
		Root = CurrentRoot
		SourceFile = CurrentFile
		
		If Location Then
			StartLine = Location[START_LINE]
			StartCol = Location[START_COL]
			EndLine = Location[END_LINE]
			EndCol = Location[END_COL]
		EndIf
		
		Children = New TList
	End Method
	
	Method GenerateCode(Writer:TCodeWriter, Data:Object = Null)
	End Method
	
	Method GenerateDeclaration(Writer:TCodeWriter, Prefix:String = "")
	End Method
	
	Method GenerateImplementation(Writer:TCodeWriter, Prefix:String = "")
	End Method
	
	Method SetExtern()
		IsExtern = True
		
		For Local Child:TASTNode = EachIn Children
			Child.SetExtern()
		Next
	End Method
	
	Method SetParentNode(Parent:TASTNode)
		Self.Parent = Parent
		
		'IsExtern = Parent.IsExtern
	End Method
	
	Method HasParentNode:Int(SearchNode:TASTNode)
		Local P:TASTNode = Parent
		While P
			If P = SearchNode Then Return True
			
			P = P.Parent
		Wend
	End Method
	
	Method AddChildNode(Node:TASTNode)
		Children.AddLast(Node)
		Node.SetParentNode(Self)
	End Method
	
	Method AddChildNodeList(List:TList)
		For Local Node:TASTNode = EachIn List
			AddChildNode(Node)
		Next
	End Method
	
	Method AddChildNodeArray(Children:TASTNode[])
		For Local Node:TASTNode = EachIn Children
			AddChildNode(Node)
		Next
	End Method
	
	Method ReplaceChildNode(Original:TASTNode, Replacement:TASTNode)
		Local ID:TTypeId = TTypeId.ForName(GetName())
		
		For Local TypeField:TField = EachIn ID.EnumFields()
			If TypeField.Get(Self) = Original Then TypeField.Set(Self, Replacement)
		Next
		
		Local Link:TLink = Children.FirstLink()
		While Link
			If Link.Value() = Original Then Link._Value = Replacement
			
			Link = Link.NextLink()
		Wend
		
		Replacement.SetParentNode(Self)
	End Method
	
	Method AddDependency(Source:TASTNode, Cause:TASTNode)
		Parent.AddDependency(Source, Cause)
	End Method
	
	Method CheckCorrectness(Scope:TScope) Abstract
	
	Method ToString:String()
		Return StartLine + "-" + EndLine + ":" + StartCol + "-" + EndCol
	End Method
	
	Method LocationFromString(Source:String)
		Local A:String[] = Source.Split(":")
		
		Local Lines:String[] = A[0].Split("-")
		StartLine = Lines[0].ToInt()
		  EndLine = Lines[1].ToInt()
		
		Local Cols:String[] = A[1].Split("-")
		StartCol = Cols[0].ToInt()
		  EndCol = Cols[1].ToInt()
	End Method
	
	Method CopyLocation(Source:TASTNode)
		SourceFile = Source.SourceFile
		
		StartLine = Source.StartLine
		  EndLine = Source.  EndLine
		
		StartCol  = Source.StartCol
		  EndCol  = Source.  EndCol
	End Method
	
	Function Abort(Message:String, Cause:TASTNode)
		TCompiler.Abort("Semantic error: " + Message + ..
			"~nIn file "   + Cause.SourceFile + ..
			"~nIn line "   + Cause.StartLine + "-" + Cause.EndLine + ..
			"~nIn column " + Cause.StartCol  + "-" + Cause.EndCol)
	End Function
	
	Function SetRuleLocation(L:Int Ptr)
		Location = L
	End Function
	
	Function SetCommonRoot(Root:TProgramNode)
		CurrentRoot = Root
		CurrentFile = Root.SourceFile
	End Function
End Type

Type TProgramNode Extends TASTNode
	Field Dependencies  :TDependencyNode
	Field FileImports   :TList
	Field ModuleImports :TList
	Field Statements    :TList
	Field Functions     :TList
	Field Classes       :TList
	
	Field Includes      :TList
	Field LinkTargets   :TList
	Field CompileTargets:TList
	Field Libraries     :TList
	
	Field IsModule:Int
	
	Field ChildCount:Int
	
	Method New()
		Dependencies   = New TDependencyNode.Create(SourceFile)
		FileImports    = New TList
		ModuleImports  = New TList
		Statements     = New TList
		Functions      = New TList
		Classes        = New TList
		Includes       = New TList
		LinkTargets    = New TList
		CompileTargets = New TList
		Libraries      = New TList
	End Method
	
	Method GetName:String()
		Return "TProgramNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		For Local Class:TClassNode = EachIn Classes
			If Not Scope.Insert(Class.Identifier, Class) Then Abort("Redefinition of '" + Class.Identifier + "'", Class)
		Next
		
		For Local VarDecl:TVariableDeclNode = EachIn Statements
			If VarDecl.DeclKind = KEY_CONST Then VarDecl.CheckCorrectness(Scope)
		Next
		
		For Local VarDecl:TVariableDeclNode = EachIn Statements
			VarDecl.CheckSignatureCorrectness(Scope)
			
			If VarDecl.DeclKind = KEY_GLOBAL Then
				If Not Scope.Insert(VarDecl.Identifier, VarDecl) Then ..
					Abort("Redefinition of '" + VarDecl.Identifier + "'", VarDecl)
			EndIf
		Next
		
		For Local FunctionNode:TFunctionNode = EachIn Functions
			FunctionNode.CheckSignatureCorrectness(Scope)
			
			If Not Scope.Insert(FunctionNode.Identifier, FunctionNode.Declaration) Then ..
				Abort("Redefinition of '" + FunctionNode.Identifier + "'", FunctionNode)
		Next
		
		For Local Class:TClassNode = EachIn Classes
			Class.ResolveParents(Scope)
		Next
		
		For Local Class:TClassNode = EachIn Classes
			If Not Class.CycleColor Then
				Class.CycleColor = Class.TableID
				
				If CheckInheritanceCycles(Class) Then Abort("Cyclic class inheritance", Class)
			EndIf
		Next
		
		For Local Class:TClassNode = EachIn Classes
			Class.CheckSignatureCorrectness(Scope)
		Next
		
		For Local Class:TClassNode = EachIn Classes
			Class.CheckCorrectness(Scope)
		Next
		
		For Local FunctionNode:TFunctionNode = EachIn Functions
			FunctionNode.CheckCorrectness(Scope)
		Next
		
		For Local Statement:TStatementNode = EachIn Statements
			Local VarDecl:TVariableDeclNode = TVariableDeclNode(Statement)
			
			If (Not VarDecl) Or VarDecl.DeclKind <> KEY_CONST Then Statement.CheckCorrectness(Scope)
		Next
	End Method
	
	Method CheckInheritanceCycles:Int(Class:TClassNode)
		For Local Parent:TClassNode = EachIn Class.ClassParents
			If Not Parent.CycleColor Then
				Parent.CycleColor = Class.CycleColor
				
				If CheckInheritanceCycles(Parent) Then Return True
			ElseIf Parent.CycleColor = Class.CycleColor Then
				Return True
			EndIf
		Next
	End Method
	
	'Note: Changed from ToString to ToStream, which writes directly do a stream instead of building a string first.
	'This is due to performance considerations when compiling massive files
	Method ToStream(Stream:TStream)
		For Local VarDecl:TVariableDeclNode = EachIn Statements
			If VarDecl.DeclKind = KEY_CONST Or VarDecl.DeclKind = KEY_GLOBAL Then
				If VarDecl.IsExtern Then Stream.WriteLine("Extern")
				Stream.WriteLine(VarDecl.ToString())
			EndIf
		Next
		
		For Local FunctionNode:TFunctionNode = EachIn Functions
			If FunctionNode.IsExtern Then Stream.WriteLine("Extern")
			Stream.WriteLine(FunctionNode.ToString())
		Next
		
		For Local Class:TClassNode = EachIn Classes
			If Class.IsExtern Then Stream.WriteLine("Extern")
			Class.ToStream(Stream)
		Next
	End Method
	
	Method SetupChildren()
		AddChildNodeList(Classes)
		AddChildNodeList(Functions)
		AddChildNodeList(Statements)
	End Method
	
	Method FromString:TProgramNode(Source:String)
		Local Lines:String[] = Source.Split("~n")
		
		IsCompiled = True
		
		CurrentRoot = Self
		
		Local CurrentClass:TClassNode, LastNode:TASTNode, ExternFlag:Int
		For Local I:Int = 0 Until Lines.Length
			Local Tokens:String[] = Lines[I].Split(" ")
			
			'If line starts with a code location ("xxx-xxx:xxx-xxx"), skip to the second token
			Local StartToken:Int = (Tokens[0].Find("-") >= 0)
			
			Select Tokens[StartToken]
				Case "File"
					FileImports.AddLast(Lines[I][Lines[I].Find(" ") + 1 ..])
				Case "Module"
					ModuleImports.AddLast(Lines[I][Lines[I].Find(" ") + 1 ..])
				Case "Function"
					Local FunctionNode:TFunctionNode = CurrentTarget.CreateFunctionNode().FromString(Lines[I])
					If CurrentClass Then
						CurrentClass.Functions.AddLast(FunctionNode)
					Else
						Functions.AddLast(FunctionNode)
					EndIf
					LastNode = FunctionNode
				Case "Method"
					Local FunctionNode:TFunctionNode = CurrentTarget.CreateFunctionNode().FromString(Lines[I])
					If FunctionNode.Identifier.ToLower() = CurrentClass.Identifier.ToLower() Then
						'FunctionNode.Identifier  = CurrentClass.GenerateConstructorIdentifier()
						FunctionNode.TransformToConstructor(CurrentClass)
						CurrentClass.Constructor = FunctionNode
					Else
						CurrentClass.Methods.AddLast(FunctionNode)
					EndIf
					LastNode = FunctionNode
				Case "Class"
					CurrentClass = CurrentTarget.CreateClassNode().FromString(Lines[I])
					Classes.AddLast(CurrentClass)
					
					LastNode = CurrentClass
				Case "EndClass"
					CurrentClass.SetupChildren()
					CurrentClass = Null
					TClassNode.SetVisibility(KEY_PUBLIC)
				Case "Global", "Const", "Field"
					Local Statement:TStatementNode = CurrentTarget.CreateVariableDeclNode().FromString(Lines[I])
					If CurrentClass Then
						CurrentClass.Variables.AddLast(Statement)
					Else
						Statements.AddLast(Statement)
					EndIf
					LastNode = Statement
				Case "Extern"
					ExternFlag = True
				Case "Private"
					TClassNode.SetVisibility(KEY_PRIVATE)
				Case "Protected"
					TClassNode.SetVisibility(KEY_PRIVATE)
			End Select
			
			If LastNode Then
				If ExternFlag Then LastNode.SetExtern()
				ExternFlag = False
				LastNode = Null
			EndIf
		Next
		
		SetupChildren()
		
		Return Self
	End Method
	
	Method Merge(Other:TProgramNode)
		Statements = MergeLists(Statements, Other.Statements)
		Functions  = MergeLists(Functions,  Other.Functions)
		Classes    = MergeLists(Classes,    Other.Classes)
		Libraries  = MergeLists(Libraries,  Other.Libraries)
	End Method
	
	Method AddDependency(Source:TASTNode, Cause:TASTNode)
		Local TargetMap:TMap
		
		If Dependencies.Contains(Source.Root.Dependencies) Then
			TargetMap = TMap(Dependencies.ValueForKey(Source.Root.Dependencies))
		Else
			TargetMap = New TMap
			
			Dependencies.Insert(Source.Root.Dependencies, TargetMap)
		EndIf
		
		If (Not TargetMap.Contains(Source)) Or TASTNode(TargetMap.ValueForKey(Source)).IsExtern Then ..
			TargetMap.Insert(Source, Cause)
	End Method
	
	Method Create:TProgramNode()
		Return Self
	End Method
	
	Function Emit:Int()
		Local ProgramNode:TProgramNode = CurrentTarget.CreateProgramNode()
		
		CurrentRoot = ProgramNode
		
		Return ProgramNode.TableID
	EndFunction
	
	Function EmitProgramFileImport(Program:Int, File:Int)
		Local ProgramNode:TProgramNode = TProgramNode(RetrieveASTNode(Program))
		
		ProgramNode.FileImports.AddLast(RetrieveString(File))
	End Function
	
	Function EmitProgramModuleImport(Program:Int, ModulePath:Int)
		Local ProgramNode:TProgramNode = TProgramNode(RetrieveASTNode(Program))
		
		ProgramNode.ModuleImports.AddLast(RetrieveString(ModulePath))
	End Function

	Function EmitProgramGlobal(Program:Int, GlobalList:Int)
		Local ProgramNode:TProgramNode = TCPPProgramNode(RetrieveASTNode(Program))
		Local NodeList:TList = RetrieveList(GlobalList)
		
		ProgramNode.AddChildNodeList(NodeList)
		ProgramNode.Statements = MergeLists(ProgramNode.Statements, NodeList)
	EndFunction
	
	Function EmitProgramStatement(Program:Int, Statement:Int)
		Local ProgramNode:TProgramNode = TProgramNode(RetrieveASTNode(Program))
		Local Node:TASTNode = RetrieveASTNode(Statement)
		
		If Node Then
			ProgramNode.AddChildNode(Node)
			ProgramNode.Statements.AddLast(TStatementNode(Node))
		Else
			Local NodeList:TList = RetrieveList(Statement)
			
			ProgramNode.AddChildNodeList(NodeList)
			ProgramNode.Statements = MergeLists(ProgramNode.Statements, NodeList)
		EndIf
	EndFunction
	
	Function EmitProgramFunction(Program:Int, FunctionIndex:Int)
		Local ProgramNode:TProgramNode = TProgramNode(RetrieveASTNode(Program))
		Local Node:TASTNode = RetrieveASTNode(FunctionIndex)
		
		ProgramNode.AddChildNode(Node)
		ProgramNode.Functions.AddLast(TFunctionNode(Node))
	EndFunction
	
	Function EmitProgramClass(Program:Int, Class:Int)
		Local ProgramNode:TProgramNode = TProgramNode(RetrieveASTNode(Program))
		Local Node:TASTNode = RetrieveASTNode(Class)
		
		ProgramNode.AddChildNode(Node)
		ProgramNode.Classes.AddLast(TClassNode(Node))
	EndFunction

	Function EmitProgramExtern(Program:Int, ExternBlock:Int)
		Local ProgramNode:TProgramNode = TProgramNode(RetrieveASTNode(Program))
		
		Local DeclarationList:TList = RetrieveList(ExternBlock)
		
		For Local Node:TASTNode = EachIn DeclarationList
			ProgramNode.AddChildNode(Node)
			Node.SetExtern()
			
			If TVariableDeclNode(Node) Then
				ProgramNode.Statements.AddLast(TVariableDeclNode(Node))
			ElseIf TClassNode(Node) Then
				ProgramNode.Classes.AddLast(TClassNode(Node))
			ElseIf TFunctionNode(Node) Then
				ProgramNode.Functions.AddLast(TFunctionNode(Node))
			EndIf
		Next
	EndFunction
End Type

Type TFunctionNode Extends TASTNode
	Field IsMethod:Int
	Field IsAbstract:Int
	Field IsFinal:Int
	Field IsConstructor:Int
	
	Field Identifier:String
	Field Datatype:TDatatype
	Field Declaration:TVariableDeclNode
	
	Field ParentClass:TClassNode
	
	Field ReturnDatatype:TDatatype
	Field Parameters:TList
	Field ParameterCount:Int
	
	Field Statements:TStatementList
	
	Method GetName:String()
		Return "TFunctionNode"
	End Method

	Method GetParameter:TVariableDeclNode(Index:Int)
		Return TVariableDeclNode(Parameters.ValueAtIndex(Index))
	End Method
	
	Method CheckSignatureCorrectness(Scope:TScope)
		If Not ReturnDatatype.IsVoid() Then ReturnDatatype.CheckCorrectness(Scope)
		
		If IsAbstract And (Not IsMethod) Then ..
			Abort("Functions can not be declared as abstract", Self)
		
		Local LastWasOptional:Int
		For Local VarDecl:TVariableDeclNode = EachIn Parameters
			VarDecl.CheckSignatureCorrectness(Scope)
			
			If VarDecl.Value Then
				LastWasOptional = True
			ElseIf LastWasOptional Then
				Abort("Non-optional parameter following optional parameter", VarDecl)
			EndIf
		Next
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		If IsMethod Then
			Local P:TASTNode = Parent, ClassNode:TClassNode
			While P
				ClassNode = TClassNode(P)
				If ClassNode Then
					ParentClass = ClassNode
					
					Exit
				EndIf
				
				P = P.Parent
			Wend
			
			If Not ClassNode Then Abort("Methods can only be declared inside classes", Self)
		EndIf
		
		If Not ReturnDatatype.IsVoid() Then ReturnDatatype.CheckCorrectness(Scope)
		
		Scope = New TScope.Create(Scope)
		
		'TODO: Optional parameter values referencing previous parameters are not caught
		For Local VarDecl:TVariableDeclNode = EachIn Parameters
			VarDecl.CheckCorrectness(Scope)
		Next
		
		Statements.CheckCorrectness(Scope)
	End Method
	
	Method TransformToConstructor(Class:TClassNode)
		SetCommonRoot(Root)
		
		If Not ReturnDatatype.IsVoid() Then Abort("Specifying return type for constructors is not allowed", Self)
		
		Local ClassType:TDatatype = CurrentTarget.CreateDatatype().Create(TYPE_COMPOUND, Class, 0)
		
		Datatype.ReplaceChildNode(Datatype.ReturnType, ClassType.Copy())
		ReplaceChildNode(ReturnDatatype, ClassType)
		
		ParentClass = Class
		Identifier = Class.GenerateConstructorIdentifier()
		Declaration.Identifier = Identifier
		
		IsConstructor = True
	End Method
	
	Method ToString:String()
		Local Out:String = Super.ToString() + " "
		
		If IsMethod Then
			Out :+ "Method "
		Else
			Out :+ "Function "
		EndIf
		
		If IsConstructor Then Out :+ ParentClass.Identifier Else Out :+ Identifier
		
		If (Not ReturnDatatype.IsVoid()) And (Not IsConstructor) Then ..
			Out :+ ":" + ReturnDatatype.ToString()
		
		Out :+ "("
		
		For Local Param:TVariableDeclNode = EachIn Parameters
			Out :+ Param.Identifier + ":" + Param.Datatype.ToString()
			
			If Param.Value Then Out :+ "="
			
			If Param <> Parameters.Last() Then Out :+ ", "
		Next
		
		Out :+ ")"
		
		If IsAbstract Then Out :+ " Abstract"
		If IsFinal    Then Out :+ " Final"
		
		Return Out
	End Method
	
	Method ReplaceChildNode(Original:TASTNode, Replacement:TASTNode)
		For Local Param:TASTNode = EachIn Parameters
			If Param = Original Then
				Local Link:TLink = Parameters.FindLink(Param)
				
				Link._value = Replacement
				Return
			EndIf
		Next
		
		Super.ReplaceChildNode(Original, Replacement)
	End Method
	
	Method SetupChildren()
		AddChildNodeList(Parameters)
		AddChildNode(Statements)
		AddChildNode(ReturnDatatype)
		AddChildNode(Datatype)
		AddChildNode(Declaration)
	End Method
	
	Method FromString:TFunctionNode(Source:String)
		Local Space:Int = Source.Find(" ")
		Super.LocationFromString(Source[ .. Space])
		
		Source = Source[Space + 1 ..]
		
		If Source.StartsWith("Method") Then IsMethod = True
		
		IsCompiled = True
		
		Parameters     = New TList
		Statements     = New TStatementList.Create(New TList)
		ParameterCount = 0
		
		Local ColonPos:Int = Source.Find(":")
		Local BracePos:Int = Source.Find("(")
		While FindAtLevel(Source, ")(", BracePos + 1) >= 0
			BracePos = FindAtLevel(Source, ")(", BracePos + 1) + 1
		Wend
		
		If ColonPos = -1 Or ColonPos > BracePos Then ColonPos = BracePos
		
		Identifier = Source[Source.Find(" ") + 1 .. ColonPos]
		
		ReturnDatatype = CurrentTarget.CreateDatatype().FromString(Source[ColonPos + 1 .. BracePos])
		
		If Source.Find("()") < 0 Then
			Local Start:Int = BracePos
			Local Stop :Int = FindAtLevel(Source, ",", BracePos, 1)
			While True
				If Stop < 0 Then Stop = FindAtLevel(Source, ")", BracePos, 1)
				
				Local ColonPos:Int        = FindAtLevel(Source, ":", Start + 1)
				Local VarIdent:String     = Source[Start + 1 .. ColonPos]
				Local VarRest :String     = Source[ColonPos + 1 .. Stop]
				Local VarValue:TLiteralNode
				If VarRest.EndsWith("=") Then
					VarRest = VarRest[.. VarRest.Length - 1]
					
					'HACK: Make AST believe there actually is an optional value even though we don't know what it is
					VarValue = CurrentTarget.CreateLiteralNode()
					VarValue.LiteralKind = KEY_NULL
				EndIf
				
				Local VarType :TDatatype  = CurrentTarget.CreateDatatype().FromString(VarRest)
				
				Parameters.AddLast(CurrentTarget.CreateVariableDeclNode().Create(KEY_LOCAL, VarIdent, VarType, VarValue))
				ParameterCount :+ 1
				
				If Source[Stop] = ")"[0] Then Exit
				
				Start = Stop + 1
				Stop = FindAtLevel(Source, ",", Stop + 1)
			Wend
		EndIf
		
		If Source.EndsWith("Abstract") Then IsAbstract = True
		If Source.EndsWith("Final"   ) Then IsFinal    = True
		
		Datatype = CurrentTarget.CreateDatatype().Create(TYPE_FUNCTION, Self)
		
		Declaration = CurrentTarget.CreateVariableDeclNode().Create(-1, Identifier, Datatype, Null)
		Declaration.CopyLocation(Self)
		
		SetupChildren()
		
		Return Self
	End Method

	Function Emit:Int(VariableIndex:Int, Statements:Int, IsMethod:Int, InheritanceFlags:Int)
		Local FunctionNode:TFunctionNode = CurrentTarget.CreateFunctionNode()
		Local Variable:TVariableNode = TVariableNode(RetrieveASTNode(VariableIndex))
		
		FunctionNode.IsMethod        = IsMethod
		FunctionNode.Identifier      = Variable.Identifier
		FunctionNode.ReturnDatatype  = Variable.Datatype.ReturnType
		FunctionNode.Parameters      = New TList
		FunctionNode.Statements      = New TStatementList.Create(RetrieveList(Statements))
		FunctionNode.IsAbstract      = (InheritanceFlags & FLAG_ABSTRACT) <> 0
		FunctionNode.IsFinal         = (InheritanceFlags & FLAG_FINAL)    <> 0
		
		FunctionNode.ParameterCount  = Variable.Datatype.ParameterCount
		
		If Not FunctionNode.ReturnDatatype Then FunctionNode.ReturnDatatype = CurrentTarget.CreateDatatype().Create(TYPE_VOID, Null)
		
		For Local I:Int = 0 Until FunctionNode.ParameterCount
			Local VarDecl:TVariableDeclNode = CurrentTarget.CreateVariableDeclNode().Create( ..
					KEY_LOCAL, ..
					Variable.Datatype.Identifiers[I], ..
					Variable.Datatype.Parameters [I], ..
					Variable.Datatype.ParamValues[I]  ..
				)
			
			VarDecl.CopyLocation(Variable)
			
			FunctionNode.Parameters.AddLast(VarDecl)
		Next
		
		FunctionNode.Datatype = CurrentTarget.CreateDatatype().Create(TYPE_FUNCTION, FunctionNode)
		
		FunctionNode.Declaration = CurrentTarget.CreateVariableDeclNode().Create(-1, FunctionNode.Identifier, FunctionNode.Datatype, Null)
		FunctionNode.Declaration.CopyLocation(FunctionNode)
		
		FunctionNode.SetupChildren()
		
		Return FunctionNode.TableID
	End Function

End Type

Type TClassNode Extends TASTNode
	Global CurrentVisibility:Int
	
	Field Identifier:String
	
	Field ClassParents:TList
	Field ClassChildren:TList
	
	Field Dependencies:TDependencyNode
	Field CheckingSignature:Int
	
	Field Variables:TList
	
	Field IntermediateScope:TScope
	Field Methods:TList
	Field Functions:TList
	Field Constructor:TFunctionNode
	
	Field AbstractFunctions:TList
	Field AbstractFlag:Int
	Field AbstractChecked:Int
	
	Field IsFinal:Int
	
	Field CycleColor:Int 'Used for detecting inheritance cycles
	
	Method New()
		Dependencies      = New TDependencyNode.Create(SourceFile)
		AbstractFunctions = New TList
		ClassChildren     = New TList
	End Method
	
	Method GetName:String()
		Return "TClassNode"
	End Method
	
	Method IsAbstract:Int()
		If Not AbstractChecked Then
			AbstractChecked = True
			
			For Local Parent:TClassNode = EachIn ClassParents
				If Parent.IsAbstract() Then
					For Local AbstractFunction:TFunctionNode = EachIn Parent.AbstractFunctions
						If Not IntermediateScope.Find(AbstractFunction.Identifier) Then ..
							AbstractFunctions.AddLast(AbstractFunction)
					Next
				EndIf
			Next
		EndIf
		
		Return AbstractFlag Or (Not AbstractFunctions.IsEmpty())
	End Method
	
	Method HasParent:Int(P:TClassNode)
		For Local Parent:TClassNode = EachIn ClassParents
			If Parent = P Or Parent.HasParent(P) Then Return True
		Next
		
		Return False
	End Method
	
	Method FindMember:TVariableDeclNode(Identifier:String, IsStatic:Int)
		If Constructor And Constructor.Identifier.ToLower() = Identifier.ToLower() Then Return Constructor.Declaration
		
		For Local Variable:TVariableDeclNode = EachIn Variables
			If Variable.Identifier.ToLower() = Identifier.ToLower() Then
				If (Variable.DeclKind = KEY_CONST Or Variable.DeclKind = KEY_GLOBAL) Or (Not IsStatic) Then Return Variable
			EndIf
		Next
		
		If Not IsStatic Then
			For Local FunctionNode:TFunctionNode = EachIn Methods
				If FunctionNode.Identifier.ToLower() = Identifier.ToLower() Then Return FunctionNode.Declaration
			Next
		EndIf
		
		For Local FunctionNode:TFunctionNode = EachIn Functions
			If FunctionNode.Identifier.ToLower() = Identifier.ToLower() Then Return FunctionNode.Declaration
		Next
		
		Local FoundNode:TVariableDeclNode
		For Local Parent:TClassNode = EachIn ClassParents
			Local Node:TVariableDeclNode = Parent.FindMember(Identifier, IsStatic)
			
			If Node Then
				If FoundNode Then
					Abort("Class member '" + Identifier + "' is ambiguous", Self)
				Else
					FoundNode = Node
				EndIf
			EndIf
		Next
		
		Return FoundNode
	End Method
	
	Method ResolveParents(Scope:TScope)
		If (Not IsExtern) And ClassParents.IsEmpty() Then ClassParents.AddLast("Object")
		
		Local ActualParents:TList = New TList
		
		For Local Parent:TClassNode = EachIn ClassParents 'Save resolved parents from previous calls to this method
			ActualParents.AddLast(Parent)
		Next
		
		For Local Parent:String = EachIn ClassParents
			Local Class:TClassNode = TClassNode(Scope.Find(Parent))
			
			If Class Then
				If Class.IsFinal Then Abort("Unable to extend final class '" + Class.Identifier + "'", Self)
				
				ActualParents.AddLast(Class)
				Class.ClassChildren.AddLast(Self)
				
				AddDependency(Class, Self)
			Else
				If IgnoreSemanticErrors Then
					ActualParents.AddLast(Parent)
				Else
					Abort("Class '" + Parent + "' not found", Self)
				EndIf
			EndIf
		Next
		ClassParents = ActualParents
	End Method
	
	Method CheckSignatureCorrectness(Scope:TScope)
		Scope = New TScope.Create(Scope)
		
		CheckingSignature = True
		
		If Constructor Then
			Constructor.CheckSignatureCorrectness(Scope)
			InsertMember(Scope, Constructor.Declaration, True)
		EndIf
		
		For Local FunctionNode:TFunctionNode = EachIn Functions
			FunctionNode.CheckSignatureCorrectness(Scope)
			
			If FunctionNode.IsAbstract Then AbstractFunctions.AddLast(FunctionNode)
			
			InsertMember(Scope, FunctionNode.Declaration, False)
		Next
		
		For Local VariableDecl:TVariableDeclNode = EachIn Variables
			VariableDecl.CheckSignatureCorrectness(Scope)
			
			InsertMember(Scope, VariableDecl, False)
		Next
		
		For Local FunctionNode:TFunctionNode = EachIn Methods
			FunctionNode.CheckSignatureCorrectness(Scope)
			
			If FunctionNode.IsAbstract Then AbstractFunctions.AddLast(FunctionNode)
			
			InsertMember(Scope, FunctionNode.Declaration, True)
		Next
		
		CheckingSignature = False
		
		IntermediateScope = Scope
	End Method
	
	Method AddDependency(Source:TASTNode, Cause:TASTNode)
		If CheckingSignature Then
			Local TargetMap:TMap
			
			If Dependencies.Contains(Source.Root.Dependencies) Then
				TargetMap = TMap(Dependencies.ValueForKey(Source.Root.Dependencies))
			Else
				TargetMap = New TMap
				
				Dependencies.Insert(Source.Root.Dependencies, TargetMap)
			EndIf
			
			If (Not TargetMap.Contains(Source)) Or TASTNode(TargetMap.ValueForKey(Source)).IsExtern Then ..
				TargetMap.Insert(Source, Cause)
		EndIf
		
		Parent.AddDependency(Source, Cause)
	End Method
	
	Method InsertMember(Scope:TScope, VariableDeclNode:TVariableDeclNode, IsMethod:Int) 'Oh god the innuendo
		If Not Scope.Insert(VariableDeclNode.Identifier, VariableDeclNode) Then ..
			Abort("Redefinition of '" + VariableDeclNode.Identifier + "'", VariableDeclNode)
		
		For Local Parent:TClassNode = EachIn ClassParents
			Local VarDeclNode:TVariableDeclNode = TVariableDeclNode(Parent.FindMember(VariableDeclNode.Identifier, False))
			
			If VarDeclNode Then
				If IsMethod Then
					If Not VariableDeclNode.Datatype.IsCastable(VarDeclNode.Datatype) Then ..
						Abort("Overriding method '" + VariableDeclNode.Identifier + "' differs by type", VariableDeclNode)
					
					If VarDeclNode.Datatype.FunctionNode.IsFinal Then ..
						Abort("Unable to override final method '" + VariableDeclNode.Identifier + "'", VariableDeclNode)
				Else
					Abort("Redefinition of '" + VariableDeclNode.Identifier + "' inherited from '" + Parent.Identifier + "'", VariableDeclNode)
				EndIf
			EndIf
		Next
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Local StaticScope:TClassScope = New TClassScope.CreateClassScope(Scope, Self, True)
		Local  ClassScope:TClassScope = New TClassScope.CreateClassScope(Scope, Self, False)
		
		For Local VarDecl:TVariableDeclNode = EachIn Variables
			If VarDecl.DeclKind = KEY_FIELD Then
				VarDecl.CheckCorrectness(ClassScope)
			Else
				VarDecl.CheckCorrectness(StaticScope)
			EndIf
		Next
		
		If Constructor Then
			If Not Constructor.IsMethod Then ..
				Abort("Constructor has to be a method", Constructor)
			
			Constructor.CheckCorrectness(ClassScope)
		EndIf
		
		For Local FunctionNode:TFunctionNode = EachIn Functions
			FunctionNode.CheckCorrectness(StaticScope)
		Next
		
		For Local FunctionNode:TFunctionNode = EachIn Methods
			FunctionNode.CheckCorrectness(ClassScope)
		Next
	End Method
	
	Method ToStream(Stream:TStream)
		Local Out:String = Super.ToString()
		
		Out :+ " Class " + Identifier
		
		If Not ClassParents.IsEmpty() Then
			Out :+ " Extends "
			
			For Local Parent:TClassNode = EachIn ClassParents
				If Parent <> ClassParents.First() Then Out :+ " "
				
				Out :+ Parent.Identifier
			Next
			
			For Local Parent:String = EachIn ClassParents
				If Parent <> ClassParents.First() Then Out :+ " "
				
				Out :+ Parent
			Next
		EndIf
		
		If AbstractFlag Then Out :+ " Abstract"
		If IsFinal      Then Out :+ " Final"
		
		Stream.WriteLine(Out)
		
		WriteWithVisibility(Stream, KEY_PUBLIC,    Variables, Methods, Functions, Constructor)
		Stream.WriteLine("Private")
		WriteWithVisibility(Stream, KEY_PRIVATE,   Variables, Methods, Functions, Constructor)
		Stream.WriteLine("Protected")
		WriteWithVisibility(Stream, KEY_PROTECTED, Variables, Methods, Functions, Constructor)
		Stream.WriteLine("EndClass")
		
		Function WriteWithVisibility(Stream:TStream, Visibility:Int, Variables:TList, Methods:TList, Functions:TList, Constructor:TFunctionNode)
			For Local Variable:TVariableDeclNode = EachIn Variables
				If Variable.Visibility = Visibility Then ..
					Stream.WriteLine(Variable.ToString())
			Next
			
			If Constructor Then WriteFunction(Stream, Constructor, Visibility)
			
			For Local FunctionNode:TFunctionNode = EachIn Methods
				WriteFunction(Stream, FunctionNode, Visibility)
			Next
			
			For Local FunctionNode:TFunctionNode = EachIn Functions
				WriteFunction(Stream, FunctionNode, Visibility)
			Next
			
			Function WriteFunction(Stream:TStream, F:TFunctionNode, Visibility:Int)
				If F.Declaration.Visibility = Visibility Then ..
					Stream.WriteLine(F.ToString())
			End Function
		End Function
	End Method
	
	Method SetupChildren()
		If Constructor Then AddChildNode(Constructor)
		AddChildNodeList(Variables)
		AddChildNodeList(Methods)
		AddChildNodeList(Functions)
	End Method
	
	Method FromString:TClassNode(Source:String)
		Local Space:Int = Source.Find(" ")
		Super.LocationFromString(Source[ .. Space])
		
		Source = Source[Space + 1 ..]
		
		Local Parts:String[] = Source.Split(" ")
		
		Identifier = Parts[1]
		
		IsCompiled = True
		
		Variables = New TList
		Methods   = New TList
		Functions = New TList
		ClassParents = New TList
		
		Local Index:Int = 3
		If Parts.Length >= 3 And Parts[2] = "Extends" Then
			While Index < Parts.Length And Parts[Index] <> "Abstract" And Parts[Index] <> "Final"
				ClassParents.AddLast(Parts[Index])
				Index :+ 1
			Wend
		EndIf
		
		If Index < Parts.Length And Parts[Index] = "Abstract" Then AbstractFlag = True; Index :+ 1
		If Index < Parts.Length And Parts[Index] = "Final"    Then IsFinal      = True
		
		Return Self
	End Method
	
	Method GenerateConstructorIdentifier:String()
		Return "__Create" + Identifier
	End Method
	
	Function SetVisibility(Mode:Int)
		CurrentVisibility = Mode
	End Function
	
	Function Emit:Int(Identifier:Int, ExtendList:Int, DeclarationList:Int, InheritanceFlags:Int)
		Local ClassNode:TClassNode = CurrentTarget.CreateClassNode()
		ClassNode.Identifier   = RetrieveString(Identifier)
		ClassNode.ClassParents = RetrieveList(ExtendList)
		
		If (Not ClassNode.ClassParents) Then ClassNode.ClassParents = New TList
		
		ClassNode.Variables = New TList
		ClassNode.Methods   = New TList
		ClassNode.Functions = New TList
		
		ClassNode.AbstractFlag = (InheritanceFlags & FLAG_ABSTRACT) <> 0
		ClassNode.IsFinal      = (InheritanceFlags & FLAG_FINAL)    <> 0
		
		Local DeclList:TList = RetrieveList(DeclarationList)
		
		For Local Variable:TVariableDeclNode = EachIn DeclList
			ClassNode.Variables.AddLast(Variable)
		Next
		
		For Local FunctionNode:TFunctionNode = EachIn DeclList
			If FunctionNode.Identifier.ToLower() = ClassNode.Identifier.ToLower() Then
				ClassNode.Constructor = FunctionNode
				FunctionNode.TransformToConstructor(ClassNode)
			ElseIf FunctionNode.IsMethod Then
				ClassNode.Methods.AddLast(FunctionNode)
			Else
				ClassNode.Functions.AddLast(FunctionNode)
			EndIf
		Next
		
		ClassNode.SetupChildren()
		
		SetVisibility(KEY_PUBLIC)
		
		Return ClassNode.TableID
	End Function
End Type

Type TStatementList Extends TASTNode
	Field Constants:TList
	Field Statements:TList
	
	Method GetName:String()
		Return "TStatementList"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		For Local VarDecl:TVariableDeclNode = EachIn Constants
			VarDecl.CheckCorrectness(Scope)
		Next
		
		For Local Statement:TStatementNode = EachIn Statements
			Statement.CheckCorrectness(Scope)
		Next
	End Method
	
	Method GenerateCode(Writer:TCodeWriter, Data:Object = Null)
		For Local VarDecl:TVariableDeclNode = EachIn Constants
			Writer.AddImplementation(VarDecl)
		Next
		
		For Local Statement:TASTNode = EachIn Statements
			Local VarDecl:TVariableDeclNode = TVariableDeclNode(Statement)
			
			If VarDecl Then
				If VarDecl.DeclKind = KEY_GLOBAL Then
					Writer.BeginStatement(VarDecl)
					Writer.AddRaw(VarDecl.Identifier + " = ")
					
					If VarDecl.Value Then
						VarDecl.Value.GenerateCode(Writer)
					Else
						VarDecl.Datatype.GenerateAutoValue(Writer)
					EndIf
					
					Writer.EndStatement()
				ElseIf VarDecl.DeclKind <> KEY_CONST Then
					Writer.AddStatement(VarDecl)
				EndIf
			Else
				Writer.AddStatement(Statement)
			EndIf
		Next
	End Method
	
	Method ReplaceChildNode(Original:TASTNode, Replacement:TASTNode)
		Super.ReplaceChildNode(Original, Replacement)
		
		Local Link:TLink = Constants.FirstLink()
		While Link
			If Link.Value() = Original Then Link._Value = Replacement
			
			Link = Link.NextLink()
		Wend
		
		Link = Statements.FirstLink()
		While Link
			If Link.Value() = Original Then Link._Value = Replacement
			
			Link = Link.NextLink()
		Wend
	End Method
	
	Method SetupChildren()
		AddChildNodeList(Constants)
		AddChildNodeList(Statements)
	End Method
	
	Method Create:TStatementList(List:TList)
		Constants  = New TList
		Statements = New TList
		
		For Local Statement:TStatementNode = EachIn List
			Local VarDecl:TVariableDeclNode = TVariableDeclNode(Statement)
			
			If VarDecl And VarDecl.DeclKind = KEY_CONST Then
				Constants.AddLast(VarDecl)
			Else
				Statements.AddLast(Statement)
			EndIf
		Next
		
		SetupChildren()
		
		Return Self
	End Method
End Type

Type TStatementNode Extends TASTNode
End Type

Type TIfNode Extends TStatementNode
	Field Expression:TExpressionNode
	
	Field Statements:TStatementList
	
	Field ElseIfList:TList
	Field ElseBranch:TElseNode
	
	Method GetName:String()
		Return "TIfNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Local OldScope:TScope = Scope
		
		Expression.CheckCorrectness(Scope)
		
		Scope = New TScope.Create(Scope)
		
		Statements.CheckCorrectness(Scope)
		
		For Local ElseIfNode:TElseIfNode = EachIn ElseIfList
			ElseIfNode.CheckCorrectness(OldScope)
		Next
		
		If ElseBranch Then ElseBranch.CheckCorrectness(OldScope)
	End Method
	
	Method SetupChildren()
		AddChildNode(Expression)
		AddChildNode(Statements)
		AddChildNodeList(ElseIfList)
		If ElseBranch Then AddChildNode(ElseBranch)
	End Method
	
	Method Create:TIfNode(Expression:TExpressionNode, Statements:TStatementList, ElseIfList:TList, ElseBranch:TElseNode)
		Self.Expression = Expression
		Self.Statements = Statements
		Self.ElseIfList = ElseIfList
		Self.ElseBranch = ElseBranch
		
		SetupChildren()
		
		Return Self
	End Method
	
	Function Emit:Int(Expression:Int, Statements:Int, ElseIfList:Int, ElseBranch:Int)
		Local IfStatement:TIfNode = CurrentTarget.CreateIfNode()
		IfStatement.Expression = TExpressionNode(RetrieveASTNode(Expression))
		IfStatement.ElseBranch = TElseNode      (RetrieveASTNode(ElseBranch))
		IfStatement.ElseIfList = RetrieveList(ElseIfList)
		IfStatement.Statements = New TStatementList.Create(RetrieveList(Statements))
		
		If Not IfStatement.ElseIfList Then IfStatement.ElseIfList = New TList
		
		IfStatement.SetupChildren()
		
		Return IfStatement.TableID
	End Function
End Type

Type TElseIfNode Extends TASTNode
	Field Expression:TExpressionNode
	
	Field Statements:TStatementList
	
	Method GetName:String()
		Return "TElseIfNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Expression.CheckCorrectness(Scope)
		
		Scope = New TScope.Create(Scope)
		
		Statements.CheckCorrectness(Scope)
	End Method
	
	Method SetupChildren()
		AddChildNode(Expression)
		AddChildNode(Statements)
	End Method
	
	Method Create:TElseIfNode(Expression:TExpressionNode, Statements:TStatementList)
		Self.Expression = Expression
		Self.Statements = Statements
		
		SetupChildren()
		
		Return Self
	End Method
	
	Function Emit:Int(Expression:Int, Statements:Int)
		Local ElseIfStatement:TElseIfNode = CurrentTarget.CreateElseIfNode()
		ElseIfStatement.Expression = TExpressionNode(RetrieveASTNode(Expression))
		ElseIfStatement.Statements = New TStatementList.Create(RetrieveList(Statements))
		
		ElseIfStatement.SetupChildren()
		
		Return ElseIfStatement.TableID
	End Function
End Type

Type TElseNode Extends TASTNode
	Field Statements:TStatementList
	
	Method GetName:String()
		Return "TElseNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Scope = New TScope.Create(Scope)
		
		Statements.CheckCorrectness(Scope)
	End Method
	
	Method SetupChildren()
		AddChildNode(Statements)
	End Method
	
	Method Create:TElseNode(Statements:TStatementList)
		Self.Statements = Statements
		
		SetupChildren()
		
		Return Self
	End Method
	
	Function Emit:Int(Statements:Int)
		Local ElseStatement:TElseNode = CurrentTarget.CreateElseNode()
		ElseStatement.Statements = New TStatementList.Create(RetrieveList(Statements))
		
		ElseStatement.SetupChildren()
		
		Return ElseStatement.TableID
	End Function
End Type

Type TWhileNode Extends TStatementNode
	Field Expression:TExpressionNode
	
	Field Statements:TStatementList
	
	Method GetName:String()
		Return "TWhileNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Expression.CheckCorrectness(Scope)
		
		Scope = New TScope.Create(Scope)
		
		Statements.CheckCorrectness(Scope)
	End Method
	
	Method SetupChildren()
		AddChildNode(Expression)
		AddChildNode(Statements)
	End Method
	
	Function Emit:Int(Expression:Int, Statements:Int)
		Local WhileNode:TWhileNode = CurrentTarget.CreateWhileNode()
		WhileNode.Expression = TExpressionNode(RetrieveASTNode(Expression))
		WhileNode.Statements = New TStatementList.Create(RetrieveList(Statements))
		
		WhileNode.SetupChildren()
		
		Return WhileNode.TableID
	End Function
End Type

Type TRepeatNode Extends TStatementNode
	Field Expression:TExpressionNode
	
	Field Statements:TStatementList
	
	Method GetName:String()
		Return "TRepeatNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		If Expression Then Expression.CheckCorrectness(Scope)
		
		Scope = New TScope.Create(Scope)
		
		Statements.CheckCorrectness(Scope)
	End Method
	
	Method SetupChildren()
		If Expression Then AddChildNode(Expression)
		AddChildNode(Statements)
	End Method
	
	Function Emit:Int(Expression:Int, Statements:Int)
		Local RepeatNode:TRepeatNode = CurrentTarget.CreateRepeatNode()
		RepeatNode.Expression = TExpressionNode(RetrieveASTNode(Expression))
		RepeatNode.Statements = New TStatementList.Create(RetrieveList(Statements))
		
		RepeatNode.SetupChildren()
		
		Return RepeatNode.TableID
	End Function
End Type

Type TSelectNode Extends TStatementNode
	Field Expression:TExpressionNode
	
	Field Datatype:TDatatype
	
	Field CaseList:TList
	
	Field DefaultCase:TDefaultNode
	
	Method GetName:String()
		Return "TSelectNode"
	End Method
	
	Method PerformASTTransformation(Scope:TScope)
		SetCommonRoot(Root)
		
		Local VarDecl:TVariableDeclNode = CurrentTarget.CreateVariableDeclNode().Create(KEY_LOCAL, "__select" + TableID, Datatype, Expression)
		
		Local FirstCase:TCaseNode = TCaseNode(CaseList.RemoveFirst())
		
		Local ElseIfList:TList = New TList
		For Local CaseNode:TCaseNode = EachIn CaseList
			Local ElseIfNode:TElseIfNode = CurrentTarget.CreateElseIfNode().Create( ..
					EqualExpression(VarDecl, CaseNode.Expression), ..
					CaseNode.Statements ..
				)
			
			ElseIfList.AddLast(ElseIfNode)
		Next
		
		Local ElseNode:TElseNode
		If DefaultCase Then ElseNode = CurrentTarget.CreateElseNode().Create(DefaultCase.Statements)
		
		Local IfNode:TIfNode = CurrentTarget.CreateIfNode().Create( ..
				EqualExpression(VarDecl, FirstCase.Expression), ..
				FirstCase.Statements, ..
				ElseIfList, ..
				ElseNode ..
			)
		
		Local Statements:TList = New TList
		
		Statements.AddLast(VarDecl)
		Statements.AddLast(IfNode)
		
		Local StatementList:TStatementList = New TStatementList.Create(Statements)
		
		Parent.ReplaceChildNode(Self, StatementList)
		
		StatementList.CheckCorrectness(Scope)
		
		Function EqualExpression:TBinaryOpExpression(A:TVariableDeclNode, B:TExpressionNode)
			Local Variable:TVariableNode = CurrentTarget.CreateVariableNode().Create(A.Identifier, A.Datatype.Copy())
			
			Return CurrentTarget.CreateBinaryOpExpression().Create(TOK_EQUAL, Variable, B)
		End Function
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Expression.CheckCorrectness(Scope)
		
		Datatype = Expression.EvaluateDatatype()
		
		For Local CaseNode:TCaseNode = EachIn CaseList
			CaseNode.CheckCorrectness(Scope)
			
			If Not CaseNode.Datatype.IsCastable(Datatype) Then ..
				Abort( ..
					"Unable to compare expression of type '" + ..
					CaseNode.Datatype.ToString() + ..
					"' to Select expression of type '" + ..
					Datatype.ToString() + ..
					"'", Self)
		Next
		
		If DefaultCase Then DefaultCase.CheckCorrectness(Scope)
		
		PerformASTTransformation(Scope)
	End Method
	
	Method SetupChildren()
		AddChildNode(Expression)
		If DefaultCase Then AddChildNode(DefaultCase)
		AddChildNodeList(CaseList)
	End Method
	
	Function Emit:Int(Expression:Int, CaseList1:Int, DefaultCase:Int, CaseList2:Int)
		Local SelectStatement:TSelectNode = CurrentTarget.CreateSelectNode()
		SelectStatement.Expression  = TExpressionNode(RetrieveASTNode(Expression))
		SelectStatement.DefaultCase = TDefaultNode   (RetrieveASTNode(DefaultCase))
		
		If CaseList2 >= 0 Then
			SelectStatement.CaseList = MergeLists(RetrieveList(CaseList1), RetrieveList(CaseList2))
		Else
			SelectStatement.CaseList = RetrieveList(CaseList1)
		EndIf
		
		SelectStatement.SetupChildren()
		
		Return SelectStatement.TableID
	End Function
End Type

Type TCaseNode Extends TASTNode
	Field Expression:TExpressionNode
	
	Field Datatype:TDatatype
	
	Field Statements:TStatementList
	
	Method GetName:String()
		Return "TCaseNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Expression.CheckCorrectness(Scope)
		
		Datatype = Expression.EvaluateDatatype()
		
		Scope = New TScope.Create(Scope)
		
		Statements.CheckCorrectness(Scope)
	End Method
	
	Method SetupChildren()
		AddChildNode(Expression)
		AddChildNode(Statements)
	End Method
	
	Function Emit:Int(Expression:Int, Statements:Int)
		Local CaseNode:TCaseNode = CurrentTarget.CreateCaseNode()
		CaseNode.Expression = TExpressionNode(RetrieveASTNode(Expression))
		CaseNode.Statements = New TStatementList.Create(RetrieveList(Statements))
		
		CaseNode.SetupChildren()
		
		Return CaseNode.TableID
	End Function
End Type

Type TDefaultNode Extends TASTNode
	Field Statements:TStatementList
	
	Method GetName:String()
		Return "TDefaultNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Scope = New TScope.Create(Scope)
		
		Statements.CheckCorrectness(Scope)
	End Method
	
	Method SetupChildren()
		AddChildNode(Statements)
	End Method
	
	Function Emit:Int(Statements:Int)
		Local DefaultNode:TDefaultNode = CurrentTarget.CreateDefaultNode()
		DefaultNode.Statements = New TStatementList.Create(RetrieveList(Statements))
		
		DefaultNode.SetupChildren()
		
		Return DefaultNode.TableID
	End Function
End Type

Type TForWhileNode Extends TStatementNode
	Field ForStatements:TStatementList
	Field WhileExpression:TExpressionNode
	Field Statements:TStatementList
	
	Method GetName:String()
		Return "TForWhileNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Scope = New TScope.Create(Scope)
		
		ForStatements.CheckCorrectness(Scope)
		WhileExpression.CheckCorrectness(Scope)
		WhileExpression.EvaluateDatatype()
		Statements.CheckCorrectness(Scope)
	End Method
	
	Method SetupChildren()
		AddChildNode(ForStatements)
		AddChildNode(WhileExpression)
		AddChildNode(Statements)
	End Method
	
	Function Emit:Int(ForStatements:Int, WhileExpression:Int, Statements:Int)
		Local ForWhileNode:TForWhileNode = CurrentTarget.CreateForWhileNode()
		
		ForWhileNode.ForStatements = New TStatementList.Create(RetrieveList(ForStatements))
		ForWhileNode.Statements    = New TStatementList.Create(RetrieveList(Statements))
		ForWhileNode.WhileExpression = TExpressionNode(RetrieveASTNode(WhileExpression))
		
		ForWhileNode.SetupChildren()
		
		Return ForWhileNode.TableID
	End Function
End Type

Type TForNode Extends TStatementNode
	Field DefineVariable:Int
	Field Variable:TExpressionNode
	
	Field FromExpression:TExpressionNode
	Field ToExpression:TExpressionNode
	Field StepExpression:TExpressionNode
	Field SkipLast:Int
	
	Field Statements:TStatementList
	
	Field LoopVariable:TStatementNode
	
	Method GetName:String()
		Return "TForNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Scope = New TScope.Create(Scope)
		
		If Not DefineVariable Then
			Variable.CheckCorrectness(Scope)
			
			Local Equality:TBinaryOpExpression = TBinaryOpExpression(Variable)
			
			Variable       = Equality.LeftExp
			FromExpression = Equality.RightExp
			
			If Not Variable.IsWritable() Then Abort("Loop variable must be writable", Variable)
			
			LoopVariable = Variable
		Else
			SetCommonRoot(Root)
			LoopVariable = CurrentTarget.CreateVariableDeclNode().Create(KEY_LOCAL, TVariableNode(Variable).Identifier, Variable.EvaluateDatatype(), FromExpression)
			AddChildNode(LoopVariable)
			
			TVariableNode(Variable).Origin = TVariableDeclNode(LoopVariable)
			
			LoopVariable.CheckCorrectness(Scope)
		EndIf
		
		If Not Variable.EvaluateDatatype().IsNumeric() Then Abort("Loop variable must be of numeric type", Variable)
		
		FromExpression.CheckCorrectness(Scope)
		ToExpression  .CheckCorrectness(Scope)
		
		If Not FromExpression.EvaluateDatatype().IsNumeric() Then Abort("From expression must be of numeric type", FromExpression)
		If Not   ToExpression.EvaluateDatatype().IsNumeric() Then Abort(  "To expression must be of numeric type", ToExpression)
		
		If StepExpression Then
			StepExpression.CheckCorrectness(Scope)
			If Not StepExpression.EvaluateDatatype().IsNumeric() Then Abort("Step expression must be of numeric type", StepExpression)
			If Not StepExpression.IsConstant() Then Abort("Step expression must be constant", StepExpression)
		EndIf
		
		Statements.CheckCorrectness(Scope)
	End Method
	
	Method SetupChildren()
		AddChildNode(FromExpression)
		AddChildNode(ToExpression)
		If StepExpression Then AddChildNode(StepExpression)
		AddChildNode(Statements)
		AddChildNode(Variable)
	End Method
	
	Function Emit:Int(DefineVariable:Int, LoopVariable:Int, FromExpression:Int, ToExpression:Int, StepExpression:Int, SkipLast:Int, Statements:Int)
		Local ForStatement:TForNode = CurrentTarget.CreateForNode()
		ForStatement.FromExpression = TExpressionNode(RetrieveASTNode(FromExpression))
		ForStatement.  ToExpression = TExpressionNode(RetrieveASTNode(  ToExpression))
		ForStatement.StepExpression = TExpressionNode(RetrieveASTNode(StepExpression))
		ForStatement.Variable       = TExpressionNode(RetrieveASTNode(  LoopVariable))
		ForStatement.SkipLast       = SkipLast
		ForStatement.DefineVariable = DefineVariable
		ForStatement.Statements     = New TStatementList.Create(RetrieveList(Statements))
		
		ForStatement.SetupChildren()
		
		Return ForStatement.TableID
	End Function
End Type

Type TExitNode Extends TStatementNode
	Method GetName:String()
		Return "TExitNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Local P:TASTNode = Parent
		While P
			If TWhileNode(P) Or TRepeatNode(P) Or TForNode(P) Then Return
			
			Local ForWhileNode:TForWhileNode = TForWhileNode(P)
			If ForWhileNode And (Not HasParentNode(ForWhileNode.ForStatements)) Then Return
			
			P = P.Parent
		Wend
		
		Abort("Exit must appear inside a loop", Self)
	End Method
	
	Function Emit:Int()
		Return CurrentTarget.CreateExitNode().TableID
	End Function
End Type

Type TContinueNode Extends TStatementNode
	Method GetName:String()
		Return "TContinueNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Local P:TASTNode = Parent
		While P
			If TWhileNode(P) Or TRepeatNode(P) Or TForNode(P) Then Return
			
			Local ForWhileNode:TForWhileNode = TForWhileNode(P)
			If ForWhileNode And (Not HasParentNode(ForWhileNode.ForStatements)) Then Return
			
			P = P.Parent
		Wend
		
		Abort("Continue must appear inside a loop", Self)
	End Method
	
	Function Emit:Int()
		Return CurrentTarget.CreateContinueNode().TableID
	End Function
End Type

Type TAssignmentNode Extends TStatementNode
	Field LeftExp:TExpressionNode
	Field RightExp:TExpressionNode
	
	Method GetName:String()
		Return "TAssignmentNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		RightExp.CheckCorrectness(Scope)
		LeftExp .CheckCorrectness(Scope)
		
		Local RightType:TDatatype = RightExp.EvaluateDatatype()
		Local  LeftType:TDatatype =  LeftExp.EvaluateDatatype()
		
		If Not RightType.IsCastable(LeftType) Then ..
			Abort("Incompatible datatypes in assignment: " + ..
				"Unable to convert from " + RightType.ToString() + " to " + ..
				LeftType.ToString(), RightExp)
		
		If Not LeftExp.IsWritable() Then Abort("Assignment target is not writable", LeftExp)
		
		SetCommonRoot(Root)
		If LeftExp.EvaluateDatatype().IsUnequal(RightExp.EvaluateDatatype()) Then ..
			RightExp = CurrentTarget.CreateTypeCastExpression().Create(RightExp, LeftExp.EvaluateDatatype().Copy(), Scope)
	End Method
	
	Method SetupChildren()
		AddChildNode(LeftExp)
		AddChildNode(RightExp)
	End Method
	
	Function Emit:Int(BinaryOp:Int)
		Local Assignment:TAssignmentNode = CurrentTarget.CreateAssignmentNode()
		
		Local Original:TBinaryOpExpression = TBinaryOpExpression(RetrieveASTNode(BinaryOp))
		
		Assignment.LeftExp  = Original.LeftExp
		Assignment.RightExp = Original.RightExp
		
		Assignment.SetupChildren()
		
		Return Assignment.TableID
	End Function
End Type

Type TReturnNode Extends TStatementNode
	Field TargetExp:TExpressionNode
	
	Method GetName:String()
		Return "TReturnNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Local P:TASTNode = Parent, FunctionNode:TFunctionNode
		While P
			FunctionNode = TFunctionNode(P)
			
			If FunctionNode Then Exit
			
			P = P.Parent
		Wend
		
		If FunctionNode Then
			If FunctionNode.IsConstructor Then
				If TargetExp Then ..
					Abort("Constructor does not return a value", Self)
				
				SetCommonRoot(Root)
				Local SelfNode:TLiteralNode = CurrentTarget.CreateLiteralNode().Create(FunctionNode.ReturnDatatype.Copy(), Null)
				SelfNode.LiteralKind = KEY_SELF
				
				TargetExp = SelfNode
				AddChildNode(SelfNode)
			ElseIf FunctionNode.ReturnDatatype.IsVoid() Then
				If TargetExp Then Abort("Function does not return a value", Self)
			Else
				If Not TargetExp Then Abort("Function must return a value", Self)
				
				TargetExp.CheckCorrectness(Scope)
				If Not TargetExp.EvaluateDatatype().IsCastable(FunctionNode.ReturnDatatype) Then ..
					Abort("Return datatype does not match function datatype", Self)
				
				SetCommonRoot(Root)
				If TargetExp.EvaluateDatatype().IsUnequal(FunctionNode.ReturnDatatype) Then ..
					TargetExp = CurrentTarget.CreateTypeCastExpression().Create(TargetExp, FunctionNode.ReturnDatatype.Copy(), Scope)
			EndIf
		Else
			Abort("Return may only be used inside functions or methods", Self)
		EndIf
	End Method
	
	Method SetupChildren()
		If TargetExp Then AddChildNode(TargetExp)
	End Method
	
	Function Emit:Int(TargetExp:Int)
		Local ReturnNode:TReturnNode = CurrentTarget.CreateReturnNode()
		ReturnNode.TargetExp = TExpressionNode(RetrieveASTNode(TargetExp))
		
		ReturnNode.SetupChildren()
		
		Return ReturnNode.TableID
	End Function
End Type

Type TDeleteNode Extends TStatementNode
	Field TargetExp:TExpressionNode
	
	Method GetName:String()
		Return "TDeleteNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		TargetExp.CheckCorrectness(Scope)
		
		If Not TargetExp.EvaluateDatatype().IsCompound() Then ..
			Abort("Delete can only be applied to class instances", TargetExp)
	End Method
	
	Method SetupChildren()
		AddChildNode(TargetExp)
	End Method
	
	Function Emit:Int(TargetExp:Int)
		Local DeleteNode:TDeleteNode = CurrentTarget.CreateDeleteNode()
		DeleteNode.TargetExp = TExpressionNode(RetrieveASTNode(TargetExp))
		
		DeleteNode.SetupChildren()
		
		Return DeleteNode.TableID
	End Function
End Type

'Das hier ist extremst beschissen. Expressions sind zwar keine Statements und werden vom Parser auch nicht erlaubt,
'Aber das muss ich leider so machen, da ein FunctionCall sowohl ein Ausdruck als auch ein Statement ist
'Da BMax keine Mehrfachvererbung untersttzt, msste ich entweder FunctionCall duplizieren oder halt wie jetzt
'die Hierarchie zurechtbiegen
Type TExpressionNode Extends TStatementNode
	Field GenericResult:Long
	
	Method EvaluateDatatype:TDatatype() Abstract
	Method EvaluateResult() Abstract
	Method IsLValue  :Int() Abstract
	Method IsWritable:Int() Abstract
	Method IsConstant:Int() Abstract
	
	Method SetIntResult(Data:Long)
		GenericResult = Data
	End Method
	
	Method SetFloatResult(Data:Float)
		(Float Ptr (Varptr GenericResult))[0] = Data
	End Method
	
	Method SetDoubleResult(Data:Double)
		(Double Ptr (Varptr GenericResult))[0] = Data
	End Method
	
	Method GetIntResult:Long()
		Return GenericResult
	End Method
	
	Method GetFloatResult:Float()
		Return (Float Ptr (Varptr GenericResult))[0]
	End Method
	
	Method GetDoubleResult:Double()
		Return (Double Ptr (Varptr GenericResult))[0]
	End Method
	
	Method GetCastedIntResult:Long()
		Local Datatype:TDatatype = EvaluateDatatype()
		
		If Datatype.IsInteger() Then
			Return GenericResult
		ElseIf Datatype.TypeKind = KEY_FLOAT Then
			Return GetFloatResult()
		ElseIf Datatype.TypeKind = KEY_DOUBLE Then
			Return GetDoubleResult()
		EndIf
	End Method
	
	Method GetCastedDoubleResult:Double()
		Local Datatype:TDatatype = EvaluateDatatype()
		
		If Datatype.IsInteger() Then
			Return GenericResult
		ElseIf Datatype.TypeKind = KEY_FLOAT Then
			Return GetFloatResult()
		ElseIf Datatype.TypeKind = KEY_DOUBLE Then
			Return GetDoubleResult()
		EndIf
	End Method
	
	Method GetCastedStringResult:String()
		Local Datatype:TDatatype = EvaluateDatatype()
		
		If Datatype.IsInteger() Then
			Return GenericResult
		ElseIf Datatype.TypeKind = KEY_FLOAT Then
			Return GetFloatResult()
		ElseIf Datatype.TypeKind = KEY_DOUBLE Then
			Return GetDoubleResult()
		EndIf
	End Method
End Type

Type TAddressExpression Extends TExpressionNode
	Field TargetExp:TExpressionNode
	Field Datatype:TDatatype
	
	Method GetName:String()
		Return "TAddressExpression"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
	End Method
	
	Method IsLValue:Int()
		Return False
	End Method
	
	Method IsWritable:Int()
		Return False
	End Method
	
	Method IsConstant:Int()
		Return False
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		TargetExp.CheckCorrectness(Scope)
		
		If TargetExp.IsLValue() Then
			SetCommonRoot(Root)
			Datatype = CurrentTarget.CreateDatatype().Create(TYPE_POINTER, TargetExp.EvaluateDatatype().Copy())
			
			AddChildNode(Datatype)
		Else
			Abort("Expression is not addressable", TargetExp)
		EndIf
	End Method
	
	Method SetupChildren()
		AddChildNode(TargetExp)
	End Method
	
	Function Emit:Int(Expression:Int)
		Local AddressExp:TAddressExpression = CurrentTarget.CreateAddressExpression()
		
		AddressExp.TargetExp = TExpressionNode(RetrieveASTNode(Expression))
		
		AddressExp.SetupChildren()
		
		Return AddressExp.TableID
	End Function
End Type

Type TAutoArrayExpression Extends TExpressionNode
	Field Values:TExpressionNode[]
	Field Datatype:TDatatype
	
	Method GetName:String()
		Return "TAutoArrayExpression"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
	End Method
	
	Method IsLValue:Int()
		Return False
	End Method
	
	Method IsWritable:Int()
		Return False
	End Method
	
	Method IsConstant:Int()
		Return False
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Local ElementType:TDatatype
		For Local Value:TExpressionNode = EachIn Values
			Value.CheckCorrectness(Scope)
			
			If ElementType Then
				If ElementType.IsUnequal(Value.EvaluateDatatype()) Then ..
					Abort("Auto-array elements must have identical types", Self)
			Else
				ElementType = Value.EvaluateDatatype()
			EndIf
		Next
		
		SetCommonRoot(Root)
		If Not ElementType Then ElementType = CurrentTarget.CreateDatatype().Create(TYPE_ANY, Null)
		
		Datatype = CurrentTarget.CreateDatatype().Create(TYPE_ARRAY, ElementType, 1)
		AddChildNode(Datatype)
		
		Datatype.CheckCorrectness(Scope) 'Make sure we even have an array class available
	End Method
	
	Method SetupChildren()
		AddChildNodeArray(Values)
	End Method
	
	Function Emit:Int(Values:Int)
		Local AutoArray:TAutoArrayExpression = CurrentTarget.CreateAutoArrayExpression()
		
		Local ValueList:TList = RetrieveList(Values)
		AutoArray.Values = New TExpressionNode[ValueList.Count()]
		
		Local I:Int = 0
		For Local Value:TExpressionNode = EachIn ValueList
			AutoArray.Values[I] = Value
			I :+ 1
		Next
		
		AutoArray.SetupChildren()
		
		Return AutoArray.TableID
	End Function
End Type

Type TNewArrayExpression Extends TExpressionNode
	Field Sizes:TExpressionNode[]
	Field Datatype:TDatatype
	
	Method GetName:String()
		Return "TNewArrayExpression"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
	End Method
	
	Method IsLValue:Int()
		Return False
	End Method
	
	Method IsWritable:Int()
		Return False
	End Method
	
	Method IsConstant:Int()
		Return False
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		For Local Size:TExpressionNode = EachIn Sizes
			Size.CheckCorrectness(Scope)
		Next
		
		Datatype.CheckCorrectness(Scope) 'Make sure we even have an array class available
	End Method
	
	Method SetupChildren()
		AddChildNodeArray(Sizes)
		AddChildNode(Datatype)
	End Method
	
	Function Emit:Int(Datatype:Int, Sizes:Int)
		Local NewArray:TNewArrayExpression = CurrentTarget.CreateNewArrayExpression()
		
		Local SizeList:TList = RetrieveList(Sizes)
		NewArray.Sizes = New TExpressionNode[SizeList.Count()]
		
		Local I:Int = 0
		For Local Size:TExpressionNode = EachIn SizeList
			NewArray.Sizes[I] = Size
			I :+ 1
		Next
		
		NewArray.Datatype = CurrentTarget.CreateDatatype().Create(TYPE_ARRAY, RetrieveASTNode(Datatype), NewArray.Sizes.Length)
		
		NewArray.SetupChildren()
		
		Return NewArray.TableID
	End Function
End Type

Type TIndexExpression Extends TExpressionNode
	Field IsArray:Int
	
	Field TargetExp:TExpressionNode
	Field Indices:TExpressionNode[]
	
	Field Datatype:TDatatype
	
	Method GetName:String()
		Return "TIndexExpression"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
	End Method
	
	Method IsLValue:Int()
		Return True
	End Method
	
	Method IsWritable:Int()
		Return True
	End Method
	
	Method IsConstant:Int()
		Return False
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		TargetExp.CheckCorrectness(Scope)
		
		For Local Index:TExpressionNode = EachIn Indices
			Index.CheckCorrectness(Scope)
			
			If Not Index.EvaluateDatatype().IsInteger() Then ..
				Abort("Index expression must be an integer", Index)
		Next
		
		Local TargetType:TDatatype = TargetExp.EvaluateDatatype()
		If TargetType.IsPointer() Then
			If Indices.Length <> 1 Then Abort("Invalid number of dimensions", Self)
			
			Datatype = TargetType.NestedType
			
			If Datatype.IsVoid() Then Abort("Dereferencing void pointers is not allowed", Self)
		ElseIf TargetType.IsArray() Then
			IsArray = True
			
			If TargetType.Dimension <> Indices.Length Then ..
				Abort("Invalid number of dimensions", Self)
			
			Datatype = TargetType.NestedType
		Else
			Abort("Expression is not indexable", TargetExp)
		EndIf
	End Method
	
	Method SetupChildren()
		AddChildNode(TargetExp)
		AddChildNodeArray(Indices)
	End Method
	
	Function Emit:Int(TargetExp:Int, Index:Int)
		Local IndexExp:TIndexExpression = CurrentTarget.CreateIndexExpression()
		
		IndexExp.TargetExp = TExpressionNode(RetrieveASTNode(TargetExp))
		
		Local IndexList:TList = RetrieveList(Index)
		IndexExp.Indices = New TExpressionNode[IndexList.Count()]
		
		Local I:Int
		For Local Expression:TExpressionNode = EachIn IndexList
			IndexExp.Indices[I] = Expression
			I :+ 1
		Next
		
		IndexExp.SetupChildren()
		
		Return IndexExp.TableID
	End Function
End Type

Type TFunctionCall Extends TExpressionNode
	Field IsStatement:Int
	
	Field TargetExp:TExpressionNode
	Field Datatype:TDatatype
	
	Field Parameters:TList
	
	Field TargetFunction:TFunctionNode
	
	Method GetName:String()
		Return "TFunctionCall"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		If Datatype.IsVoid() Then Abort("Function does not return a value", Self)
		
		Return Datatype
	End Method
	
	Method EvaluateResult()
	End Method
	
	Method IsLValue:Int()
		Return False
	End Method
	
	Method IsWritable:Int()
		Return False
	End Method
	
	Method IsConstant:Int()
		Return False
	End Method
	
	'Typecasts to a class are recognized as function calls by the parser
	'This method replaces the function call by a typecast, if required
	Method TransformToTypecast:Int(Scope:TScope)
		If Parameters.Count() <> 1 Then Return False
		
		Local VariableNode:TVariableNode = TVariableNode(TargetExp)
		If Not VariableNode Then Return False
		
		Local ClassNode:TClassNode = TClassNode(Scope.Find(VariableNode.Identifier))
		If Not ClassNode Then Return False
		
		SetCommonRoot(Root)
		
		Local Datatype:TDatatype = CurrentTarget.CreateDatatype().Create(TYPE_COMPOUND, ClassNode)
		Local TypeCast:TTypeCastExpression = CurrentTarget.CreateTypeCastExpression().Create(TExpressionNode(Parameters.First()), Datatype, Scope)
		TypeCast.CopyLocation(Self)
		
		Parent.ReplaceChildNode(Self, TypeCast)
		
		TypeCast.CheckCorrectness(Scope)
		
		Return True
	End Method
	
	'Constructor calls of the form Super(....) or Self(....) need to be caught and transformed as well
	Method TransformToConstructorCall(Scope:TScope)
		Local LiteralNode:TLiteralNode = TLiteralNode(TargetExp)
		
		If LiteralNode And (LiteralNode.IsSelf() Or LiteralNode.IsSuper()) Then
			LiteralNode.CheckCorrectness(Scope) 'Required to make the call to EvaluateDatatype safe
			
			'TODO: Function calls of form Super() should always work - the default constructor should always exist, even if the user doesn't define one
			Local ClassNode:TClassNode = LiteralNode.EvaluateDatatype().Compound
			If Not ClassNode.Constructor Then Abort("No constructor available for class '" + ClassNode.Identifier + "'", Self)
			
			SetCommonRoot(Root)
			
			Local MemberAccess:TMemberAccessExpression = CurrentTarget.CreateMemberAccessExpression().Create( ..
					LiteralNode, ..
					ClassNode.GenerateConstructorIdentifier() ..
				)
			
			MemberAccess.CopyLocation(LiteralNode)
			
			ReplaceChildNode(LiteralNode, MemberAccess)
		EndIf
	End Method
	
	Method ReplaceChildNode(Original:TASTNode, Replacement:TASTNode)
		For Local Param:TASTNode = EachIn Parameters
			If Param = Original Then
				Local Link:TLink = Parameters.FindLink(Param)
				
				Link._value = Replacement
			EndIf
		Next
		
		Super.ReplaceChildNode(Original, Replacement)
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		If TransformToTypecast(Scope) Then Return
		TransformToConstructorCall(Scope)
		
		SetCommonRoot(Root)
		
		TargetExp.CheckCorrectness(Scope)
		
		Local FunctionType:TDatatype = TargetExp.EvaluateDatatype()
		
		If Not FunctionType.IsFunction() Then Abort("Expression of type '" + FunctionType.ToString() + "' cannot be called as a function", TargetExp)
		
		For Local Expression:TExpressionNode = EachIn Parameters
			Expression.CheckCorrectness(Scope)
		Next
		
		If Parameters.Count() > FunctionType.ParameterCount Then Abort("Too many arguments", Self)
		
		Local NewParameters:TList = New TList
		
		Local I:Int = 0
		Local Link:TLink = Parameters.FirstLink()
		While Link
			Local Expression:TExpressionNode = TExpressionNode(Link.Value())
			Local ExpType:TDatatype = Expression.EvaluateDatatype()
			
			If Not ExpType.IsCastable(FunctionType.Parameters[I]) Then ..
				Abort( ..
					"Invalid datatype for argument '" + ..
					FunctionType.Identifiers[I] + ..
					"': Expected '" + ..
					FunctionType.Parameters[I].ToString() + ..
					"' but received '" + ..
					ExpType.ToString() + ..
					"'", ..
					Expression ..
				)
			
			If ExpType.IsUnequal(FunctionType.Parameters[I]) Then
				NewParameters.AddLast(CurrentTarget.CreateTypeCastExpression().Create(Expression, FunctionType.Parameters[I].Copy(), Scope))
			Else
				NewParameters.AddLast(Expression)
			EndIf
			
			I :+ 1
			Link = Link.NextLink()
		Wend
		
		If I < FunctionType.MandatoryParamCount Then Abort("Too few arguments", Self)
		
		Parameters = NewParameters
		
		Datatype = FunctionType.ReturnType
	End Method
	
	Method SetupChildren()
		AddChildNode(TargetExp)
		AddChildNodeList(Parameters)
	End Method
	
	Method Create:TFunctionCall(TargetExp:TExpressionNode, Parameters:TList, IsStatement:Int)
		Self.TargetExp   = TargetExp
		Self.Parameters  = Parameters
		Self.IsStatement = IsStatement
		
		If Not Parameters Then Self.Parameters = New TList
		
		SetupChildren()
		
		Return Self
	End Method
	
	Function Emit:Int(ExpressionIndex:Int, ParameterList:Int, IsStatement:Int)
		Local FunctionCall:TFunctionCall = CurrentTarget.CreateFunctionCall()
		
		If IsStatement = -1 Then
			Local Original:TFunctionCall = TFunctionCall(RetrieveASTNode(ExpressionIndex))
			
			FunctionCall.IsStatement = 1
			FunctionCall.TargetExp   = Original.TargetExp
			FunctionCall.Parameters  = Original.Parameters
		Else
			FunctionCall.IsStatement = IsStatement
			FunctionCall.TargetExp   = TExpressionNode(RetrieveASTNode(ExpressionIndex))
			FunctionCall.Parameters  = RetrieveList(ParameterList)
		EndIf
		
		If Not FunctionCall.Parameters Then FunctionCall.Parameters = New TList
		
		FunctionCall.SetupChildren()
		
		Return FunctionCall.TableID
	End Function
End Type

Type TTypeCastExpression Extends TExpressionNode
	Field Datatype:TDatatype
	Field TargetExp:TExpressionNode
	
	Method GetName:String()
		Return "TTypeCastExpression"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		TargetExp.CheckCorrectness(Scope)
		
		Local SourceType:TDatatype = TargetExp.EvaluateDatatype()
		
		If Not SourceType.IsExplicitCastable(Datatype) Then ..
			Abort("Unable to cast from '" + SourceType.ToString() + "' to '" + Datatype.ToString() + "'", Self)
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
		TargetExp.EvaluateResult()
		
		Local SourceType:TDatatype = TargetExp.EvaluateDatatype()
		
		If SourceType.IsInteger() Then
			Select Datatype.TypeKind
				Case TYPE_BOOL   ; SetIntResult(0 <> (TargetExp.GenericResult))
				Case TYPE_BYTE   ; SetIntResult(Byte (TargetExp.GenericResult))
				Case TYPE_SHORT  ; SetIntResult(Short(TargetExp.GenericResult))
				Case TYPE_INT    ; SetIntResult(Int  (TargetExp.GenericResult))
				Case TYPE_LONG   ; SetIntResult(Long (TargetExp.GenericResult))
				Case TYPE_FLOAT  ; SetFloatResult    (TargetExp.GenericResult)
				Case TYPE_DOUBLE ; SetDoubleResult   (TargetExp.GenericResult)
				Case TYPE_POINTER; SetIntResult(Long (TargetExp.GenericResult))
			End Select
		ElseIf SourceType.IsReal() Then
			Local Source:Double = TargetExp.GetCastedDoubleResult()
			
			Select Datatype.TypeKind
				Case TYPE_BOOL   ; SetIntResult(0 <> (TargetExp.GenericResult))
				Case TYPE_BYTE   ; SetIntResult(Byte (Source))
				Case TYPE_SHORT  ; SetIntResult(Short(Source))
				Case TYPE_INT    ; SetIntResult(Int  (Source))
				Case TYPE_LONG   ; SetIntResult(Long (Source))
				Case TYPE_FLOAT  ; SetFloatResult    (Source)
				Case TYPE_DOUBLE ; SetDoubleResult   (Source)
			End Select
		EndIf
	End Method
	
	Method IsLValue:Int()
		Return False
	End Method
	
	Method IsWritable:Int()
		Return False
	End Method
	
	Method IsConstant:Int()
		Return (TargetExp.IsConstant() And Datatype.IsNumeric())
	End Method
	
	Method SetupChildren()
		AddChildNode(TargetExp)
		AddChildNode(Datatype)
	End Method
	
	Method Create:TTypeCastExpression(TargetExp:TExpressionNode, Datatype:TDatatype, Scope:TScope)
		Self.Datatype = Datatype
		Self.TargetExp = TargetExp
		
		SetParentNode(TargetExp.Parent)
		
		SetupChildren()
		
		Datatype.CheckCorrectness(Scope)
		
		Return Self
	End Method
	
	Function Emit:Int(Datatype:Int, TargetExp:Int)
		Local TypeCast:TTypeCastExpression = CurrentTarget.CreateTypeCastExpression()
		
		TypeCast.Datatype  = TDatatype      (RetrieveASTNode(Datatype))
		TypeCast.TargetExp = TExpressionNode(RetrieveASTNode(TargetExp))
		
		TypeCast.SetupChildren()
		
		Return TypeCast.TableID
	End Function
End Type

Type TBinaryOpExpression Extends TExpressionNode
	Field Operator:Int
	
	Field  LeftExp:TExpressionNode
	Field RightExp:TExpressionNode
	
	Field Datatype:TDatatype
	
	Method GetName:String()
		Return "TBinaryOpExpression"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
		 LeftExp.EvaluateResult()
		RightExp.EvaluateResult()
		
		Local  LeftType:TDatatype =  LeftExp.EvaluateDatatype()
		Local RightType:TDatatype = RightExp.EvaluateDatatype()
		
		If LeftType.IsPointer() Or RightType.IsPointer() Then Abort("Not yet implemented", Self)
		
		If LeftType.IsInteger() And RightType.IsInteger() Then
			Local  LeftResult:Long =  LeftExp.GenericResult
			Local RightResult:Long = RightExp.GenericResult
			
			Local Mask:Long
			Select Datatype.TypeKind
				Case TYPE_BYTE
					Mask = $FF
				Case TYPE_SHORT
					Mask = $FFFF
				Case TYPE_INT
					Mask = $FFFFFFFF
				Case TYPE_LONG
					Mask = $FFFFFFFFFFFFFFFF:Long
			End Select
			
			Select Operator
				Case TOK_PLUS        ; GenericResult = LeftResult +   RightResult
				Case TOK_MINUS       ; GenericResult = LeftResult -   RightResult
				Case TOK_ASTERISK    ; GenericResult = LeftResult *   RightResult
				Case TOK_SLASH       ; GenericResult = LeftResult /   RightResult
				Case TOK_TILDE       ; GenericResult = LeftResult ~   RightResult
				Case TOK_AMPERSAND   ; GenericResult = LeftResult &   RightResult
				Case TOK_PIPE        ; GenericResult = LeftResult |   RightResult
				Case TOK_SMALLEREQUAL; GenericResult = LeftResult <=  RightResult
				Case TOK_GREATEREQUAL; GenericResult = LeftResult >=  RightResult
				Case TOK_SMALLER     ; GenericResult = LeftResult <   RightResult
				Case TOK_GREATER     ; GenericResult = LeftResult >   RightResult
				Case KEY_MOD         ; GenericResult = LeftResult Mod RightResult
				Case KEY_AND         ; GenericResult = LeftResult And RightResult
				Case KEY_OR          ; GenericResult = LeftResult Or  RightResult
				Case KEY_SHL         ; GenericResult = LeftResult Shl RightResult
				Case KEY_SHR         ; GenericResult = LeftResult Shr RightResult
				Case TOK_EQUAL       ; GenericResult = LeftResult =   RightResult
				Case TOK_UNEQUAL     ; GenericResult = LeftResult <>  RightResult
				Case KEY_SAR         ; GenericResult = LeftResult Sar RightResult
			End Select
			
			If Datatype.TypeKind = TYPE_BOOL Then
				GenericResult = GenericResult <> 0
			Else
				GenericResult :& Mask
			EndIf
		ElseIf LeftType.IsNumeric() And RightType.IsNumeric() Then
			Local  LeftResult:Double =  LeftExp.GetCastedDoubleResult()
			Local RightResult:Double = RightExp.GetCastedDoubleResult()
			
			Local Result:Double
			Select Operator
				Case TOK_PLUS        ; Result = LeftResult +   RightResult
				Case TOK_MINUS       ; Result = LeftResult -   RightResult
				Case TOK_ASTERISK    ; Result = LeftResult *   RightResult
				Case TOK_SLASH       ; Result = LeftResult /   RightResult
				Case TOK_SMALLEREQUAL; Result = LeftResult <=  RightResult
				Case TOK_GREATEREQUAL; Result = LeftResult >=  RightResult
				Case TOK_SMALLER     ; Result = LeftResult <   RightResult
				Case TOK_GREATER     ; Result = LeftResult >   RightResult
				Case KEY_MOD         ; Result = LeftResult Mod RightResult
				Case KEY_AND         ; Result = LeftResult And RightResult
				Case KEY_OR          ; Result = LeftResult Or  RightResult
				Case TOK_EQUAL       ; Result = LeftResult =   RightResult
				Case TOK_UNEQUAL     ; Result = LeftResult <>  RightResult
			End Select
			
			If Datatype.IsInteger() Then
				SetIntResult(Result)
			ElseIf Datatype.TypeKind = TYPE_FLOAT Then
				SetFloatResult(Result)
			ElseIf Datatype.TypeKind = TYPE_DOUBLE Then
				SetDoubleResult(Result)
			EndIf
		EndIf
	End Method
	
	Method IsLValue:Int()
		Return False
	End Method
	
	Method IsWritable:Int()
		Return False
	End Method
	
	Method IsConstant:Int()
		Return (LeftExp.IsConstant() And RightExp.IsConstant()) And Datatype.TypeKind <> TYPE_STRING
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		SetCommonRoot(Root)
		
		LeftExp.CheckCorrectness(Scope)
		RightExp.CheckCorrectness(Scope)
		
		Local LeftType:TDatatype  = LeftExp .EvaluateDatatype()
		Local RightType:TDatatype = RightExp.EvaluateDatatype()
		
		If (LeftType.IsInteger() Or LeftType.IsPointer()) And (RightType.IsInteger() Or RightType.IsPointer()) Then
			Select Operator
				Case TOK_CARET
					If LeftType.IsPointer() Or RightType.IsPointer() Then Abort("Operator ^ not applicable to pointers", Self)
					
					Datatype = CurrentTarget.CreateDatatype().Create(TYPE_DOUBLE, Null)
					
					Local PowerOf:TVariableDeclNode = TVariableDeclNode(Scope.Find("PowerOf"))
					If Not PowerOf Then
						Abort("PowerOf function not available!", Self)
					Else
						AddDependency(PowerOf, Self)
					EndIf
				Case TOK_EQUAL, TOK_UNEQUAL, TOK_GREATER, TOK_SMALLER, TOK_GREATEREQUAL, TOK_SMALLEREQUAL, KEY_AND, KEY_OR
					Datatype = CurrentTarget.CreateDatatype().Create(TYPE_BOOL, Null)
				Default
					If LeftType.IsPointer() And RightType.IsPointer() And LeftType.IsUnequal(RightType) Then ..
						Abort("Pointer arithmetic on pointers of different types is not allowed", Self)
					
					Datatype = TDatatype.Promote(LeftType, RightType)
			End Select
		ElseIf LeftType.IsReal() Or RightType.IsReal() Then
			If LeftType.IsPointer() Or RightType.IsPointer() Then ..
				Abort("Pointer arithmetic on floating point types is not allowed", Self)
			
			Select Operator
				Case TOK_TILDE, TOK_AMPERSAND, TOK_PIPE, KEY_SHL, KEY_SHR, KEY_SAR
					Abort("'" + TLexer.TokenToString(Operator) + "' only applicable to integers", Self)
				Case TOK_EQUAL, TOK_UNEQUAL, TOK_GREATER, TOK_SMALLER, TOK_GREATEREQUAL, TOK_SMALLEREQUAL, KEY_AND, KEY_OR
					Datatype = CurrentTarget.CreateDatatype().Create(TYPE_BOOL, Null)
				Default
					Datatype = TDatatype.Promote(LeftType, RightType)
			End Select
		ElseIf LeftType.IsString() Or RightType.IsString() Then
			If Operator = TOK_PLUS Then
				Datatype = CurrentTarget.CreateDatatype().Create(TYPE_STRING, Null)
			ElseIf Operator = TOK_EQUAL Then
				Datatype = CurrentTarget.CreateDatatype().Create(TYPE_BOOL, Null)
			Else
				Abort("'" + TLexer.TokenToString(Operator) + "' is not applicable to strings", Self)
			EndIf
		ElseIf (LeftType.IsCompound() Or LeftType.IsAny()) And (RightType.IsCompound() Or RightType.IsAny()) Then
			Select Operator
				Case TOK_EQUAL, TOK_UNEQUAL
					Datatype = CurrentTarget.CreateDatatype().Create(TYPE_BOOL, Null)
				Default
					Abort("'" + TLexer.TokenToString(Operator) + "' is not applicable to classes", Self)
			End Select
		Else
			Abort(..
				"'" + TLexer.TokenToString(Operator) + ..
				"' is not applicable to expressions of type '" + ..
				LeftType.ToString() + "' and '" + RightType.ToString() + ..
				"'", ..
				Self ..
			)
		EndIf
		
		AddChildNode(Datatype)
	End Method
	
	Method SetupChildren()
		AddChildNode(LeftExp)
		AddChildNode(RightExp)
	End Method
	
	Method Create:TBinaryOpExpression(Operator:Int, LeftExp:TExpressionNode, RightExp:TExpressionNode)
		Self.Operator = Operator
		Self.LeftExp  = LeftExp
		Self.RightExp = RightExp
		
		SetupChildren()
		
		Return Self
	End Method
	
	Function Emit:Int(Operator:Int, LeftExp:Int, RightExp:Int)
		Local Expression:TBinaryOpExpression = CurrentTarget.CreateBinaryOpExpression()
		Expression.Operator = Operator
		Expression.LeftExp  = TExpressionNode(RetrieveASTNode(LeftExp))
		Expression.RightExp = TExpressionNode(RetrieveASTNode(RightExp))
		
		Expression.SetupChildren()
		
		Return Expression.TableID
	End Function
End Type

Type TUnaryOpExpression Extends TExpressionNode
	Field Operator:Int
	
	Field TargetExp:TExpressionNode
	
	Field Datatype:TDatatype
	
	Method GetName:String()
		Return "TUnaryOpExpression"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
		TargetExp.EvaluateResult()
		
		Select Operator
			Case TOK_MINUS
				If Datatype.IsInteger() Or Datatype.IsPointer() Then
					GenericResult = -TargetExp.GenericResult
				ElseIf Datatype.TypeKind = KEY_FLOAT Then
					SetFloatResult(-TargetExp.GetFloatResult())
				ElseIf Datatype.TypeKind = KEY_DOUBLE Then
					SetDoubleResult(-TargetExp.GetDoubleResult())
				EndIf
			Case TOK_TILDE
				Select Datatype.TypeKind
					Case TYPE_BOOL
						GenericResult = Not TargetExp.GenericResult
					Case TYPE_BYTE
						GenericResult = (~TargetExp.GenericResult) & $FF
					Case TYPE_SHORT
						GenericResult = (~TargetExp.GenericResult) & $FFFF
					Case TYPE_INT, TYPE_POINTER 'TODO: 64bit pointers
						GenericResult = (~TargetExp.GenericResult) & $FFFFFFFF
					Case TYPE_LONG
						GenericResult = ~TargetExp.GenericResult
				End Select
			Case KEY_NOT
				If Datatype.IsInteger() Or Datatype.IsPointer() Then
					GenericResult = Not TargetExp.GenericResult
				ElseIf Datatype.TypeKind = TYPE_FLOAT Then
					SetFloatResult(Not TargetExp.GetFloatResult())
				ElseIf Datatype.TypeKind = TYPE_DOUBLE Then
					SetDoubleResult(Not TargetExp.GetDoubleResult())
				EndIf
		End Select
	End Method
	
	Method IsLValue:Int()
		Return False
	End Method
	
	Method IsWritable:Int()
		Return False
	End Method
	
	Method IsConstant:Int()
		Return TargetExp.IsConstant()
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		TargetExp.CheckCorrectness(Scope)
		
		Datatype = TargetExp.EvaluateDatatype()
		
		Select Operator
			Case TOK_MINUS
				If Not (Datatype.IsNumeric()  Or Datatype.IsPointer()) Then Abort("Negation is only applicable to numbers", Self)
			Case TOK_TILDE
				If Not (Datatype.IsInteger() Or Datatype.IsPointer()) Then Abort("Bitwise complement is only applicable to integers", Self)
		End Select
	End Method
	
	Method SetupChildren()
		AddChildNode(TargetExp)
	End Method
	
	Function Emit:Int(Operator:Int, TargetExp:Int)
		Local Expression:TUnaryOpExpression = CurrentTarget.CreateUnaryOpExpression()
		Expression.Operator = Operator
		Expression.TargetExp = TExpressionNode(RetrieveASTNode(TargetExp))
		
		Expression.SetupChildren()
		
		Return Expression.TableID
	End Function
End Type

Type TGroupingExpression Extends TExpressionNode
	Field TargetExp:TExpressionNode
	
	Field Datatype:TDatatype
	
	Method GetName:String()
		Return "TGroupingExpression"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
		TargetExp.EvaluateResult()
		
		GenericResult = TargetExp.GenericResult
	End Method
	
	Method IsLValue:Int()
		Return TargetExp.IsLValue()
	End Method
	
	Method IsWritable:Int()
		Return TargetExp.IsWritable()
	End Method
	
	Method IsConstant:Int()
		Return TargetExp.IsConstant()
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		TargetExp.CheckCorrectness(Scope)
		
		Datatype = TargetExp.EvaluateDatatype()
	End Method
	
	Method SetupChildren()
		AddChildNode(TargetExp)
	End Method
	
	Function Emit:Int(TargetExp:Int)
		Local Expression:TGroupingExpression = CurrentTarget.CreateGroupingExpression()
		Expression.TargetExp = TExpressionNode(RetrieveASTNode(TargetExp))
		
		Expression.SetupChildren()
		
		Return Expression.TableID
	End Function
End Type

Type TMemberAccessExpression Extends TExpressionNode
	Field TargetExp:TExpressionNode
	Field Identifier:String
	
	Field TargetClass:TClassNode
	Field Origin:TVariableDeclNode
	
	Field Datatype:TDatatype
	
	Field IsStaticAccess:Int
	Field Writable:Int
	
	Method GetName:String()
		Return "TMemberAccessExpression"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
		If Not Origin.Value Then Abort("Extern constants are not allowed in constant expressions", Self)
		
		Origin.Value.EvaluateResult()
		
		GenericResult = Origin.Value.GenericResult
	End Method
	
	Method IsLValue:Int()
		Return True
	End Method
	
	Method IsWritable:Int()
		Return Writable
	End Method
	
	Method IsConstant:Int()
		Return (Origin And Origin.DeclKind = KEY_CONST)
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		If TVariableNode(TargetExp) Then
			TargetClass = TClassNode(Scope.Find(TVariableNode(TargetExp).Identifier))
			
			If TargetClass Then IsStaticAccess = True
		EndIf
		
		If Not TargetClass Then
			TargetExp.CheckCorrectness(Scope)
			
			Local ExpType:TDatatype = TargetExp.EvaluateDatatype()
			
			If ExpType.TypeKind <> TYPE_COMPOUND And (Not ExpType.IsString()) And (Not ExpType.IsArray()) Then
				Abort("Member access is only valid for classes", Self)
			Else
				TargetClass = ExpType.Compound
			EndIf
		EndIf
			
		AddDependency(TargetClass, Self)
		
		Local Member:TASTNode = TargetClass.FindMember(Identifier, False)
		If Not Member Then Abort("'" + Identifier + "' is not a valid member of '" + TargetClass.Identifier + "'", Self)
		
		Local P:TASTNode = Parent, ParentClass:TClassNode
		While P And Not ParentClass
			ParentClass = TClassNode(P)
			P = P.Parent
		Wend
		
		Local VariableDecl:TVariableDeclNode = TVariableDeclNode(Member)
		
		Select VariableDecl.Visibility
			Case KEY_PUBLIC
				'Should always be accessible
			Case KEY_PROTECTED
				'TODO
			Case KEY_PRIVATE
				If (Not ParentClass) Or (Not VariableDecl.HasParentNode(ParentClass)) Then ..
					Abort("Member '" + Identifier + "' is private and not accessible from here", Self)
		End Select
		
		Identifier = VariableDecl.Identifier
		Datatype = VariableDecl.Datatype
		
		If VariableDecl.DeclKind = KEY_FIELD Or VariableDecl.DeclKind = KEY_GLOBAL Then Writable = True
		
		Origin = VariableDecl
	End Method
	
	Method SetupChildren()
		AddChildNode(TargetExp)
	End Method
	
	Method Create:TMemberAccessExpression(TargetExp:TExpressionNode, Identifier:String)
		Self.TargetExp  = TargetExp
		Self.Identifier = Identifier
		
		Return Self
	End Method
	
	Function Emit:Int(TargetExp:Int, Identifier:Int)
		Local Expression:TMemberAccessExpression = CurrentTarget.CreateMemberAccessExpression()
		Expression.TargetExp = TExpressionNode(RetrieveASTNode(TargetExp))
		Expression.Identifier = RetrieveString(Identifier)
		
		Expression.SetupChildren()
		
		Return Expression.TableID
	End Function
End Type

Type TLiteralNode Extends TExpressionNode
	Field LiteralKind:Int
	
	Field IntValue:Long
	Field FloatValue:Float
	Field DoubleValue:Double
	Field StringValue:String
	
	Field Datatype:TDatatype
	
	Method GetName:String()
		Return "TLiteralNode"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
		Select LiteralKind
			Case KEY_TRUE
				SetIntResult(1)
			Case KEY_FALSE
				SetIntResult(0)
			Case TOK_BYTE, TOK_SHORT, TOK_INT, TOK_LONG
				SetIntResult(IntValue)
			Case TOK_FLOAT
				SetFloatResult(FloatValue)
			Case TOK_DOUBLE
				SetDoubleResult(DoubleValue)
			Case KEY_NULL
				GenericResult = 0
		End Select
	End Method
	
	Method IsLValue:Int()
		Return False
	End Method
	
	Method IsWritable:Int()
		Return False
	End Method
	
	Method IsConstant:Int()
		Return LiteralKind <> TOK_STRING
	End Method
	
	Method IsSelf:Int()
		Return LiteralKind = KEY_SELF
	End Method
	
	Method IsSuper:Int()
		Return LiteralKind = KEY_SUPER
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		SetCommonRoot(Root)
		
		If LiteralKind = KEY_SELF Or LiteralKind = KEY_SUPER Then
			Local FunctionNode:TFunctionNode, ClassNode:TClassNode
			
			Local P:TASTNode = Parent
			While P
				If Not FunctionNode Then FunctionNode = TFunctionNode(P)
				If Not    ClassNode Then    ClassNode =    TClassNode(P) Else Exit
				
				P = P.Parent
			Wend
			
			Local Keyword:String
			If LiteralKind = KEY_SELF Then Keyword = "Self" Else Keyword = "Super"
			
			If Not ClassNode Then ..
				Abort("Use of " + Keyword + " is only allowed within classes", Self)
			
			If Not FunctionNode Then ..
				Abort("Use of " + Keyword + " is only allowed within functions", Self)
			
			If LiteralKind = KEY_SELF And (Not FunctionNode.IsMethod) Then ..
				Abort("Use of Self is only allowed within methods", Self)
			
			If LiteralKind = KEY_SELF Then
				Datatype = CurrentTarget.CreateDatatype().Create(TYPE_COMPOUND, ClassNode)
			Else
				Datatype = CurrentTarget.CreateDatatype().Create(TYPE_COMPOUND, TClassNode(ClassNode.ClassParents.First()))
			EndIf
		ElseIf Not Datatype Then
			Global TokenToType:Int[] = [TYPE_BYTE, TYPE_SHORT, TYPE_INT, TYPE_LONG, TYPE_FLOAT, TYPE_DOUBLE, TYPE_STRING, TYPE_UBYTE, TYPE_USHORT, TYPE_UINT, TYPE_ULONG]
			
			Select LiteralKind
				Case KEY_TRUE, KEY_FALSE
					Datatype = CurrentTarget.CreateDatatype().Create(TYPE_BOOL, Null)
				Case KEY_NULL
					Datatype = CurrentTarget.CreateDatatype().Create(TYPE_ANY, Null)
				Default
					Datatype = CurrentTarget.CreateDatatype().Create(TokenToType[LiteralKind - TOK_BYTE], Null)
			End Select
		EndIf
		
		Children.Clear()
		AddChildNode(Datatype)
		
		If Scope Then Datatype.CheckCorrectness(Scope)
	End Method
	
	Method Create:TLiteralNode(Datatype:TDatatype, Value:String)
		Self.Datatype = Datatype
		
		LiteralKind = Datatype.TypeKind
		
		Select LiteralKind
			Case TYPE_BYTE, TYPE_SHORT, TYPE_INT, TYPE_LONG, TOK_UBYTE, TOK_USHORT, TOK_UINT, TOK_ULONG
				IntValue = Value.ToLong()
			Case TYPE_FLOAT
				FloatValue = Value.ToFloat()
			Case TYPE_DOUBLE
				DoubleValue = Value.ToDouble()
			Case TYPE_COMPOUND
				If Datatype.IsString() Then StringValue = Value
		End Select
		
		Return Self
	End Method
	
	Function Emit:Int(LiteralKind:Int, Data:Long)
		Local Literal:TLiteralNode = CurrentTarget.CreateLiteralNode()
		Literal.LiteralKind = LiteralKind
		
		Select LiteralKind
			Case TOK_BYTE, TOK_SHORT, TOK_INT, TOK_LONG, TOK_UBYTE, TOK_USHORT, TOK_UINT, TOK_ULONG
				Literal.IntValue = RetrieveInt(Data)
			Case TOK_FLOAT
				Literal.FloatValue = RetrieveFloat(Data)
			Case TOK_DOUBLE
				Literal.DoubleValue = RetrieveDouble(Data)
			Case TOK_STRING
				Literal.StringValue = RetrieveString(Data)
		End Select
		
		If LiteralKind <> KEY_SELF And LiteralKind <> KEY_SUPER Then Literal.CheckCorrectness(Null) 'Get datatype early on
		
		Return Literal.TableID
	End Function
End Type

Type TVariableNode Extends TExpressionNode
	Field Identifier:String
	
	Field Datatype:TDatatype
	
	Field Origin:TVariableDeclNode
	
	Method GetName:String()
		Return "TVariableNode"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
		If Not Origin.Value Then Abort("Extern constants are not allowed in constant expressions", Self)
		
		Origin.Value.EvaluateResult()
		
		GenericResult = Origin.Value.GenericResult
	End Method
	
	Method IsLValue:Int()
		Return True
	End Method
	
	Method IsWritable:Int()
		Return Origin.DeclKind <> KEY_CONST
	End Method
	
	Method IsConstant:Int()
		Return Origin.DeclKind = KEY_CONST
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		If Datatype Then Datatype.CheckCorrectness(Scope)
		
		Origin = TVariableDeclNode(Scope.Find(Identifier))
		
		If Not Origin Then Abort("'" + Identifier + "' is undeclared", Self)
		If Datatype And Datatype.IsUnequal(Origin.Datatype) Then Abort("Variable datatype does not match declared datatype", Self)
		
		AddDependency(Origin, Self)
		
		Datatype = Origin.Datatype
	End Method
	
	Method SetupChildren()
		If Datatype Then AddChildNode(Datatype)
	End Method
	
	Method Create:TVariableNode(Identifier:String, Datatype:TDatatype)
		Self.Identifier = Identifier
		Self.Datatype   = Datatype
		
		SetupChildren()
		
		Return Self
	End Method
	
	Function Emit:Int(Identifier:Int, Datatype:Int)
		Local Variable:TVariableNode = CurrentTarget.CreateVariableNode()
		Variable.Identifier = RetrieveString(Identifier)
		Variable.Datatype   = TDatatype(RetrieveASTNode(Datatype))
		
		Variable.SetupChildren()
		
		Return Variable.TableID
	End Function
End Type

Type TNewExpression Extends TExpressionNode
	Field Identifier:String
	
	Field TargetClass:TClassNode
	Field Parameters :TList
	
	Field Datatype:TDatatype
	
	Method GetName:String()
		Return "TNewExpression"
	End Method
	
	Method EvaluateDatatype:TDatatype()
		Return Datatype
	End Method
	
	Method EvaluateResult()
	End Method
	
	Method IsLValue:Int()
		Return False
	End Method
	
	Method IsWritable:Int()
		Return False
	End Method
	
	Method IsConstant:Int()
		Return False
	End Method
	
	Method PerformASTTransformation:Int(Scope:TScope)
		If Parameters And TargetClass.Constructor Then
			SetCommonRoot(Root)
			
			Local MemberAccess:TMemberAccessExpression = CurrentTarget.CreateMemberAccessExpression().Create(Self, TargetClass.GenerateConstructorIdentifier())
			Local FunctionCall:TFunctionCall = CurrentTarget.CreateFunctionCall().Create(MemberAccess, Parameters, False)
			
			Parent.ReplaceChildNode(Self, FunctionCall)
			
			Parameters = Null 'Avoid performing AST transformations in an endless loop
			
			MemberAccess.CopyLocation(Self)
			FunctionCall.CopyLocation(Self)
			FunctionCall.CheckCorrectness(Scope)
			
			Return True
		EndIf
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		SetCommonRoot(Root)
		
		TargetClass = TClassNode(Scope.Find(Identifier))
		
		If Not TargetClass Then Abort("Class '" + Identifier + "' not found", Self)
		
		If TargetClass.IsAbstract() Then
			Local Msg:String = "Unable to instantiate abstract class '" + Identifier + "'"
			
			If Not (TargetClass.AbstractFlag Or TargetClass.AbstractFunctions.IsEmpty()) Then
				Msg :+ " due to abstract function"
				
				If TargetClass.AbstractFunctions.Count() > 1 Then Msg :+ "s:~n" Else Msg :+ " "
				
				For Local FunctionNode:TFunctionNode = EachIn TargetClass.AbstractFunctions
					If FunctionNode <> TargetClass.AbstractFunctions.First() Then Msg :+ "~n"
					
					Msg :+ "'" + FunctionNode.Identifier + "'"
				Next
			EndIf
			
			Abort(Msg, Self)
		EndIf
		
		AddDependency(TargetClass, Self)
		
		Datatype = CurrentTarget.CreateDatatype().Create(TYPE_COMPOUND, TargetClass)
		AddChildNode(Datatype)
		
		If Parameters And Not Parameters.IsEmpty() Then
			If Not TargetClass.Constructor Then ..
				Abort("No constructor available for class '" + TargetClass.Identifier + "'", Self)
		EndIf
		
		PerformASTTransformation(Scope)
	End Method
	
	Function Emit:Int(Identifier:Int, Parameters:Int)
		Local NewExpression:TNewExpression = CurrentTarget.CreateNewExpression()
		NewExpression.Identifier = RetrieveString(Identifier)
		NewExpression.Parameters = RetrieveList(Parameters)
		
		If Not NewExpression.Parameters Then NewExpression.Parameters = New TList
		
		Return NewExpression.TableID
	End Function
End Type

Type TVariableDeclNode Extends TStatementNode
	Field DeclKind:Int
	
	Field Identifier:String
	
	Field Datatype:TDatatype
	
	Field Value:TExpressionNode
	
	Field Visibility:Int
	
	Field CorrectnessChecked:Int
	
	Method New()
		Visibility = TClassNode.CurrentVisibility
	End Method
	
	Method GetName:String()
		Return "TVariableDeclNode"
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		Datatype.CheckCorrectness(Scope)
		
		If Value Then
			Value.CheckCorrectness(Scope)
			
			Local ValueType:TDatatype = Value.EvaluateDatatype()
			
			If Not ValueType.IsCastable(Datatype) Then ..
				Abort("Unable to convert from '" + ValueType.ToString() + "' to '" + Datatype.ToString() + "'", Self)
			
			If DeclKind = KEY_CONST And Not Value.IsConstant() Then Abort("Expression must be constant", Value)
			
			SetCommonRoot(Root)
			If ValueType.IsUnequal(Datatype) Then ..
				Value = CurrentTarget.CreateTypeCastExpression().Create(Value, Datatype.Copy(), Scope)
		Else
			If DeclKind = KEY_CONST And Not (IsExtern Or IsCompiled) Then Abort("Constants must be intialized", Self)
		EndIf
		
		If Not (Scope.Insert(Identifier, Self) Or DeclKind = KEY_GLOBAL) Then Abort("Redefinition of '" + Identifier + "'", Self)
		
		CorrectnessChecked = True
	End Method
	
	Method CheckSignatureCorrectness(Scope:TScope)
		Datatype.CheckCorrectness(Scope)
	End Method
	
	Method ToString:String()
		Local Out:String = Super.ToString() + " "
		
		Select DeclKind
			Case KEY_CONST
				Out :+ "Const"
			Case KEY_LOCAL
				Out :+ "Local"
			Case KEY_FIELD
				Out :+ "Field"
			Case KEY_GLOBAL
				Out :+ "Global"
			Default
				Abort("Invalid variable scope", Self)
		End Select
		
		Out :+ " " + Identifier + ":" + Datatype.ToString()
		
		If Value And DeclKind = KEY_CONST And CorrectnessChecked Then
			Value.EvaluateResult()
			
			Out :+ "="
			
			Select Datatype.TypeKind
				Case TYPE_BYTE, TYPE_SHORT, TYPE_INT, TYPE_LONG
					Out :+ String.FromLong(Value.GetCastedIntResult())
				Case TYPE_FLOAT, TYPE_DOUBLE
					Out :+ String.FromDouble(Value.GetCastedDoubleResult())
				Case TYPE_STRING
					Out :+ Value.GetCastedStringResult()
			End Select
		EndIf
		
		Return Out
	End Method
	
	Method FromString:TVariableDeclNode(Source:String)
		IsCompiled = True
		
		Local Space:Int = Source.Find(" ")
		Super.LocationFromString(Source[ .. Space])
		
		Source = Source[Space + 1 ..]
		
		If Source.StartsWith("Const") Then
			DeclKind = KEY_CONST
		ElseIf Source.StartsWith("Global") Then
			DeclKind = KEY_GLOBAL
		ElseIf Source.StartsWith("Field") Then
			DeclKind = KEY_FIELD
		EndIf
		
		Local Stop:Int = Source.Find("=")
		If Stop < 0 Then Stop = Source.Length
		
		Identifier = Source[Source.Find(" ") + 1 .. Source.Find(":")]
		
		Datatype = CurrentTarget.CreateDatatype().FromString(Source[Source.Find(":") + 1 .. Stop])
		
		If Source.Find("=") Then Value = CurrentTarget.CreateLiteralNode().Create(Datatype, Source[Stop + 1 ..])
		
		SetupChildren()
		
		Return Self
	End Method
	
	Method SetupChildren()
		If Value Then AddChildNode(Value)
		AddChildNode(Datatype)
	End Method
	
	Method Create:TVariableDeclNode(DeclKind:Int, Identifier:String, Datatype:TDatatype, Value:TExpressionNode)
		Self.DeclKind = DeclKind
		Self.Identifier = Identifier
		Self.Datatype = Datatype
		Self.Value = Value
		
		SetupChildren()
		
		Return Self
	End Method
	
	Function Emit:Int(VariableIndex:Int, Value:Int)
		Local VariableDecl:TVariableDeclNode = CurrentTarget.CreateVariableDeclNode()
		
		Local Variable:TVariableNode = TVariableNode(RetrieveASTNode(VariableIndex))
		VariableDecl.Identifier = Variable.Identifier
		VariableDecl.Datatype   = Variable.Datatype
		VariableDecl.Value      = TExpressionNode(RetrieveASTNode(Value))
		
		VariableDecl.SetupChildren()
		
		Return VariableDecl.TableID
	End Function
	
	Function ScopeVarDecl(DeclKind:Int, DeclList:Int)
		For Local VariableDecl:TVariableDeclNode = EachIn RetrieveList(DeclList)
			VariableDecl.DeclKind = DeclKind
		Next
	End Function
End Type

Type TDatatype Extends TASTNode
	Field TypeKind:Int
	
	Field IsUnsigned:Int
	
	Field Identifier:String
	
	Field Compound:TClassNode
	Field FunctionNode:TFunctionNode
	
	Field ReturnType:TDatatype
	Field ParameterCount:Int
	Field Identifiers:String[]
	Field Parameters:TDatatype[]
	Field ParamValues:TExpressionNode[]
	Field MandatoryParamCount:Int
	
	Field NestedType:TDatatype
	
	Field Dimension:Int
	
	Method GetName:String()
		Return "TDatatype"
	End Method
	
	Method GenerateAutoValue(Writer:TCodeWriter) Abstract
	
	Method Copy:TDatatype()
		SetCommonRoot(Root)
		
		If TypeKind = TYPE_FUNCTION Then Abort("Not yet implemented", Self)
		
		Local Data:TASTNode
		If TypeKind = TYPE_COMPOUND Then
			Data = Compound
		ElseIf TypeKind = TYPE_POINTER Then
			Data = NestedType
		ElseIf TypeKind = TYPE_ARRAY Then
			Data = NestedType
		EndIf
		
		Local KindAdd:Int
		If IsUnsigned Then KindAdd = TYPE_UBYTE - TYPE_BYTE
		
		Return CurrentTarget.CreateDatatype().Create(TypeKind + KindAdd, Data, Dimension)
	End Method
	
	Method IsUnequal:Int(Other:TDatatype)
		If TypeKind = TYPE_ANY Or Other.TypeKind = TYPE_ANY Then Return False
		
		If TypeKind   <> Other.TypeKind   Then Return True
		If IsUnsigned <> Other.IsUnsigned Then Return True
		
		If TypeKind = TYPE_COMPOUND And Other.IsCompound() And Identifier.ToLower() <> Other.Identifier.ToLower() Then Return True
		If TypeKind = TYPE_POINTER Or TypeKind = TYPE_ARRAY Then
			If Other.TypeKind <> TypeKind Then Return True
			
			If Other.Dimension <> Dimension Then Return True
			
			Return NestedType.IsUnequal(Other.NestedType)
		EndIf
		
		If TypeKind = TYPE_FUNCTION Then
			If (FunctionNode And Other.FunctionNode) And (FunctionNode.IsMethod <> Other.FunctionNode.IsMethod) Then Return True
			If ParameterCount <> Other.ParameterCount Then Return True
			If ReturnType.IsUnequal(Other.ReturnType) Then Return True
			
			For Local I:Int = 0 Until ParameterCount
				If Parameters[I].IsUnequal(Other.Parameters[I]) Then Return True
			Next
		EndIf
		
		Return False
	End Method
	
	Method IsAny:Int()
		Return TypeKind = TYPE_ANY
	End Method
	
	Method IsVoid:Int()
		Return TypeKind = TYPE_VOID
	End Method
	
	Method IsPointer:Int()
		Return TypeKind = TYPE_POINTER
	End Method
	
	Method IsVoidPointer:Int()
		Return IsPointer() And NestedType.IsVoid()
	End Method
	
	Method IsNumeric:Int()
		Return IsInteger() Or IsReal()
	End Method
	
	Method IsInteger:Int()
		Select TypeKind
			Case TYPE_BOOL, TYPE_BYTE, TYPE_SHORT, TYPE_INT, TYPE_LONG
				Return True
			Default
				Return False
		End Select
	End Method
	
	Method IsReal:Int()
		If TypeKind = TYPE_FLOAT Or TypeKind = TYPE_DOUBLE Then Return True
		
		Return False
	End Method
	
	Method IsPrimitive:Int()
		Return Not (IsCompound() Or IsArray())
	End Method
	
	Method IsCompound:Int()
		Return TypeKind = TYPE_COMPOUND Or TypeKind = TYPE_STRING Or TypeKind = TYPE_ARRAY
	End Method
	
	Method IsString:Int()
		Return TypeKind = TYPE_STRING
	End Method
	
	Method IsFunction:Int()
		Return TypeKind = TYPE_FUNCTION
	End Method
	
	Method IsArray:Int()
		Return TypeKind = TYPE_ARRAY
	End Method
	
	Method IsBoolean:Int()
		Return TypeKind = TYPE_BOOL
	End Method
	
	Method IsCastable:Int(Other:TDatatype)
		If TypeKind = TYPE_ANY                                     Then Return True
		If Not IsUnequal(Other)                                    Then Return True
		If IsNumeric() And (Other.IsNumeric() Or Other.IsString()) Then Return True
		If IsPointer() And Other.IsVoidPointer()                   Then Return True
		If Other.IsBoolean()                                       Then Return True
		
		If IsString() And Other.IsPointer() And ..
		   Other.NestedType.TypeKind = TYPE_BYTE Then Return True
		
		If IsFunction() And Other.IsFunction() And ReturnType.IsCompound() And ..
		   Other.ReturnType.IsCompound() And ReturnType.IsCastable(Other.ReturnType) Then ..
			Return True
		
		If (IsCompound() And Other.IsCompound()) And Other.IsUnequal(Self) Then ..
			Return Compound.HasParent(Other.Compound)
		
		If (IsArray() Or IsString()) And Other.IsCompound() And Other.Identifier = "Object" Then Return True
		
		Return False
	End Method
	
	Method IsExplicitCastable:Int(Other:TDatatype)
		If IsCastable(Other) Then Return True
		
		If IsString() And Other.IsNumeric() Then Return True
		If IsPointer() And Other.IsInteger() Then Return True
		If Other.IsPointer() And IsInteger() Then Return True
		If IsPointer() And Other.IsPointer() Then Return True
		If (IsCompound() Or IsString()) And (Other.IsCompound() Or Other.IsString() Or Other.IsArray()) Then Return True
		
		Return False
	End Method
	
	Method Create:TDatatype(TypeKind:Int, Data:TASTNode, Dimension:Int = 0)
		Self.TypeKind = TypeKind
		Self.Dimension = Dimension
		
		If TypeKind = TYPE_STRING Then
			Identifier = "String"
		ElseIf TypeKind = TYPE_COMPOUND Then
			Compound = TClassNode(Data)
			
			Identifier = Compound.Identifier
			
			If Identifier = "String" Then
				Self.TypeKind = TYPE_STRING
				Self.Compound = Null
			EndIf
		ElseIf TypeKind = TYPE_FUNCTION Then
			FunctionNode = TFunctionNode(Data)
			
			ReturnType = FunctionNode.ReturnDatatype
			ParameterCount = FunctionNode.ParameterCount

			Parameters  = New TDatatype[ParameterCount]
			Identifiers = New String[ParameterCount]
			
			For Local I:Int = 0 Until ParameterCount
				Parameters [I] = FunctionNode.GetParameter(I).Datatype
				Identifiers[I] = FunctionNode.GetParameter(I).Identifier
			Next
		ElseIf TypeKind = TYPE_POINTER Then
			NestedType = TDatatype(Data)
		ElseIf TypeKind = TYPE_ARRAY Then
			Identifier = "Array"
			
			NestedType = TDatatype(Data)
		ElseIf TypeKind >= TYPE_UBYTE And TypeKind <= TYPE_ULONG
			Self.TypeKind :- TYPE_UBYTE - TYPE_BYTE
			IsUnsigned = True
		EndIf
		
		SetupChildren()
		
		Return Self
	End Method
	
	Method CheckCorrectness(Scope:TScope)
		If IsCompound() Then
			Local Node:TASTNode = Scope.Find(Identifier)
			Compound = TClassNode(Node)
			
			If Compound Then
				AddDependency(Compound, Self)
			Else
				If Not IgnoreSemanticErrors Then
					If Node Then
						Abort("'" + Identifier + "' is not a class", Self)
					Else
						Abort("Class '" + Identifier + "' not found", Self)
					EndIf
				EndIf
			EndIf
			
			If TypeKind = TYPE_ARRAY Then NestedType.CheckCorrectness(Scope)
		ElseIf TypeKind = TYPE_FUNCTION Then
			If Not ReturnType.IsVoid() Then ReturnType.CheckCorrectness(Scope)
			
			For Local I:TDatatype = EachIn Parameters
				I.CheckCorrectness(Scope)
			Next
		ElseIf TypeKind = TYPE_POINTER Then
			If Not NestedType.IsVoid() Then NestedType.CheckCorrectness(Scope)
		ElseIf TypeKind = TYPE_VOID Then
			Abort("Void is only allowed in conjunction with pointers or functions", Self)
		EndIf
	End Method
	
	Method ToString:String()
		Local Out:String
		
		If IsUnsigned Then Out = "U"
		
		Select TypeKind
			Case TYPE_BOOL
				Out :+ "Bool"
			Case TYPE_BYTE
				Out :+ "Byte"
			Case TYPE_SHORT
				Out :+ "Short"
			Case TYPE_INT
				Out :+ "Int"
			Case TYPE_LONG
				Out :+ "Long"
			Case TYPE_FLOAT
				Out :+ "Float"
			Case TYPE_LONG
				Out :+ "Long"
			Case TYPE_DOUBLE
				Out :+ "Double"
			Case TYPE_STRING
				Out :+ "String"
			Case TYPE_COMPOUND
				Out :+ Identifier
			Case TYPE_FUNCTION
				Out :+ ReturnType.ToString() + "("
				
				For Local I:Int = 0 Until ParameterCount
					Out :+ Identifiers[I] + ":" + Parameters[I].ToString()
					
					If I < ParameterCount - 1 Then Out :+ ", "
				Next
				
				Out :+ ")"
			Case TYPE_POINTER
				Out :+ NestedType.ToString() + " Ptr"
			Case TYPE_ARRAY
				Out :+ NestedType.ToString() + "[" + ",,,,,,,,,,"[.. Dimension - 1] + "]"
			Case TYPE_VOID
				Out :+ "Void"
			Case TYPE_ANY
				Out :+ "Null"
			Default
				Abort("Invalid Datatype", Self)
		End Select
		
		Return Out
	End Method
	
	Method SetupChildren()
		If NestedType Then AddChildNode(NestedType)
		If ReturnType Then AddChildNode(ReturnType)
		If ParameterCount Then AddChildNodeArray(Parameters)
	End Method
	
	Method FromString:TDatatype(Source:String)
		IsCompiled = True
		
		Select Source
			Case "", "Void"
				TypeKind = TYPE_VOID
			Case "Bool"
				TypeKind = TYPE_BOOL
			Case "Byte"
				TypeKind = TYPE_BYTE
			Case "Short"
				TypeKind = TYPE_SHORT
			Case "Int"
				TypeKind = TYPE_INT
			Case "Long"
				TypeKind = TYPE_LONG
			Case "UByte"
				TypeKind = TYPE_BYTE
				IsUnsigned = True
			Case "UShort"
				TypeKind = TYPE_SHORT
				IsUnsigned = True
			Case "UInt"
				TypeKind = TYPE_INT
				IsUnsigned = True
			Case "ULong"
				TypeKind = TYPE_LONG
				IsUnsigned = True
			Case "Float"
				TypeKind = TYPE_FLOAT
			Case "Double"
				TypeKind = TYPE_DOUBLE
			Case "String"
				TypeKind = TYPE_STRING
				Identifier = "String"
			Default
				If Source.EndsWith(" Ptr") Then
					TypeKind = TYPE_POINTER
					
					NestedType = CurrentTarget.CreateDatatype().FromString(Source[.. Source.Length - 4])
				ElseIf Source.EndsWith("]") Then
					TypeKind = TYPE_ARRAY
					Dimension = Source.FindLast("]") - Source.FindLast("[")
					
					NestedType = CurrentTarget.CreateDatatype().FromString(Source[.. Source.FindLast("[")])
					Identifier = "Array"
				ElseIf Source.EndsWith(")") Then
					TypeKind = TYPE_FUNCTION
					
					Local Start:Int = FindAtLevel(Source, "(")
					ReturnType = CurrentTarget.CreateDatatype().FromString(Source[.. Start])
					Source = Source[Start + 1 .. Source.Length - 1]
					
					ParameterCount = 0
					
					While Source
						Local Offset:Int = FindAtLevel(Source, ",")
						If Offset < 0 Then Offset = Source.Length
						
						ParameterCount :+ 1
						Local ColonOffset:Int = Source.Find(":")
						Identifiers :+ [Source[.. ColonOffset]]
						Parameters  :+ [CurrentTarget.CreateDatatype().FromString(Source[ColonOffset + 1 .. Offset])]
						
						Source = Source[Offset + 2 ..]
					Wend
				Else
					TypeKind = TYPE_COMPOUND
					Identifier = Source
				EndIf
		End Select
		
		SetupChildren()
		
		Return Self
	End Method
	
	Function Promote:TDatatype(A:TDatatype, B:TDatatype)
		If A.TypeKind = TYPE_ANY Then Return B.Copy()
		If B.TypeKind = TYPE_ANY Then Return A.Copy()
		
		If A.IsCompound() Or B.IsCompound() Then Return Null
		
		For Local I:Int = EachIn [TYPE_STRING, TYPE_POINTER, TYPE_DOUBLE, TYPE_FLOAT, TYPE_LONG]
			If A.TypeKind = I Then Return A.Copy()
			If B.TypeKind = I Then Return B.Copy()
		Next
		
		Return CurrentTarget.CreateDatatype().Create(TYPE_INT, Null)
	End Function
	
	Function Emit:Int(TypeKind:Int, Data:Int, Dimension:Int)
		Local Datatype:TDatatype = CurrentTarget.CreateDatatype()
		Datatype.TypeKind = TypeKind
		Datatype.Dimension = Dimension
		
		If TypeKind = TYPE_COMPOUND Then
			Datatype.Identifier = RetrieveString(Data)
			
			If Datatype.Identifier.ToLower() = "string" Then Datatype.TypeKind = TYPE_STRING
		ElseIf TypeKind = TYPE_POINTER Or TypeKind = TYPE_ARRAY Then
			Datatype.NestedType = TDatatype(RetrieveASTNode(Data))
			
			If TypeKind = TYPE_ARRAY Then Datatype.Identifier = "Array"
		ElseIf TypeKind = TYPE_FUNCTION Then
			Datatype.ReturnType = TDatatype(RetrieveASTNode(Data))
			
			Local ParamList:TList = RetrieveList(Dimension)
			
			Datatype.ParameterCount = ParamList.Count()
			Datatype.Parameters  = New TDatatype      [Datatype.ParameterCount]
			Datatype.Identifiers = New String         [Datatype.ParameterCount]
			Datatype.ParamValues = New TExpressionNode[Datatype.ParameterCount]
			
			Local I:Int = 0
			For Local Param:TVariableDeclNode = EachIn ParamList
				Datatype.Parameters [I] = Param.Datatype
				Datatype.Identifiers[I] = Param.Identifier
				Datatype.ParamValues[I] = Param.Value
				Datatype.MandatoryParamCount :+ (Param.Value <> Null)
				
				I :+ 1
			Next
		ElseIf TypeKind >= TYPE_UBYTE And TypeKind <= TYPE_ULONG
			Datatype.TypeKind :- TYPE_UBYTE - TYPE_BYTE
			Datatype.IsUnsigned = True
		EndIf
		
		Datatype.SetupChildren()
		
		Return Datatype.TableID 
	End Function
End Type