%{
	#define YYDEBUG 1
	int (*yylex)(void);
	void (*SetRuleLocation)(void*);
	void yyerror (char const *);

	typedef struct {
		int From;
		int To;
		int Step;
		int SkipLast;
	} ForToExp;

	int (*EmitProgram)();
	int (*EmitProgramFileImport)(int, int);
	int (*EmitProgramModuleImport)(int, int);
	int (*EmitProgramStatement)(int, int);
	int (*EmitProgramFunction)(int, int);
	int (*EmitProgramClass)(int, int);
	int (*EmitProgramGlobal)(int, int);
	int (*EmitProgramExtern)(int, int);
	int (*EmitFunction)(int, int, int, int);
	int (*EmitEmptyList)();
	int (*EmitSingleList)(int);
	int (*EmitClass)(int, int, int, int);
	int (*EmitReturn)(int);
	int (*EmitExit)();
	int (*EmitContinue)();
	int (*EmitDelete)(int);
	int (*EmitIf)(int, int, int, int);
	int (*EmitElseIf)(int, int);
	int (*EmitElse)(int);
	int (*EmitWhile)(int, int);
	int (*EmitRepeat)(int, int);
	int (*EmitSelect)(int, int, int, int);
	int (*EmitCase)(int, int);
	int (*EmitDefault)(int);
	int (*EmitForWhile)(int, int, int);
	int (*EmitFor)(int, int, int, int, int, int, int);
	int (*EmitAssignment)(int);
	int (*EmitAddress)(int);
	int (*EmitAutoArray)(int);
	int (*EmitNewArray)(int, int);
	int (*EmitIndex)(int, int);
	int (*EmitFunctionCall)(int, int, int);
	int (*EmitBinaryOp)(int, int, int);
	int (*EmitUnaryOp)(int, int);
	int (*EmitGrouping)(int);
	int (*EmitMemberAccess)(int, int);
	int (*EmitNew)(int, int);
	int (*EmitVariableDecl)(int, int);
	int (*EmitVariable)(int, int);
	int (*EmitDatatype)(int, int, int);
	int (*EmitLiteral)(int, long long int);
	int (*EmitTypeCast)(int, int);
	void (*ScopeVarDecl)(int, int);
	void (*AppendToList)(int, int);
	void (*AppendStringToList)(int, int);
	void (*PrependToList)(int, int);
	int (*AppendIdentifier)(int, int);
	int (*AppendConstChar)(int, const char*);
	int (*SetVisibility)(int);

	# define YYLLOC_DEFAULT(Current, Rhs, N) \
		do {\
			if (YYID (N)) { \
				(Current).first_line   = YYRHSLOC (Rhs, 1).first_line; \
				(Current).first_column = YYRHSLOC (Rhs, 1).first_column; \
				(Current).last_line    = YYRHSLOC (Rhs, N).last_line; \
				(Current).last_column  = YYRHSLOC (Rhs, N).last_column; \
			} else { \
				(Current).first_line   = (Current).last_line   = YYRHSLOC (Rhs, 0).last_line; \
				(Current).first_column = (Current).last_column = YYRHSLOC (Rhs, 0).last_column; \
			} \
			SetRuleLocation((void*)&(Current));\
		} while (YYID (0))
%}

%union {
	int keyword;
	int token;
	int id;
	long long int literal;
	ForToExp ForExp;
}

%token <keyword> KEY_ABSTRACT    259
%token <keyword> KEY_AND         260
%token <keyword> KEY_BYTE        263
%token <keyword> KEY_CASE        264
%token <keyword> KEY_CONST       267
%token <keyword> KEY_CONTINUE    268
%token <keyword> KEY_DEFAULT     269
%token <keyword> KEY_DELETE      271
%token <keyword> KEY_DO          356
%token <keyword> KEY_DOUBLE      272
%token <keyword> KEY_ELSE        274
%token <keyword> KEY_ELSEIF      275
%token <keyword> KEY_ENDEXTERN   277
%token <keyword> KEY_ENDFUNCTION 278
%token <keyword> KEY_ENDIF       279
%token <keyword> KEY_ENDMETHOD   280
%token <keyword> KEY_ENDSELECT   282
%token <keyword> KEY_ENDCLASS    284
%token <keyword> KEY_ENDWHILE    285
%token <keyword> KEY_EXIT        286
%token <keyword> KEY_EXTENDS     287
%token <keyword> KEY_EXTERN      288
%token <keyword> KEY_FALSE       289
%token <keyword> KEY_FIELD       290
%token <keyword> KEY_FINAL       291
%token <keyword> KEY_FLOAT       292
%token <keyword> KEY_FOR         293
%token <keyword> KEY_FOREVER     294
%token <keyword> KEY_FUNCTION    296
%token <keyword> KEY_GLOBAL      297
%token <keyword> KEY_IF          299
%token <keyword> KEY_IMPORT      300
%token <keyword> KEY_INT         305
%token <keyword> KEY_LOCAL       307
%token <keyword> KEY_LONG        308
%token <keyword> KEY_METHOD      310
%token <keyword> KEY_MOD         312
%token <keyword> KEY_NEW         315
%token <keyword> KEY_NEXT        316
%token <keyword> KEY_NOT         317
%token <keyword> KEY_NULL        318
%token <keyword> KEY_OR          320
%token <keyword> KEY_PRIVATE     322
%token <keyword> KEY_PTR         323
%token <keyword> KEY_PUBLIC      324
%token <keyword> KEY_REPEAT      328
%token <keyword> KEY_RETURN      330
%token <keyword> KEY_SAR         331
%token <keyword> KEY_SELECT      332
%token <keyword> KEY_SELF        333
%token <keyword> KEY_SHL         335
%token <keyword> KEY_SHORT       336
%token <keyword> KEY_SHR         337
%token <keyword> KEY_STEP        339
%token <keyword> KEY_SUPER       342
%token <keyword> KEY_THEN        344
%token <keyword> KEY_TO          346
%token <keyword> KEY_TRUE        347
%token <keyword> KEY_CLASS       349
%token <keyword> KEY_UNTIL       350
%token <keyword> KEY_VARPTR      352
%token <keyword> KEY_WEND        353
%token <keyword> KEY_WHILE       354
%token <keyword> KEY_BOOL        355
%token <keyword> KEY_UBYTE       357
%token <keyword> KEY_USHORT      358
%token <keyword> KEY_UINT        359
%token <keyword> KEY_ULONG       360
%token <keyword> KEY_VOID        361
%token <keyword> KEY_PROTECTED   362

%token <token> TOK_EOL 511
%token <token> TOK_EOF 0

%token <literal> TOK_IDENTIFIER 512
%token <literal> TOK_BYTE       513
%token <literal> TOK_SHORT      514
%token <literal> TOK_INT        515
%token <literal> TOK_LONG       516
%token <literal> TOK_FLOAT      517
%token <literal> TOK_DOUBLE     518
%token <literal> TOK_STRING     519
%token <literal> TOK_UBYTE      520
%token <literal> TOK_USHORT     521
%token <literal> TOK_UINT       522
%token <literal> TOK_ULONG      523

%token <token> TOK_SHARP    428 <token> TOK_COMMA    418 <token> TOK_EXCLAMATION 422 <token> TOK_COLON     417
%token <token> TOK_PERIOD   419 <token> TOK_PERCENT  424 <token> TOK_AT          423 <token> TOK_SEMICOLON 420
%token <token> TOK_LPARAN   406 <token> TOK_RPARAN   407 <token> TOK_LBRACKET    426 <token> TOK_RBRACKET  427
%token <token> TOK_GE       411 <token> TOK_LE       412 <token> TOK_G           408 <token> TOK_L         409
%token <token> TOK_NE       413 <token> TOK_E        410 <token> TOK_PLUS        401 <token> TOK_MINUS     402
%token <token> TOK_BIT_XOR  414 <token> TOK_BIT_OR   415 <token> TOK_BIT_AND     416 <token> TOK_DIVIDE    404
%token <token> TOK_ASTERISK 403 <token> TOK_POWER    405

%token <token> TYPE_BYTE     1025
%token <token> TYPE_SHORT    1026
%token <token> TYPE_INT      1027
%token <token> TYPE_LONG     1028
%token <token> TYPE_FLOAT    1029
%token <token> TYPE_DOUBLE   1030
%token <token> TYPE_STRING   1031
%token <token> TYPE_UBYTE    1032
%token <token> TYPE_USHORT   1033
%token <token> TYPE_UINT     1034
%token <token> TYPE_ULONG    1035
%token <token> TYPE_COMPOUND 1036
%token <token> TYPE_FUNCTION 1037
%token <token> TYPE_BOOL     1038
%token <token> TYPE_VOID     1039
%token <token> TYPE_POINTER  1040
%token <token> TYPE_ARRAY    1041

%token <token> FLAG_ABSTRACT 0x1
%token <token> FLAG_FINAL    0x2

%type <id> Program
%type <id> ExternBlock
%type <literal> ModuleImport
%type <literal> NonWildcardModuleImport
%type <id> StatementList
%type <id> Statement
%type <id> SimpleStatement
%type <id> StatementLine
%type <id> ControlStatement
%type <id> Equality
%type <id> FunctionCall
%type <id> ExpressionList
%type <id> NonEmptyExpressionList
%type <id> Expression
%type <id> LogicalExp
%type <id> BooleanExp
%type <id> Modular
%type <id> Sum
%type <id> BinaryExp
%type <id> Product
%type <id> Unary
%type <id> Exponential
%type <id> Factor
%type <id> NonVariableFactor
%type <id> Literal
%type <id> OneLineIf
%type <id> OneLineElseIfList
%type <id> OneLineElse
%type <id> IfStatement
%type <id> ElseIfList
%type <id> OptionalElse
%type <id> WhileStatement
%type <id> RepeatStatement
%type <id> SelectStatement
%type <id> DefaultStatement
%type <id> CaseStatementList
%type <id> ForStatement
%type <ForExp> ForToExpression
%type <id> OptionalStep
%type <id> ClassDeclaration
%type <id> DeclarationList
%type <id> MethodDeclaration
%type <id> FunctionDeclaration
%type <id> ExternClass
%type <id> ExternDeclarationList
%type <id> ExternMethod
%type <id> ExternFunction
%type <id> GlobalDeclaration
%type <id> LocalDeclaration
%type <id> ConstDeclaration
%type <id> FieldDeclaration
%type <id> VarDeclList
%type <id> NonEmptyVarDeclList
%type <id> OptionalAssignment
%type <id> VariableDecl
%type <id> Variable
%type <id> TypeSpecifier
%type <id> Datatype
%type <literal> AmbiguousBracket
%type <literal> AmbiguousParan
%type <id> CommaList
%type <id> LineDelimiter
%type <id> ExtendList
%type <id> NonEmptyExtendList
%type <id> InheritanceFlags
%type <id> OptionalFinal
%type <id> ForWhileStatement
%type <id> OptionalLineDelimiter
%type <id> FunctionDatatype
%type <id> DatatypeOrIdentifier
%type <id> FunctionDecl

%%
	Program:
		/* empty */                                     {$$ = EmitProgram()}
		| Program KEY_IMPORT TOK_STRING LineDelimiter   {EmitProgramFileImport($1, $3)}
		| Program ModuleImport          LineDelimiter   {EmitProgramModuleImport($1, $2)}
		| Program Statement             LineDelimiter   {EmitProgramStatement($1, $2)}
		| Program OneLineIf             TOK_EOL         {EmitProgramStatement($1, $2)}
		| Program FunctionDeclaration   LineDelimiter   {EmitProgramFunction($1, $2)}
		| Program GlobalDeclaration     LineDelimiter   {EmitProgramGlobal($1, $2)}
		| Program ClassDeclaration      LineDelimiter   {EmitProgramClass($1, $2)}
		| Program KEY_EXTERN LineDelimiter ExternBlock KEY_ENDEXTERN LineDelimiter {EmitProgramExtern($1, $4)}
	;

	ExternBlock:
		  /* empty */                                 {$$ = EmitEmptyList()}
		| ExternBlock GlobalDeclaration LineDelimiter {AppendToList($1, $2)}
		| ExternBlock ConstDeclaration  LineDelimiter {AppendToList($1, $2)}
		| ExternBlock ExternClass       LineDelimiter {AppendToList($1, $2)}
		| ExternBlock ExternFunction    LineDelimiter {AppendToList($1, $2)}
	;

	ModuleImport:
		  NonWildcardModuleImport
		| NonWildcardModuleImport '.' '*' {$$ = AppendConstChar($1, ".*")}
	;

	NonWildcardModuleImport:
		  KEY_IMPORT TOK_IDENTIFIER                  {$$ = $2}
		| NonWildcardModuleImport '.' TOK_IDENTIFIER {$$ = AppendIdentifier(AppendConstChar($1, "."), $3)}
	;

	StatementList:
		  /* empty */                           {$$ = EmitEmptyList()}
		| StatementList Statement LineDelimiter {AppendToList($1, $2)}
		| StatementList OneLineIf TOK_EOL       {AppendToList($1, $2)}
	;

	Statement:
		  ConstDeclaration
		| ControlStatement
		| SimpleStatement
	;

	SimpleStatement:
		  LocalDeclaration
		| Modular '=' Expression {$$ = EmitAssignment(EmitBinaryOp('=', $1, $3))}
		| FunctionCall           {$$ = EmitFunctionCall($1, -1, -1)}
		| KEY_RETURN             {$$ = EmitReturn(-1)}
		| KEY_RETURN Expression  {$$ = EmitReturn($2)}
		| KEY_DELETE Expression  {$$ = EmitDelete($2)}
        | KEY_EXIT               {$$ = EmitExit()}
        | KEY_CONTINUE           {$$ = EmitContinue()}
	;

	StatementLine:
		  SimpleStatement                   {$$ = EmitSingleList($1)}
		| StatementLine ';' SimpleStatement {AppendToList($1, $3)}
	;

	ControlStatement:
		  IfStatement
		| SelectStatement
		| WhileStatement
		| RepeatStatement
		| ForWhileStatement
		| ForStatement
	;

	Equality:
		Modular '=' Modular {$$ = EmitBinaryOp('=', $1, $3)}
	;

	FunctionCall:
		  NonVariableFactor '('  ExpressionList ')' {$$ = EmitFunctionCall($1, $3, 0)}
        | AmbiguousParan ExpressionList ')' {$$ = EmitFunctionCall(EmitVariable($1, -1), $2, 0)}
	;

	ExpressionList:
		  /* empty */            {$$ = EmitEmptyList()}
		| NonEmptyExpressionList
	;

	NonEmptyExpressionList:
		  Expression                            {$$ = EmitSingleList($1)}
		| NonEmptyExpressionList ',' Expression {AppendToList($1, $3)}

	Expression:
		LogicalExp
	;

	LogicalExp:
		  BooleanExp
		| LogicalExp KEY_AND BooleanExp {$$ = EmitBinaryOp(KEY_AND, $1, $3)}
		| LogicalExp KEY_OR  BooleanExp {$$ = EmitBinaryOp(KEY_OR , $1, $3)}
		|            KEY_NOT BooleanExp {$$ = EmitUnaryOp(KEY_NOT, $2)}
	;

	BooleanExp:
		  Modular
		| BooleanExp TOK_LE Modular {$$ = EmitBinaryOp(TOK_LE, $1, $3)}
		| BooleanExp TOK_GE Modular {$$ = EmitBinaryOp(TOK_GE, $1, $3)}
		| BooleanExp TOK_NE Modular {$$ = EmitBinaryOp(TOK_NE, $1, $3)}
		| BooleanExp '>'  Modular {$$ = EmitBinaryOp(   '>', $1, $3)}
		| BooleanExp '<'  Modular {$$ = EmitBinaryOp(   '<', $1, $3)}
		| Equality
	;

	Modular:
		  Sum
		| Modular KEY_MOD Sum  {$$ = EmitBinaryOp(KEY_MOD, $1, $3)}
	;

	Sum:
		  BinaryExp
		| Sum '+' BinaryExp {$$ = EmitBinaryOp('+', $1, $3)}
		| Sum '-' BinaryExp {$$ = EmitBinaryOp('-', $1, $3)}
	;

	BinaryExp:
		  Product
		| BinaryExp '|' Product   {$$ = EmitBinaryOp('|', $1, $3)}
		| BinaryExp '&' Product   {$$ = EmitBinaryOp('&', $1, $3)}
		| BinaryExp '~' Product   {$$ = EmitBinaryOp('~', $1, $3)}
		| BinaryExp KEY_SHL Product {$$ = EmitBinaryOp(KEY_SHL, $1, $3)}
		| BinaryExp KEY_SHR Product {$$ = EmitBinaryOp(KEY_SHR, $1, $3)}
	;

	Product:
		  Unary
		| Product '*' Unary {$$ = EmitBinaryOp('*', $1, $3)}
		| Product '/' Unary {$$ = EmitBinaryOp('/', $1, $3)}
	;

	Unary:
		  Exponential
		| KEY_VARPTR Exponential {$$ = EmitAddress($2)}
		| '-' Exponential {$$ = EmitUnaryOp('-', $2)}
		| '~' Exponential {$$ = EmitUnaryOp('~', $2)}
		| '+' Exponential {$$ = $2}
	;

	Exponential:
		  Factor
		| Factor '^' Exponential {$$ = EmitBinaryOp('^', $1, $3)}
	;

	Factor:
		  NonVariableFactor
		| Variable
	;

	NonVariableFactor:
		  '(' Expression ')' {$$ = EmitGrouping($2)}
		| '[' ExpressionList ']' {$$ = EmitAutoArray($2)}
		| FunctionCall
		| Datatype '(' Expression ')' {$$ = EmitTypeCast($1, $3)}
		| NonVariableFactor '[' NonEmptyExpressionList ']' {$$ = EmitIndex($1, $3)}
		| AmbiguousBracket      NonEmptyExpressionList ']' {$$ = EmitIndex(EmitVariable($1, -1), $2)}
		| Factor '.' TOK_IDENTIFIER {$$ = EmitMemberAccess($1, $3)}

		//Shift/reduce conflict: new Foo[5] could be (new Foo)[5] or a new array of type Foo and size 5.
        //Likewise, new Foo()[5] could be (new Foo)()[5] or a new array of function pointers returning Foo
		//Shifting is the correct resolution, although avoiding the conflict altogether would be preferrable
		| KEY_NEW TOK_IDENTIFIER {$$ = EmitNew($2, -1)}
        | KEY_NEW AmbiguousParan ExpressionList ')' {$$ = EmitNew($2, $3)}
		| KEY_NEW Datatype '[' NonEmptyExpressionList ']' {$$ = EmitNewArray($2, $4)}
        
        | KEY_NEW AmbiguousBracket NonEmptyExpressionList ']' {$$ = EmitNewArray(EmitDatatype(TYPE_COMPOUND, $2, 0), $3)}

		| Literal
	;

	OneLineIf: KEY_IF Expression KEY_THEN StatementLine OneLineElseIfList OneLineElse {$$ = EmitIf($2, $4, $5, $6)};

	OneLineElseIfList:
		  /* empty */                                                {$$ = EmitEmptyList()}
		| OneLineElseIfList KEY_ELSEIF Expression KEY_THEN StatementLine {AppendToList($1, EmitElseIf($3, $5))}
	;

	OneLineElse:
		  /* empty */            {$$ = -1}
		| KEY_ELSE StatementLine {$$ = EmitElse($2)}
	;

	IfStatement: KEY_IF Expression KEY_THEN LineDelimiter StatementList ElseIfList OptionalElse KEY_ENDIF {$$ = EmitIf($2, $5, $6, $7)};

	ElseIfList:
		  /* empty */                                                           {$$ = EmitEmptyList()}
		| ElseIfList KEY_ELSEIF Expression KEY_THEN LineDelimiter StatementList {AppendToList($1, EmitElseIf($3, $6))}
	;

	OptionalElse:
		  /* empty */                          {$$ = -1}
		| KEY_ELSE LineDelimiter StatementList {$$ = EmitElse($3)}
	;

	WhileStatement: KEY_WHILE Expression LineDelimiter StatementList KEY_WEND {$$ = EmitWhile($2, $4)};

	RepeatStatement:
		  KEY_REPEAT LineDelimiter StatementList KEY_UNTIL Expression {$$ = EmitRepeat($5, $3)}
		| KEY_REPEAT LineDelimiter StatementList KEY_FOREVER          {$$ = EmitRepeat(-1, $3)}
	;

	SelectStatement:
		  KEY_SELECT Expression LineDelimiter CaseStatementList DefaultStatement CaseStatementList KEY_ENDSELECT {$$ = EmitSelect($2, $4, $5, $6)}
		| KEY_SELECT Expression LineDelimiter CaseStatementList KEY_ENDSELECT                                    {$$ = EmitSelect($2, $4, -1, -1)}
	;

	DefaultStatement: KEY_DEFAULT LineDelimiter StatementList {$$ = EmitDefault($3)};

	CaseStatementList:
		  /* empty */                                                       {$$ = EmitEmptyList()}
		| CaseStatementList KEY_CASE Expression LineDelimiter StatementList {AppendToList($1, EmitCase($3, $5))}
	;

	ForWhileStatement: KEY_FOR StatementLine KEY_WHILE Expression KEY_DO OptionalLineDelimiter StatementList KEY_NEXT  {$$ = EmitForWhile($2, $4, $7)};

	ForStatement:
		  KEY_FOR Equality KEY_TO    Expression OptionalStep LineDelimiter StatementList KEY_NEXT {$$ = EmitFor(0, $2,      -1,    $4,      $5,           0, $7)}
		| KEY_FOR Equality KEY_UNTIL Expression OptionalStep LineDelimiter StatementList KEY_NEXT {$$ = EmitFor(0, $2,      -1,    $4,      $5,           1, $7)}
		| KEY_FOR KEY_LOCAL VariableDecl '=' ForToExpression LineDelimiter StatementList KEY_NEXT {$$ = EmitFor(1, $3, $5.From, $5.To, $5.Step, $5.SkipLast, $7)}
	;

	ForToExpression:
		  Expression KEY_TO    Expression OptionalStep {$$.From = $1; $$.To = $3; $$.Step = $4; $$.SkipLast = 0;}
		| Expression KEY_UNTIL Expression OptionalStep {$$.From = $1; $$.To = $3; $$.Step = $4; $$.SkipLast = 1;}
	;


	OptionalStep:
		  /* empty */         {$$ = -1}
		| KEY_STEP Expression {$$ = $2}
	;

	ClassDeclaration: KEY_CLASS TOK_IDENTIFIER ExtendList InheritanceFlags LineDelimiter DeclarationList KEY_ENDCLASS {$$ = EmitClass($2, $3, $6, $4)};
	
	ExtendList:
		  /* empty */ {$$ = -1}
		| NonEmptyExtendList
	;
	
	NonEmptyExtendList:
		  KEY_EXTENDS    TOK_IDENTIFIER {$$ = EmitEmptyList(); AppendStringToList($$, $2)}
		| ExtendList ',' TOK_IDENTIFIER {AppendStringToList($1, $3)}
	;
		
	InheritanceFlags:
		  /* empty */                   {$$ = 0}
		| InheritanceFlags KEY_ABSTRACT {$$ = $1 | FLAG_ABSTRACT}
		| InheritanceFlags KEY_FINAL    {$$ = $1 | FLAG_FINAL}
	;
		
	OptionalFinal:
		  /* empty */ {$$ = 0}
		| KEY_FINAL   {$$ = FLAG_FINAL}
	;

	DeclarationList:
		  /* empty */                                       {$$ = EmitEmptyList(); SetVisibility(KEY_PUBLIC)}
		| DeclarationList GlobalDeclaration   LineDelimiter {AppendToList($1, $2)}
		| DeclarationList ConstDeclaration    LineDelimiter {AppendToList($1, $2)}
		| DeclarationList FieldDeclaration    LineDelimiter {AppendToList($1, $2)}
		| DeclarationList MethodDeclaration   LineDelimiter {AppendToList($1, $2)}
		| DeclarationList FunctionDeclaration LineDelimiter {AppendToList($1, $2)}
		| DeclarationList KEY_PRIVATE         LineDelimiter {SetVisibility(KEY_PRIVATE)}
		| DeclarationList KEY_PUBLIC          LineDelimiter {SetVisibility(KEY_PUBLIC)}
        | DeclarationList KEY_PROTECTED       LineDelimiter {SetVisibility(KEY_PROTECTED)}
	;
        
    FunctionDecl:
          TOK_IDENTIFIER ':' FunctionDatatype       {$$ = EmitVariable($1, $3)}
        | TOK_IDENTIFIER ':' Datatype '(' ')'       {$$ = EmitVariable($1, EmitDatatype(TYPE_FUNCTION, $3, EmitEmptyList()))}
        | TOK_IDENTIFIER ':' TOK_IDENTIFIER '(' ')' {$$ = EmitVariable($1, EmitDatatype(TYPE_FUNCTION, EmitDatatype(TYPE_COMPOUND, $3, 0), EmitEmptyList()))}
        | TOK_IDENTIFIER '(' VarDeclList ')'        {$$ = EmitVariable($1, EmitDatatype(TYPE_FUNCTION, EmitDatatype(TYPE_VOID, -1, 0), $3))}
    ;

	MethodDeclaration  : KEY_METHOD   FunctionDecl OptionalFinal LineDelimiter StatementList KEY_ENDMETHOD   {$$ = EmitFunction($2, $5, 1, $3)}
                       | KEY_METHOD   FunctionDecl KEY_ABSTRACT                                              {$$ = EmitFunction($2, EmitEmptyList(), 1, FLAG_ABSTRACT)};
	FunctionDeclaration: KEY_FUNCTION FunctionDecl OptionalFinal LineDelimiter StatementList KEY_ENDFUNCTION {$$ = EmitFunction($2, $5, 0, $3)}
	                   | KEY_FUNCTION FunctionDecl KEY_ABSTRACT                                              {$$ = EmitFunction($2, EmitEmptyList(), 0, FLAG_ABSTRACT)};

	ExternClass: KEY_CLASS TOK_IDENTIFIER ExtendList InheritanceFlags LineDelimiter ExternDeclarationList KEY_ENDCLASS {$$ = EmitClass($2, $3, $6, $4)};

	ExternDeclarationList:
		  /* empty */                                           {$$ = EmitEmptyList(); SetVisibility(KEY_PUBLIC)}
		| ExternDeclarationList GlobalDeclaration LineDelimiter {AppendToList($1, $2)}
		| ExternDeclarationList ConstDeclaration  LineDelimiter {AppendToList($1, $2)}
		| ExternDeclarationList FieldDeclaration  LineDelimiter {AppendToList($1, $2)}
		| ExternDeclarationList ExternMethod      LineDelimiter {AppendToList($1, $2)}
		| ExternDeclarationList ExternFunction    LineDelimiter {AppendToList($1, $2)}
        | ExternDeclarationList KEY_PRIVATE       LineDelimiter {SetVisibility(KEY_PRIVATE)}
		| ExternDeclarationList KEY_PUBLIC        LineDelimiter {SetVisibility(KEY_PUBLIC)}
        | ExternDeclarationList KEY_PROTECTED     LineDelimiter {SetVisibility(KEY_PROTECTED)}
	;

	ExternMethod  : KEY_METHOD   FunctionDecl InheritanceFlags {$$ = EmitFunction($2, EmitEmptyList(), 1, $3)};
	ExternFunction: KEY_FUNCTION FunctionDecl InheritanceFlags {$$ = EmitFunction($2, EmitEmptyList(), 0, $3)};

	GlobalDeclaration  : KEY_GLOBAL NonEmptyVarDeclList {ScopeVarDecl(KEY_GLOBAL, $2); $$ = $2};
	LocalDeclaration   : KEY_LOCAL  NonEmptyVarDeclList {ScopeVarDecl(KEY_LOCAL , $2); $$ = $2};
	ConstDeclaration   : KEY_CONST  NonEmptyVarDeclList {ScopeVarDecl(KEY_CONST , $2); $$ = $2};
	FieldDeclaration   : KEY_FIELD  NonEmptyVarDeclList {ScopeVarDecl(KEY_FIELD , $2); $$ = $2};

	VarDeclList:
		  /* empty */         {$$ = EmitEmptyList()}
		| NonEmptyVarDeclList
	;

	NonEmptyVarDeclList:
		  VariableDecl OptionalAssignment                         {$$ = EmitSingleList(EmitVariableDecl($1, $2))}
		| NonEmptyVarDeclList ',' VariableDecl OptionalAssignment {AppendToList($1, EmitVariableDecl($3, $4))}
	;

	OptionalAssignment:
		  /* empty */    {$$ = -1}
		| '=' Expression {$$ = $2}
	;

	VariableDecl:
		  TOK_IDENTIFIER TypeSpecifier {$$ = EmitVariable($1, $2)}
	;

	Variable:
		  TOK_IDENTIFIER               {$$ = EmitVariable($1, -1)}
	;

	TypeSpecifier:
		  ':' DatatypeOrIdentifier {$$ = $2}
	;
		
	FunctionDatatype:
          Datatype '('   NonEmptyVarDeclList ')' {$$ = EmitDatatype(TYPE_FUNCTION, $1, $3)}
        | AmbiguousParan NonEmptyVarDeclList ')' {$$ = EmitDatatype(TYPE_FUNCTION, EmitDatatype(TYPE_COMPOUND, $1, 0), $2)}
        | Datatype '('   KEY_VOID            ')' {$$ = EmitDatatype(TYPE_FUNCTION, $1, EmitEmptyList())}
        | AmbiguousParan KEY_VOID            ')' {$$ = EmitDatatype(TYPE_FUNCTION, EmitDatatype(TYPE_COMPOUND, $1, 0), EmitEmptyList())}
    ;
		
	DatatypeOrIdentifier:
		  TOK_IDENTIFIER {$$ = EmitDatatype(TYPE_COMPOUND, $1, 0)}
		| Datatype
	;

	Datatype:
		  KEY_VOID                         {$$ = EmitDatatype(TYPE_VOID   , -1, 0)}
		| KEY_BOOL                         {$$ = EmitDatatype(TYPE_BOOL   , -1, 0)}
		| KEY_BYTE                         {$$ = EmitDatatype(TYPE_BYTE   , -1, 0)}
		| KEY_SHORT                        {$$ = EmitDatatype(TYPE_SHORT  , -1, 0)}
		| KEY_INT                          {$$ = EmitDatatype(TYPE_INT    , -1, 0)}
		| KEY_LONG                         {$$ = EmitDatatype(TYPE_LONG   , -1, 0)}
		| KEY_UBYTE                        {$$ = EmitDatatype(TYPE_UBYTE  , -1, 0)}
		| KEY_USHORT                       {$$ = EmitDatatype(TYPE_USHORT , -1, 0)}
		| KEY_UINT                         {$$ = EmitDatatype(TYPE_UINT   , -1, 0)}
		| KEY_ULONG                        {$$ = EmitDatatype(TYPE_ULONG  , -1, 0)}
		| KEY_FLOAT                        {$$ = EmitDatatype(TYPE_FLOAT  , -1, 0)}
		| KEY_DOUBLE                       {$$ = EmitDatatype(TYPE_DOUBLE , -1, 0)}
		| Datatype       KEY_PTR           {$$ = EmitDatatype(TYPE_POINTER, $1, 0)}
		| TOK_IDENTIFIER KEY_PTR           {$$ = EmitDatatype(TYPE_POINTER, EmitDatatype(TYPE_COMPOUND, $1, 0), 0)}
		| Datatype       '[' CommaList ']' {$$ = EmitDatatype(TYPE_ARRAY, $1, $3 + 1)}
		| AmbiguousBracket   CommaList ']' {$$ = EmitDatatype(TYPE_ARRAY, EmitDatatype(TYPE_COMPOUND, $1, 0), $2 + 1)}
        | FunctionDatatype
	;

	Literal:
		  TOK_BYTE   {$$ = EmitLiteral(TOK_BYTE  , $1)}
		| TOK_SHORT  {$$ = EmitLiteral(TOK_SHORT , $1)}
		| TOK_INT    {$$ = EmitLiteral(TOK_INT   , $1)}
		| TOK_LONG   {$$ = EmitLiteral(TOK_LONG  , $1)}
		| TOK_FLOAT  {$$ = EmitLiteral(TOK_FLOAT , $1)}
		| TOK_DOUBLE {$$ = EmitLiteral(TOK_DOUBLE, $1)}
		| TOK_STRING {$$ = EmitLiteral(TOK_STRING, $1)}
        | TOK_UBYTE  {$$ = EmitLiteral(TOK_UBYTE , $1)}
		| TOK_USHORT {$$ = EmitLiteral(TOK_USHORT, $1)}
		| TOK_UINT   {$$ = EmitLiteral(TOK_UINT  , $1)}
		| TOK_ULONG  {$$ = EmitLiteral(TOK_ULONG , $1)}
		| KEY_NULL   {$$ = EmitLiteral(KEY_NULL  , -1)}
		| KEY_SELF   {$$ = EmitLiteral(KEY_SELF  , -1)}
        | KEY_SUPER  {$$ = EmitLiteral(KEY_SUPER , -1)}
		| KEY_TRUE   {$$ = EmitLiteral(KEY_TRUE  , -1)}
		| KEY_FALSE  {$$ = EmitLiteral(KEY_FALSE , -1); @1.last_line = @1.last_line} //This may look weird, but don't remove it!
		//This forces bison to actually include yylloc, which would lead to linker errors otherwise
	;

	AmbiguousBracket:
		TOK_IDENTIFIER '['
	;
        
    AmbiguousParan:
		TOK_IDENTIFIER '('
	;

	CommaList:
		 /* empty */    {$$ = 0}
		| CommaList ',' {$$ = $1 + 1}
	;

	OptionalLineDelimiter:
		  /* empty */   {$$ = 0}
		| LineDelimiter {$$ = 0}
	;

	LineDelimiter:
		  TOK_EOL {$$ = -1}
		| ';'     {$$ = -1}
	;
%%

void yyerror (char const *s) {
}
