Type CLI

	Field Arguments:TList
	Field Config:String[]
	Field InputFilePath:String
	Field InputFile:String

	Method New()
		Arguments = New TList
	EndMethod

	Method AddArgument(Arg:CLIArgument)
		Arguments.AddLast(Arg)
	EndMethod

	Method ParseCommandLine:Int(Args:String[])
		Config = New String[Arguments.Count()]
		Local LastIndex:Int = 0 'LastIndex marks where the last argument or its param has been found

		For Local a:CLIArgument = EachIn Arguments
			Config[a.ConfigValue] = a.DefaultValue

			For Local index:Int = 0 Until Args.length
				Local c:String = Args[index]

				'Strip starting dashes
				If c.StartsWith("-") Then c = c[1..]
				If c.StartsWith("-") Then c = c[1..]

				'Is c the argument we are looking for?
				If a.HasName(c) Then
					If index > LastIndex Then LastIndex = index

					If a.ParamCount = 0 Then
						Config[a.ConfigValue] = "on"
					Else
						'TODO allow and parse multiple parameters (e.g. -f file1,file2,file3)
						If index+1 < Args.length Then
							Config[a.ConfigValue] = Args[index+1]

							If index+1 > LastIndex Then LastIndex = index+1
						Else
							Print "Command line error, " + a.GetNames() + " needs a parameter!"
							Return False
						EndIf
					EndIf
				EndIf

			Next
		Next

		'Get file name
		LastIndex:+1

		If LastIndex >= Args.length Then
			Print "Command line error, input file omitted!"
			Return False
		EndIf
	
		InputFile = Args[LastIndex]
		InputFilePath = GetAbsolutePath(Args[LastIndex])

		Return True
	EndMethod

	Method GetUsageDescription:String()
		'TODO change to real binary name
		Local Out:String = "compiler "
		Local OutLong:String
		
		For Local a:CLIArgument = EachIn Arguments
			Out:+a.GetShortDescription() + " "
			OutLong:+a.GetLongDescription() + "~n"
		Next

		Return Out + "FILE ~n" + OutLong
	EndMethod

EndType

Type CLIArgument

	Field ShortName:String      'Short Option Name (e.g. -o)
	Field LongName:String        'Long Option Name (e.h. --output)
	Field Description:String    'Description of the argument
	Field ParamCount:Int        'Number of parameters
	Field Optional:Int          
	Field ConfigValue:Int        'Position in config array
	Field DefaultValue:String

	Function Create:CLIArgument(ShortN:String, LongN:String, Desc:String, ParamC:Int, Opt:Int, Def:String, Conf:Int)
		Local a:CLIArgument = New CLIArgument
		a.ShortName = ShortN
		a.LongName = LongN
		a.Description = Desc
		a.ParamCount = ParamC
		a.Optional = Opt
		a.ConfigValue = Conf
		a.DefaultValue = Def

		Return a
	EndFunction

	Method GetLongDescription:String()
		Local Out:String

		If ShortName <> "" Then Out:+ "-" + ShortName
		If LongName <> "" Then
			If Out <> "" Then Out:+ ", "
			Out:+ "--" + LongName
		EndIf

		Out:+ "~t~t" + Description
		If ParamCount > 0 Then Out:+ " (default: ~q" + DefaultValue + "~q)"
		Return Out
	EndMethod

	Method GetShortDescription:String()
		Local Out:String

		If ShortName <> "" Then Out:+ "-" + ShortName
		If LongName <> "" Then
			If Out <> "" Then Out:+ "|"
			Out:+ "--" + LongName
		EndIf

		If ParamCount > 0 Then Out:+ " <param>"
		If Optional Then Out = "[" + Out + "]"

		Return Out
	EndMethod

	Method GetNames:String()
		Local Out:String

		If ShortName <> "" Then Out:+ "-" + ShortName
		If LongName <> "" Then
			If Out <> "" Then Out:+ "|"
			Out:+ "--" + LongName
		EndIf

		Return Out
	EndMethod

	Method HasName:Int(Name:String)

		If ShortName And ShortName = Name Return True
		If LongName And LongName = Name Return True

		Return False

	EndMethod

EndType
