Type TLexer
	Global Lexer:TLexer
	
	Global KeywordMap:TMap
	Global IndexToKeyword:TMap
	
	Global TokenStartLine:Int, TokenEndLine:Int
	Global TokenStartCol:Int,  TokenEndCol:Int
	
	Global TokenLoc:Int Ptr
	
	Global LastToken:Int
	
	Field Path:String
	
	Field Code:String
	Field CodePos:Int
	
	Field CurrentLine:Int
	Field LastLinePos:Int
	
	Field LastLiteralType:Int
	
	Global IsFirstToken:Int
	
	Function CalculateLocation()
		TokenStartLine = TokenLoc[START_LINE]
		TokenStartCol = TokenLoc[START_COL]
		TokenEndLine = TokenLoc[END_LINE]
		TokenEndCol = TokenLoc[END_COL]
	End Function
	
	Method Abort(Error:String)
		WriteStdout("Lexer error: " + Error + ..
			"~nIn file " + Path + ..
			"~nIn line " + TokenLoc[START_LINE] + "-" + TokenLoc[END_LINE] + ..
			"~nIn column " + TokenLoc[START_COL] + "-" + TokenLoc[END_COL] + "~n" ..
		)
		
		QuitProgram(-1)
	End Method
	
	Function GetNextTokenStatic:Int()
		Local Token:Int = Lexer.GetNextToken()
		
		LastToken = Token
		
		'?Debug
		'	Print Lexer.TokenToString(Token)
		'?
		
		Return Token
	End Function
	
	Function Load(Path:String)
		If Not Lexer Then Lexer = New TLexer
		If Not KeywordMap Then Init()
		
		Lexer.Path = Path
		Lexer.CurrentLine = 1
		Lexer.LastLinePos = 0
		Lexer.CodePos = 0
		LastToken = TOK_EOL
		
		Try
			Lexer.Code = LoadText(Path).Replace("~r", "") + "~n"
		Catch Error:Object
			Lexer.Abort("Unable to open file " + Path)
		EndTry
	End Function
	
	Method SetStringResult(Result:String)
		Local CString:Byte Ptr = Result.ToCString()
		
		yylval = (Int Ptr (Varptr CString))[0]
		
		LastLiteralType = TOK_STRING
	End Method
	
	Method SetFloatResult(Result:Float)
		yylval = (Int Ptr (Varptr Result))[0]
		LastLiteralType = TOK_FLOAT
	End Method
	
	Method SetDoubleResult(Result:Double)
		yylval = (Long Ptr (Varptr Result))[0]
		LastLiteralType = TOK_DOUBLE
	End Method
	
	Method SetByteResult(Result:Byte, IsUnsigned:Int)
		yylval = Result
		LastLiteralType = TOK_BYTE + (TOK_UBYTE - TOK_BYTE)*IsUnsigned
	End Method
	
	Method SetShortResult(Result:Short, IsUnsigned:Int)
		yylval = Result
		LastLiteralType = TOK_SHORT + (TOK_UBYTE - TOK_BYTE)*IsUnsigned
	End Method
	
	Method SetIntResult(Result:Int, IsUnsigned:Int)
		yylval = Result
		LastLiteralType = TOK_INT + (TOK_UBYTE - TOK_BYTE)*IsUnsigned
	End Method
	
	Method SetLongResult(Result:Long, IsUnsigned:Int)
		yylval = Result
		LastLiteralType = TOK_LONG + (TOK_UBYTE - TOK_BYTE)*IsUnsigned
	End Method
	
	Method GetNextToken:Int()
		If CodePos = Code.Length Then Return TOK_EOF
		
		TokenLoc[START_LINE] = CurrentLine
		TokenLoc[START_COL]  = CodePos - LastLinePos + 1
		
		Local Char:String = GetNextChar()
		
		While Char = " " Or Char = "~t"
			Char = GetNextChar()
			
			TokenLoc[START_COL] :+ 1
		Wend
		
		TokenLoc[END_LINE] = TokenLoc[START_LINE]
		TokenLoc[END_COL]  = TokenLoc[START_COL] + 1
		
		Select Char
			Case "+"
				Return TOK_PLUS
			Case "-"
				Return TOK_MINUS
			Case "*"
				Return TOK_ASTERISK
			Case "/"
				Local C:String = GetNextChar()
				If C = "/" Then
					While C <> "~n"
						C = GetNextChar()
					Wend
					
					CodePos :- 1
					
					Return GetNextToken()
				ElseIf C = "*" Then
					While True
						If CodePos = Code.Length Then Abort("/* without matching */")
						
						C = GetNextChar()
						
						If C = "*" Then
							If GetNextChar() = "/" Then Exit Else CodePos :- 1
						ElseIf C = "~n" Then
							LastLinePos = CodePos
							CurrentLine :+ 1
						EndIf
					Wend
					
					Return GetNextToken()
				Else
					CodePos :- 1
					
					Return TOK_SLASH
				EndIf
			Case "("
				Return TOK_LEFTPARAN
			Case ")"
				Return TOK_RIGHTPARAN
			Case "^"
				Return TOK_CARET
			Case ":"
				Return TOK_COLON
			Case ";"
				Return TOK_SEMICOLON
			Case ","
				Return TOK_COMMA
			Case "."
				If GetNextChar() = "." Then
					If GetNextChar() = "~n" Then
						LastLinePos = CodePos
						CurrentLine :+ 1
						
						Return GetNextToken()
					Else
						CodePos :- 1
					EndIf
					
					TokenLoc[END_COL] :+ 1
					
					Return TOK_DOUBLEPERIOD
				Else
					CodePos :- 1
					Return TOK_PERIOD
				EndIf
			Case "&"
				Return TOK_AMPERSAND
			Case "|"
				Return TOK_PIPE
			Case "["
				Return TOK_LEFTBRACKET
			Case "]"
				Return TOK_RIGHTBRACKET
			Case "~q"
				GetNextChar()
				
				SetStringResult(GetNextString())
				
				TokenLoc[END_LINE] = CurrentLine
				TokenLoc[END_COL]  = CodePos - LastLinePos + 1
				
				Return TOK_STRING
			Case "'"
				GetNextCharConstant()
				TokenLoc[END_COL] = CodePos - LastLinePos + 1
				
				Return LastLiteralType
			Case "~~"
				Return TOK_TILDE
			Case "$"
				GetNextChar()
				
				GetNextHexNumber()
				
				TokenLoc[END_COL] = CodePos - LastLinePos + 1
				
				Return LastLiteralType
			Case "%"
				GetNextChar()
				
				GetNextBinaryNumber()
				
				TokenLoc[END_COL] = CodePos - LastLinePos + 1
				
				Return LastLiteralType
			Case "~n"
				LastLinePos = CodePos
				CurrentLine :+ 1
				If LastToken = TOK_EOL Then
					Return GetNextToken()
				Else
					Return TOK_EOL
				EndIf
			Case ">"
				If GetNextChar() = "=" Then
					TokenLoc[END_COL] :+ 1
					
					Return TOK_GREATEREQUAL
				Else
					CodePos :- 1
					Return TOK_GREATER
				EndIf
			Case "<"
				If GetNextChar() = "=" Then
					TokenLoc[END_COL] :+ 1
					
					Return TOK_SMALLEREQUAL
				Else
					CodePos :- 1
					
					If GetNextChar() = ">" Then
						TokenLoc[END_COL] :+ 1
						
						Return TOK_UNEQUAL
					Else
						CodePos :- 1
						Return TOK_SMALLER
					EndIf
				EndIf
			Case "="
				Select GetNextChar()
					Case ">"
						TokenLoc[END_COL] :+ 1
						
						Return TOK_GREATEREQUAL
					Case "<"
						TokenLoc[END_COL] :+ 1
						
						Return TOK_SMALLEREQUAL
					Default
						CodePos :- 1
						Return TOK_EQUAL
				End Select
			Default
				If IsDigit(Char) Then
					If CheckForFloat() Then
						GetNextFloat()
						
						TokenLoc[END_COL] = CodePos - LastLinePos + 1
					Else
						GetNextInteger()
						
						TokenLoc[END_COL] = CodePos - LastLinePos + 1
					EndIf
					
					Return LastLiteralType
				EndIf
				
				If IsValidChar(Char) Then
					Local Identifier:String = GetNextIdentifier()
					
					If Identifier.ToLower() = "end" Then
						Local OldPos:Int = CodePos
						
						If GetNextChar() = " " And IsValidChar(GetNextChar()) Then
							Local NextIdent:String = GetNextIdentifier()
							
							Select NextIdent.ToLower()
								Case "select", "extern", "if", "class", "function", "method"
									Identifier :+ NextIdent
								Default
									CodePos = OldPos
							End Select
						Else
							CodePos = OldPos
						EndIf
					EndIf
					
					TokenLoc[END_COL] = CodePos - LastLinePos + 1
					
					Local Result:String = String(KeywordMap.ValueForKey(Identifier.ToLower()))
					If Result Then
						Return Result.ToInt()
					Else
						SetStringResult(Identifier)
						
						Return TOK_IDENTIFIER
					EndIf
				EndIf
				
				Abort("Unknown literal")
		End Select
	End Method
	
	Method GetNextChar:String()
		If CodePos >= Code.Length Then Abort("Lexer error")
		
		Local Char:String = Chr(Code[CodePos])
		CodePos :+ 1
		
		Return Char
	End Method
	
	Method GetNextIdentifier:String()
		CodePos :- 1
		
		Local Output:String, Char:String = GetNextChar()
		
		While IsValidChar(Char)
			Output :+ Char
			
			Char = GetNextChar()
		Wend
		
		CodePos :- 1
		
		Return Output
	End Method
	
	Method GetNextString:String()
		CodePos :- 1
		
		Local Output:String, Char:String = GetNextChar()
		
		While Char <> "~q"
			If Char = "~n" Then
				LastLinePos = CodePos
				CurrentLine :+ 1
			EndIf
			
			If Char = "~~" Then
				Output :+ UnescapeChar(GetNextChar())
			Else
				Output :+ Char
			EndIf
			
			Char = GetNextChar()
		Wend
		
		Return Output
	End Method
	
	Method GetNextCharConstant()
		Local Output:Long, Char:String = GetNextChar(), CharCount:Int
		
		While Char <> "'" And Char <> "~n" And Char <> "~r"
			If Char = "~~" Then Char = UnescapeChar(GetNextChar())
			
			Output = (Output Shl 8) | Char[0]
			CharCount :+ 1
			
			Char = GetNextChar()
		Wend
		
		If CharCount = 0 Then
			Abort("Character constant has to have at least one character")
		ElseIf CharCount = 1 Then
			SetByteResult(Output, True)
		ElseIf CharCount = 2 Then
			SetShortResult(Output, True)
		ElseIf CharCount <= 4 Then
			SetIntResult(Output, True)
		ElseIf CharCount <= 8 Then
			SetLongResult(Output, True)
		Else
			Abort("Too many characters in character constant")
		EndIf
	End Method
	
	Method GetNextInteger()
		CodePos :- 1
		
		Local Output:Long, Char:String = GetNextChar()
		
		While IsDigit(Char)
			Output = Output*10 + (Char[0] - "0"[0])
			
			Char = GetNextChar()
		Wend
		
		GetLiteralFlags(Output, Char)
	End Method
	
	Method GetNextFloat()
		CodePos :- 1
		
		Local Output:Double, Char:String = GetNextChar()
		
		While IsDigit(Char)
			Output = Output*10.0 + (Char[0] - "0"[0])
			
			Char = GetNextChar()
		Wend
		
		If Char = "." Then
			Char = GetNextChar()
			
			Local Exponent:Float = 0.1
			While IsDigit(Char)
				Output :+ (Char[0] - "0"[0])*Exponent
				Exponent :* 0.1
				
				Char = GetNextChar()
			Wend
		EndIf
		
		If Char.ToLower() = "e" Then
			Char = GetNextChar()
			
			Abort("Scientific notation not yet supported")
		Else
			CodePos :- 1
		EndIf
		
		If Char.ToUpper() = "D" Then
			SetDoubleResult(Output)
			CodePos :+ 1
		Else
			SetFloatResult(Output)
		EndIf
	End Method
	
	Method GetNextHexNumber()
		CodePos :- 1
		
		Local Output:Long, Char:String = GetNextChar()
		
		While IsHexDigit(Char)
			If IsDigit(Char) Then
				Output = (Output Shl 4) + (Char[0] - "0"[0])
			Else
				Output = (Output Shl 4) + ((Char.ToUpper())[0] - "A"[0] + 10)
			EndIf
			
			Char = GetNextChar()
		Wend
		
		GetLiteralFlags(Output, Char)
	End Method
	
	Method GetNextBinaryNumber()
		CodePos :- 1
		
		Local Output:Long, Char:String = GetNextChar()
		
		While IsBinaryDigit(Char)
			Output = (Output Shl 1) + (Char[0] - "0"[0])
			
			Char = GetNextChar()
		Wend
		
		GetLiteralFlags(Output, Char)
	End Method
	
	Method GetLiteralFlags(Output:Long, Char:String)
		Local IsUnsigned:Int
		While True
			Select Char.ToUpper()
				Case "U"
					IsUnsigned = True
				Case "L"
					SetLongResult(Output, IsUnsigned);  Exit
				Case "S"
					SetShortResult(Output, IsUnsigned); Exit
				Case "B"
					SetByteResult(Output, IsUnsigned);  Exit
				Default
					SetIntResult(Output, IsUnsigned)
					
					CodePos :- 1
					
					Exit
			End Select
			
			Char = GetNextChar()
		Wend
	End Method
	
	Method CheckForFloat:Int()
		Local OldPos:Int = CodePos, Char:String = GetNextChar()
		
		While IsDigit(Char)
			Char = GetNextChar()
		Wend
		
		CodePos = OldPos
		
		If Char = "." Or Char.ToLower() = "e" Then Return True
	End Method
	
	Method UnescapeChar:String(Char:String)
		Select Char
			Case "n"
				Return "~n"
			Case "r"
				Return "~r"
			Case "~~"
				Return "~~"
			Case "t"
				Return "~t"
			Case "~q"
				Return "~q"
			Case "'"
				Return "'"
			Default
				Abort("Invalid escape character: " + Char)
		End Select
	End Method
	
	Function TokenToString:String(Token:Int)
		Local Keyword:String = String(IndexToKeyword.ValueForKey(String.FromInt(Token)))
		If Keyword Then
			Return Keyword
		ElseIf Token <= 127 Then
			Return Chr(Token)
		ElseIf Token >= 401 And Token <= 428 Then
			Local Symbols:String[] = ["+", "-", "*", "/", "^", "(", ")", ">", "<", "=", ">=", "<=", "<>", "~~", "|", "&", ":", ",", ".", ";", "$", "!", "@", "%", "'", "[", "]", "#", ".."]
			
			Return Symbols[Token - 401]
		Else
			Select Token
				Case TOK_EOL
					Return "New Line"
				Case TOK_IDENTIFIER
					Return "Identifier"
				Case TOK_INT
					Return "Int literal"
				Case TOK_FLOAT
					Return "Float literal"
				Case TOK_STRING
					Return "String literal"
			End Select
		EndIf
	End Function
	
	Function Init()
		KeywordMap = New TMap
		IndexToKeyword = New TMap
		
		Local Keywords:String[] = ["", "Abstract", "And", "", "Assert", "Byte", "Case", "Catch", "", "Const",  ..
		"Continue", "Default", "", "Delete", "Double", "EachIn", "Else", "ElseIf", "", "EndExtern", "EndFunction",  ..
		"EndIf", "EndMethod", "", "EndSelect", "EndTry", "EndClass", "EndWhile", "Exit", "Extends", "Extern","False",  ..
		"Field", "Final", "Float", "For", "Forever", "", "Function", "Global", "", "If", "Import", "Incbin",  ..
		"IncbinLen", "IncbinPtr", "", "Int", "", "Local", "Long", "", "Method", "", "Mod", "", "",  ..
		"New", "Next", "Not", "Null", "", "Or", "", "Private", "Ptr", "Public", "", "", "", "Repeat",  ..
		"" , "Return" , "" , "Select" , "Self" , "" , "Shl" , "Short" , "Shr" , "SizeOf" , "Step" , "" , "" , ..'"String",  ..
		"Super", "", "Then", "Throw", "To","True", "Try", "Class", "Until", "Var", "Varptr", "Wend", "While", "Bool", "Do", "UByte", "UShort", "UInt", "ULong", "Void", "Protected"]
		
		Local SymbolIndex:Int = KEY_ABS
		For Local I:String = EachIn Keywords
			If I Then 'TODO: Find a proper way to resolve keywords to name
				KeywordMap.Insert(I.ToLower(), String.FromInt(SymbolIndex))
				IndexToKeyword.Insert(String.FromInt(SymbolIndex), I)
			EndIf
			
			SymbolIndex :+ 1
		Next
	End Function
	
	Function IsValidChar:Int(Char:String)
		If Char[0] >= "A"[0] And Char[0] <= "Z"[0] Then Return True
		If Char[0] >= "a"[0] And Char[0] <= "z"[0] Then Return True
		If Char[0] >= "0"[0] And Char[0] <= "9"[0] Then Return True
		
		If Char[0] = "_"[0] Then Return True
	End Function
	
	Function IsHexDigit:Int(Char:String)
		If Char.ToLower() >= "a" And Char.ToLower() <= "f" Then Return True
		If Char >= "0" And Char <= "9" Then Return True
	End Function
	
	Function IsBinaryDigit:Int(Char:String)
		If Char = "1" Or Char = "0" Then Return True
	End Function
	
	Function IsDigit:Int(Char:String)
		If Asc(Char) >= 48 And Asc(Char) <= 57 Then Return True
	End Function
	
	Function IsChar:Int(Char:String)
		If Asc(Char) >= 65 And Asc(Char) <= 90 Then Return True
		If Asc(Char) >= 97 And Asc(Char) <= 122 Then Return True
	End Function
End Type
