'TODO: Classes
'TODO: Enums

SuperStrict

Import BaH.Libxml

Global TypedefMap:TMap, KeywordMap:TMap, SymbolMap:TMap

Init()

Local InputFile:String, IsTempFile:Int

'ChangeDir(LaunchDir)

If AppArgs.Length = 2 Then
	InputFile = RealPath(AppArgs[1])
ElseIf AppArgs.Length > 2 Then
	IsTempFile = True
	
	Local OutText:String
	For Local I:Int = 1 Until AppArgs.Length
		If FileType(AppArgs[I]) = FILETYPE_FILE Then
			OutText :+ LoadText(AppArgs[I]) + "~n~n"
		Else
			Print "Invalid input file(s)"
			End
		EndIf
	Next
	
	InputFile = RealPath(StripExt(AppArgs[1]) + "-Externiated.h")
	
	SaveText(OutText, InputFile)
Else
	Print "Command line error"
	End
EndIf

If FileType(InputFile) <> FILETYPE_FILE Then Print "Invalid input file"; End

Local XMLPath:String = StripExt(InputFile) + "_wrap.xml"
Local BiPath :String = StripExt(InputFile) + ".bi"

InvokeSWIG(InputFile)

Local Outstream:TStream = WriteFile(BiPath)
If Not Outstream Then Print "Failed to write *.bi file"; End

Outstream.WriteLine("Extern")

Local Doc :TxmlDoc  = TxmlDoc.ParseFile(XMLPath)
If Not Doc Then Print "Failed to read XML file"; End

'Local Result:TxmlXPathObject = Doc.newXPathContext().evalExpression("/top/include[attributelist/attribute/@value=~q" + InputFile + "~q]")
'Local Root:TxmlNode          = TxmlNode(Result.GetNodeSet().GetNodeList().First())
Local Root:TxmlNode = TxmlNode(Doc.GetRootElement().GetLastChild())

For Local Child:TxmlNode = EachIn Root.GetChildren()
	Select Child.GetName().ToLower()
		Case "constant"
			WrapConstant(Child, Outstream)
		Case "cdecl"
			WrapCDecl(Child, Outstream)
	End Select
Next

Outstream.WriteLine("EndExtern")
Outstream.Close()

xmlCleanupParser()

'DeleteFile XMLPath
If IsTempFile Then DeleteFile InputFile

End


Function InvokeSWIG(InputFile:String)
	Local CMD:String = "swig -xml -module ~q" + StripAll(InputFile) + "~q ~q" + InputFile + "~q"
	Local Proc:TProcess = CreateProcess(CMD, HIDECONSOLE)
	
	If Proc Then
		Local Output:String
		While Proc.Status()
			Output :+ ReadPipe(Proc.Err)
		Wend
		
		If Output Then
			Print "SWIG output:"
			Print Output
		EndIf
	Else
		Print "Failed to invoke SWIG"
		End
	EndIf
	
	Function ReadPipe:String(Pipe:TPipeStream)
		Local OutputSize:Int = Pipe.ReadAvail()

		If OutputSize Then
			Local OutputBytes:Byte[] = New Byte[OutputSize]
			Pipe.Read(Varptr OutputBytes[0], OutputSize)

			Return String.FromBytes(Varptr OutputBytes[0], OutputSize)
		Else
			Return Null
		EndIf
	End Function
End Function

Function TryDefineSymbol:Int(Identifier:String)
	If SymbolMap.Contains(Identifier.ToLower()) Then
		Print "Warning: Skipping duplicate identifier '" + Identifier + "'."
		Return False
	Else
		SymbolMap.Insert(Identifier.ToLower(), Identifier)
		Return True
	EndIf
End Function

Function WrapConstant(Node:TxmlNode, Outstream:TStream)
	Local Attributes:TList = Node.GetFirstChild().GetChildren()
	
	Local Name :String = FindAttribute(Attributes, "name")
	Local DType:String = FindAttribute(Attributes, "type")
	Local Value:String = FindAttribute(Attributes, "value")
	
	If TryDefineSymbol(Name) Then Outstream.WriteLine("~tConst " + Name + ":" + WrapDatatype(DType) + " = " + WrapLiteral(Value))
End Function

Function WrapCDecl(Node:TxmlNode, Outstream:TStream)
	Local Attributes:TList = Node.GetFirstChild().GetChildren()
	
	Local Kind :String = FindAttribute(Attributes, "kind")
	Local Name :String = FindAttribute(Attributes, "name")
	Local DType:String = FindAttribute(Attributes, "type")
	Local Decl :String = FindAttribute(Attributes, "decl")
	
	Select Kind.ToLower()
		Case "function"
			Local Out:String = "Function " + WrapIdentifier(Name) + ":" + WrapDatatype(DType) + "("
			
			Local ParamNode:TxmlNode = FindChild(Node.GetFirstChild(), "parmlist")
			
			If ParamNode Then
				Local Params:TList = ParamNode.GetChildren()
				
				For Local Param:TxmlNode = EachIn Params
					Local ParamAttr:TList = Param.GetFirstChild().GetChildren()
					
					Local ParamName:String = FindAttribute(ParamAttr, "name")
					Local ParamType:String = FindAttribute(ParamAttr, "type")
					
					If ParamName Then
						Out :+ WrapIdentifier(ParamName) + ":" + WrapDatatype(ParamType)
						
						If Param <> Params.Last() Then Out :+ ", "
					EndIf
				Next
			EndIf
			
			Out :+ ")"
			
			If TryDefineSymbol(WrapIdentifier(Name)) Then Outstream.WriteLine("~t" + Out)
		Case "variable"
			If TryDefineSymbol(WrapIdentifier(Name)) Then Outstream.WriteLine("~tGlobal " + WrapIdentifier(Name) + ":" + WrapDatatype(Decl + DType))
		Case "typedef"
			If Name = "GLUTproc" Then
				DebugStop
				WrapDatatype(Decl + DType)
			EndIf
			
			TypedefMap.Insert(Name, WrapDatatype(Decl + DType))
	End Select
End Function

Function WrapLiteral:String(Value:String)
	If Value.StartsWith("0x") Then
		Return "$" + Value[2 ..]
	ElseIf Value[0] = "~q"[0] Then
		Return Value
	ElseIf Value.Find(".") >= 0 Then
		If Value.EndsWith("f") Then
			Return Value[.. Value.Length - 1]
		Else
			Return Value + "D"
		EndIf
	Else
		Return Value
	EndIf
End Function

Function WrapIdentifier:String(Identifier:String)
	If KeywordMap.Contains(Identifier.ToLower()) Then Return WrapIdentifier("_" + Identifier)
	
	Return Identifier
End Function

Function WrapDatatype:String(Kind:String)
	Local OutType:String
	
	Kind = Kind.Replace("p.f", "f")
	
	Local SubParts:String[]
	Local LastOffset:Int = 0, BracketLevel:Int = 0
	For Local I:Int = 0 Until Kind.Length
		Select Chr(Kind[I])
			Case "."
				If Not BracketLevel Then
					SubParts :+ [Kind[LastOffset .. I]]
					LastOffset = I + 1
				EndIf
			Case "(", "<"
				BracketLevel :+ 1
			Case ")", ">"
				BracketLevel :- 1
		End Select
	Next
	SubParts :+ [Kind[LastOffset ..]]
	
	For Local Index:Int = 0 Until SubParts.Length
		Local I:String = SubParts[Index]
		
		If I And (I.Length = 1 Or I[1] = "("[0]) Then
			Select Chr(I[0])
				Case "p"
					OutType = " Ptr" + OutType
				Case "a"
					OutType = " Ptr" + OutType 'Arrays: Interpret as pointer
				Case "r"
					'Reference: do nothing
				Case "m"
					'Member pointer: Wtf do I do
					Print "Warning: Encountered member pointer."
					DebugStop
				Case "f"
					OutType = ")" + OutType
					
					Local Param:Int = "A"[0]
					Local FunctionType:String
					
					Local IsFirst:Int = True
					For Local T:String = EachIn I[2 .. I.Length - 1].Split(",")
						If T = "void" Or T = "" Then Continue
						
						If Not IsFirst Then FunctionType :+ ", "
						
						FunctionType :+ Chr(Param) + ":" + WrapDatatype(T)
						
						Param :+ 1
						
						IsFirst = False
					Next
					
					If IsFirst Then FunctionType :+ "Void"
					
					OutType = "(" + FunctionType + OutType
				Case "q"
					Rem
					WriteStdout "Warning: Ignoring qualifiers "
					
					For Local T:String = EachIn I[2 .. I.Length - 1].Split(",")
						WriteStdout T + " "
					Next
					
					Print ""
					End Rem
			End Select
		Else
			OutType = ResolveBasicType(I) + OutType
		EndIf
	Next
	
	Return OutType
End Function

Function ResolveBasicType:String(Original:String)
	Select Original
		Case "", "void"
			Return "Void"
		Case "char", "signed char"
			Return "Byte"
		Case "short"
			Return "Short"
		Case "int", "long"
			Return "Int"
		Case "long long"
			Return "Long"
		Case "unsigned char"
			Return "UByte"
		Case "unsigned short"
			Return "UShort"
		Case "unsigned int", "unsigned long"
			Return "UInt"
		Case "unsigned long long"
			Return "ULong"
		Case "float"
			Return "Float"
		Case "double"
			Return "Double"
		Default
			If TypedefMap.Contains(Original) Then
				Return String(TypedefMap.ValueForKey(Original))
			Else
				Return WrapIdentifier(Original) 'TODO: Templates
			EndIf
	End Select
End Function

Function FindAttribute:String(NodeList:TList, Name:String)
	Name = Name.ToLower()
	
	For Local Node:TxmlNode = EachIn NodeList
		If Node.GetAttribute("name").ToLower() = Name Then Return Node.GetAttribute("value")
	Next
End Function

Function FindChild:TxmlNode(Node:TxmlBase, Name:String)
	For Local Child:TxmlNode = EachIn Node.GetChildren()
		If Child.GetName() = Name Then Return Child
	Next
End Function

Function Init()
	Local KeywordList:String[] = ["Abs", "Abstract", "And", "Asc", "Assert", "Byte", "Case", "Catch", "Chr", "Const",  ..
		"Continue", "Default", "DefData", "Delete", "Double", "EachIn", "Else", "ElseIf", "EndExtern", "EndFunction",  ..
		"EndIf", "EndMethod", "EndRem", "EndSelect", "EndTry", "EndClass", "EndWhile", "Exit", "Extends", "Extern","False",  ..
		"Field", "Final", "Float", "For", "Forever", "Framework", "Function", "Global", "Goto", "If", "Import", "Incbin",  ..
		"IncbinLen", "IncbinPtr", "Include", "Int", "Len", "Local", "Long", "Max", "Method", "Min", "Mod", "Module", "ModuleInfo",  ..
		"New", "Next", "Not", "Null", "Or", "Pi", "Private", "Ptr", "Public", "ReadData", "Release", "Rem", "Repeat",  ..
		"RestoreData" , "Return" , "Sar" , "Select" , "Self" , "Sgn" , "Shl" , "Short" , "Shr" , "SizeOf" , "Step" , "Strict" , ..
		"Super", "SuperStrict", "Then", "Throw", "To","True", "Try", "Class", "Until", "Var", "Varptr", "Wend", "While", "Bool", ..
		"Do", "UByte", "UShort", "UInt", "ULong", "Void"]
	
	TypedefMap = New TMap
	KeywordMap = New TMap
	 SymbolMap = New TMap
	
	For Local I:String = EachIn KeywordList
		KeywordMap.Insert(I.ToLower(), I)
	Next
End Function