#ifndef BISMUTH_UTIL_ARRAY
#define BISMUTH_UTIL_ARRAY

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>

#include "Bismuth/Core/Object/.bismuth/ModuleObject.h"

template<typename T, int D> struct Array: public Object {
	T* Data;

	int Length;
	int* Sizes;

	Array(int Count, ...);
	T* At(int Index1, ...);

	void RetrieveAndSet(va_list* Values, int Index);
};

template<typename T> struct Array<T, 1>: public Object {
	T* Data;

	int Length;
	int* Sizes;

	Array(int Count, ...);
	T* At(int Index1, ...);

	void RetrieveAndSet(va_list* Values, int Index);
};

template<typename T, int D> void Array<T, D>::RetrieveAndSet(va_list* Values, int Index) {
	Data[Index] = va_arg(*Values, T);
}

template<typename T> void Array<T, 1>::RetrieveAndSet(va_list* Values, int Index) {
	Data[Index] = va_arg(*Values, T);
}

template<> inline void Array<float, 1>::RetrieveAndSet(va_list* Values, int Index) {
	Data[Index] = (float)va_arg(*Values, double);
}

template<> inline void Array<bool, 1>::RetrieveAndSet(va_list* Values, int Index) {
	Data[Index] = (bool)va_arg(*Values, int);
}

template<> inline void Array<char, 1>::RetrieveAndSet(va_list* Values, int Index) {
	Data[Index] = (char)va_arg(*Values, int);
}

template<> inline void Array<signed char, 1>::RetrieveAndSet(va_list* Values, int Index) {
	Data[Index] = (signed char)va_arg(*Values, int);
}

template<> inline void Array<unsigned char, 1>::RetrieveAndSet(va_list* Values, int Index) {
	Data[Index] = (unsigned char)va_arg(*Values, int);
}

template<> inline void Array<short, 1>::RetrieveAndSet(va_list* Values, int Index) {
	Data[Index] = (short)va_arg(*Values, int);
}

template<> inline void Array<unsigned short, 1>::RetrieveAndSet(va_list* Values, int Index) {
	Data[Index] = (unsigned short)va_arg(*Values, int);
}


template<typename T> Array<T, 1>::Array(int Count, ...) {
	if (Count == -1) {
		Sizes = &Length;

		va_list Values;
		va_start(Values, Count);
        Length = va_arg(Values, int);
		va_end(Values);

		Data = (T*)/*GC_MALLOC*/malloc(Length*sizeof(T));
        memset((void*)Data, 0, Length*sizeof(T));
	} else {
		Length = Count;
		Sizes = &Length;

		if (Count) {
			Data = (T*)/*GC_MALLOC*/malloc(Count*sizeof(T));

			va_list Values;
			va_start(Values, Count);

			for (int I = 0; I < Count; I++)
				RetrieveAndSet(&Values, I);

			va_end(Values);
		} else {
			Data = NULL;
		}
	}
}

template<typename T, int D> Array<T, D>::Array(int Count, ...) {
	if (Count == -1) {
		Sizes = (int*)malloc(D*sizeof(int));

		va_list Values;
		va_start(Values, Count);

		int ElementCount = 1;
		for (int I = 0; I < D; I++) {
			Sizes[I] = va_arg(Values, int);
			ElementCount *= Sizes[I];
		}

		va_end(Values);

		Length = Sizes[0];

		Data = (T*)/*GC_MALLOC*/malloc(ElementCount*sizeof(T));
        memset((void*)Data, 0, ElementCount*sizeof(T));
	} else {
		Length = Count;
		Sizes = &Length;

		Data = NULL;
	}
}

template<typename T, int D> T* Array<T, D>::At(int Index1, ...) {
    assert(Index1 < Length);
    
	va_list Values;
	va_start(Values, Index1);

	int Index = Index1;
	for (int I = 1; I < D; I++) {
        int SubIndex = va_arg(Values, int);
        
        assert(SubIndex < Sizes[I]);
		Index = Index*Sizes[I] + SubIndex;
    }

	va_end(Values);

	return Data + Index;
}

template<typename T> T* Array<T, 1>::At(int Index1, ...) {
    assert(Index1 < Length);
	return Data + Index1;
}

#endif
