#include "String.h"

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

static char ConversionBuffer[256];

//Custom lli-to-string routine since sprintf provides no standardized way of doing that
static int ConvertLongLong(long long int Value) {
	int Sign = Value < 0;
    Value = Sign ? -Value : Value;
	
	int Size = 0;
    do {
        ConversionBuffer[Size++] = (Value % 10) + '0';
    } while(Value /= 10);

	if (Sign)
		ConversionBuffer[Size++] = '-';

	for (int I = 0; I < Size/2; I++) {
		char Temp = ConversionBuffer[I];
		ConversionBuffer[I] = ConversionBuffer[Size - I - 1];
		ConversionBuffer[Size - I - 1] = Temp;
	}
	ConversionBuffer[Size] = '\0';

	return Size;
}

String::String(const char* pData) {
	Length = strlen(pData);
	Data = (char*)malloc(Length + 1);
    memcpy(Data, pData, Length + 1);
}

String* String::Append(char* Buffer, int Size) {
	char* NewBuffer = (char*)malloc(Length + Size + 1);
    memcpy(NewBuffer, Data, Length);
    memcpy(NewBuffer + Length, Buffer, Size);
    NewBuffer[Length + Size] = '\0';

	return new String(NewBuffer, Length + Size);
}

String* String::Prepend(char* Buffer, int Size) {
	char* NewBuffer = (char*)malloc(Length + Size + 1);
    memcpy(NewBuffer, Buffer, Size);
	memcpy(NewBuffer + Length, Data, Length);
    NewBuffer[Length + Size] = '\0';

	return new String(NewBuffer, Length + Size);
}

String* String::Append(char* Buffer) {
    return Append(Buffer, strlen(Buffer));
}

String* String::Prepend(char* Buffer) {
    return Prepend(Buffer, strlen(Buffer));
}

String* String::Append(String* Other) {
	return Append(Other->Data, Other->Length);
}

String* String::Append(bool Value) {
	return Append(ConversionBuffer, sprintf(ConversionBuffer, "%d", Value));
}

String* String::Append(int Value) {
	return Append(ConversionBuffer, sprintf(ConversionBuffer, "%d", Value));
}

String* String::Append(short Value) {
	return Append(ConversionBuffer, sprintf(ConversionBuffer, "%hd", Value));
}

String* String::Append(long long int Value) {
	return Append(ConversionBuffer, ConvertLongLong(Value));
}

String* String::Append(size_t Value) {
	return Append((long long int)Value);
}

String* String::Append(float Value) {
	return Append(ConversionBuffer, sprintf(ConversionBuffer, "%f", Value));
}

String* String::Append(double Value) {
	return Append(ConversionBuffer, sprintf(ConversionBuffer, "%f", Value));
}


String* String::Prepend(String* Other) {
	return Prepend(Other->Data, Other->Length);
}

String* String::Prepend(bool Value) {
	return Prepend(ConversionBuffer, sprintf(ConversionBuffer, "%d", Value));
}

String* String::Prepend(short Value) {
	return Prepend(ConversionBuffer, sprintf(ConversionBuffer, "%hd", Value));
}

String* String::Prepend(int Value) {
	return Prepend(ConversionBuffer, sprintf(ConversionBuffer, "%d", Value));
}

String* String::Prepend(long long int Value) {
	return Prepend(ConversionBuffer, ConvertLongLong(Value));
}

String* String::Prepend(size_t Value) {
	return Prepend((long long int)Value);
}

String* String::Prepend(float Value) {
	return Prepend(ConversionBuffer, sprintf(ConversionBuffer, "%f", Value));
}

String* String::Prepend(double Value) {
	return Prepend(ConversionBuffer, sprintf(ConversionBuffer, "%f", Value));
}

String* String::Part(int From, int To) {
    From = From < 0 ? 0 : From;
    To = To > Length ? Length : To;
    
    if (From >= To)
        return new String("");
    
    char *Buffer = (char*)malloc((To - From + 1)*sizeof(char));
    memcpy(Buffer, Data + From, To - From);
	Buffer[To - From] = '\0';
	
	return new String(Buffer, To - From);
}

int String::Find(String* Substring, int From) {
    if (!Substring->Length)
        return -1;
    
    for (int I = From, T = I; I <= Length - Substring->Length; I++, T = I) {
        while (Data[T] == Substring->Data[T - I]) {
            T++;
            if (T - I == Substring->Length)
                return I;
        }
    }
    return -1;
}

int String::FindLast(String* Substring, int From){
    if (!Substring->Length)
        return -1;
    
	for (int I = Length - Substring->Length, T = I; I >= From; I--, T = I) {
        while (Data[T] == Substring->Data[T - I]) {
            T++;
            if (T - I == Substring->Length)
                return I;
        }
    }
    return -1;
}


String* String::Replace(String* Search, String* Replacement) {
    int NewLength = Length + CountOccurrences(Search)*(Replacement->Length - Search->Length);
    char *Buffer = (char*)malloc((NewLength + 1)*sizeof(char));
    
    int LastI = 0, Index = 0, I = 0;
    do {
        I          = Find(Search, LastI);
        int Border = I >= 0 ? I : Length;
        int Span   = Border - LastI;
        memcpy(Buffer + Index, Data + LastI, Span);
        memcpy(Buffer + Index + Span, Replacement->Data, Replacement->Length);
        
        LastI  = I + Search->Length;
        Index += Span + Replacement->Length;
    } while (I >= 0);
    Buffer[NewLength] = '\0';
    
	return new String(Buffer, NewLength);
}

int String::ToInt() {
	return atoi(Data);
}

float String::ToFloat() {
	return atof(Data);
}

String* String::TrimLead() {
	int Pos;
	for (Pos = 0; Pos < Length && Data[Pos] <= 32; Pos++);
	return Part(Pos, Length);
}

String* String::TrimTrail() {
	int Pos;
	for (Pos = Length; Pos > 0 && Data[Pos - 1] <= 32; Pos--);
	return Part(0, Pos);
}

String* String::Trim() {
	int Pos1, Pos2;
    for (Pos1 = 0; Pos1 < Length && Data[Pos1] <= 32; Pos1++);
    for (Pos2 = Length; Pos2 > 0 && Data[Pos2 - 1] <= 32; Pos2--);
    return Part(Pos1, Pos2);
}

String* String::FromChar(int Code) {
	char* Buffer = (char*)malloc(2*sizeof(char));
	Buffer[0] = Code;
	Buffer[1] = '\0';
	return new String(Buffer, 1);
}

String* String::ToLower() {
	char* Buffer = (char*)malloc(Length*sizeof(char));
    
	for (int I = 0; I <= Length; I++)
		Buffer[I] = tolower(Data[I]);
    
	return new String(Buffer, Length);
}

String* String::ToUpper() {
	char* Buffer = (char*)malloc(Length*sizeof(char));
    
	for (int I = 0; I <= Length; I++)
		Buffer[I] = toupper(Data[I]);
    
	return new String(Buffer, Length);
}

int String::Contains(String *Substring) {
	return Find(Substring, 0) >= 0;
}

int String::StartsWith(String *Substring) {
	return !(Length < Substring->Length || strcmp(Data, Substring->Data));
}

int String::EndsWith(String* Substring) {
	return !(Length < Substring->Length || strcmp(Data - Substring->Length, Substring->Data));
}

String* String::FromCString(char *CString) {
	return new String(CString);
}

String* String::FromCString(const char *CString) {
	return new String((char*)CString);
}

bool String::Equals(String* Other) {
    return !(Length != Other->Length || strcmp(Data, Other->Data));
}

String* String::Reverse() {
	char *Buffer = (char*)malloc((Length + 1)*sizeof(char));
    
	for (char *P = Buffer, *Q = Data + Length - 1; Q >= Data; P++, Q--)
        *P = *Q;
    Buffer[Length] = '\0';
    
	return new String(Buffer, Length);
}

String* String::Hex(int Number) {
	char* Buffer = (char*)malloc(9*sizeof(char));
    
	for (int I = 7; I >= 0; I--, Number >>= 4) {
        int Digit = Number & 0xF;
		Buffer[I] = Digit > 9 ? Digit - 10 + 'A' : Digit + '0';
	}
	Buffer[8] = '\0';
    
	return new String(Buffer, 8);
}

String* String::Bin(int Number) {
	char* Buffer = (char*)malloc(33*sizeof(char));
    
	for (int I = 31; I >= 0; I--, Number >>= 1)
        Buffer[I] = '0' + (Number & 1);
	Buffer[32] = '\0';
    
	return new String(Buffer, 32);
}

String* String::Oct(int Number) {
	char* Buffer = (char*)malloc(12*sizeof(char));
    
	for (int I = 10; I >= 0; I--, Number >>= 3)
        Buffer[I] = '0' + (Number & 7);
	Buffer[11] = '\0';
    
	return new String(Buffer, 11);
}

String* String::FromBytes(Array<char, 1> *Bytes, int Count) {
	char *Buffer = (char*)malloc((Count + 1)*sizeof(char));
    
	memcpy(Buffer, Bytes->At(0), Count);
    Buffer[Count] = '\0';
    
	return new String(Buffer, Count);
}

char *String::ToCString() {
	return Data;
}

Array<String*,1>* String::Split(String* Separator) {
    if (!Length)
        return new Array<String*, 1>(-1, 0);
    
    int Count = CountOccurrences(Separator);
	Array<String*,1>* Result = new Array<String*, 1>(-1, Count + 1);
    
    int LastI = 0, I = Find(Separator, 0), Index = 0;
	for(; Index < Count; LastI = I + Separator->Length, I = Find(Separator, LastI), Index++)
		*Result->At(Index) = Part(LastI, I);
    
	*Result->At(Count) = Part(LastI, Length);
    
	return Result;
}

int String::CountOccurrences(String* SubString) {
	int Count = 0;
    for (int I = Find(SubString, 0); I >= 0; I = Find(SubString, I + SubString->Length))
        Count++;
    
	return Count;
}
