Import "FileStream.cpp"
Import "FileStream.h"

Extern
	Function _CFileReadBytes:Int(CFile:Void Ptr, Dest:Void Ptr, Count:Int)
	Function _CFileWriteBytes:Int(CFile:Void Ptr, Src:Void Ptr, Count:Int)
	Function _CFileEOF:Bool(CFile:Void Ptr)
	Function _CFileFlush(CFile:Void Ptr)
	Function _CFilePos:Int(CFile:Void Ptr)
	Function _CFileSeek(CFile:Void Ptr, Pos:Int)
	Function _CFileOpen:Void Ptr(Path:String, Mode:String)
	Function _CFileClose(CFile:Void Ptr)
	Function _CFileSize:Int(CFile:Void Ptr)
End Extern

Function LoadText:String(Path:String)
    Local Instream:TStream = ReadFile(Path)
    
    If Instream Then
        Local Size:Int = Instream.Size()
        Local Dest:Byte Ptr = Byte Ptr(MemAlloc(Size + 1))
        
        Instream.ReadBytes(Dest, Size)
        Dest[Size] = 0
        
        Local Result:String = String.FromCString(Dest)
        MemFree(Dest)
        Instream.Close()
        
        Return Result
    Endif
End Function

Function SaveText(Path:String, Text:String)
    Local Outstream:TStream = WriteFile(Path)
    
    If Outstream Then
        Outstream.WriteBytes(Text.ToCString(), Text.Length)
        Outstream.Close()
    Endif
End Function

Function ReadFile:TFileStream(Path:String)
    Return New TFileStream.Create(_CFileOpen(Path, "r+b"))
End Function

Function WriteFile:TFileStream(Path:String)
    Return New TFileStream.Create(_CFileOpen(Path, "w+b"))
End Function

Class TFileStream Extends TStream
    Field CFile:Void Ptr
    
    Method ReadBytes:Int(Dest:Void Ptr, Count:Int)
        Return _CFileReadBytes(CFile, Dest, Count)
    End Method
    
    Method WriteBytes:Int(Src:Void Ptr, Count:Int)
        Return _CFileWriteBytes(CFile, Src, Count)
    End Method
    
    Method Eof:Bool()
        Return _CFileEOF(CFile)
    End Method
    
    Method Flush()
        _CFileFlush(CFile)
    End Method
    
    Method Pos:Int()
        Return _CFilePos(CFile)
    End Method
    
    Method Seek(Position:Int)
        _CFileSeek(CFile, Position)
    End Method
    
    Method Size:Int()
        Return _CFileSize(CFile)
    End Method
    
    Method Close()
        If CFile Then
            _CFileClose(CFile)
            CFile = Null
        Endif
    End Method
    
    Method Create:TFileStream(CFile:Void Ptr)
        If Not CFile Then Return Null
        
        Self.CFile = CFile
        
        Return Self
    End Method
End Class