/*
	Module: 
		Bismuth.Random
	
	Author: 
		Moritz "Xeres" Karbaum
	
	Version:
		1.0 Initial Release
	
	About:
		Wrapper for Mersenne twister pseudorandom number generator
		by Makoto Matsumoto and Takuji Nishimura

		Using SFMT-src-1.3.3
		http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/SFMT/#SFMT
	
	To-Do:
		Implement optional parameters
	
	Overview:
	
		SeedRand:Int(seed:Int)
		RandInt:Int(min_value:Int, max_value:Int)
		RandDouble:Double(min_value:Double, max_value:Double)
		RandLong:Long(min_value:Long, max_value:Long)
*/

Import "SFMT/SFMT.c"
Import "SFMT/SFMT.h"
Import "SFMT/SFMT-params.h"
Import "SFMT/SFMT-params19937.h"

Extern
	Function genrand_res53:Double()
	Function init_gen_rand:Int(seed:Int)
EndExtern

Global SFMT_Initialized:Bool = False

Function SeedRand:Int(seed:Int)
	SFMT_Initialized = True
	init_gen_rand(seed)
EndFunction

/*

	Generates random integer.
	
*/
Function RandInt:Int(min_value:Int, max_value:Int) // max_value:Int = 1
	If SFMT_Initialized = False Then
		SeedRand(0)
	EndIf
	
	Local range:Int = max_value - min_value
	If range > 0 Then
		Return Int(genrand_res53()*(1+range))+min_value
	Else
		Return Int(genrand_res53()*(1-range))+max_value
	Endif	
EndFunction

/*

	Generates random double.

*/
Function RandDouble:Double(min_value:Double, max_value:Double) // min_value:Double=1, max_value:Double=0
	If SFMT_Initialized = False Then
		SeedRand(0)
	EndIf
	
	If max_value > min_value Then
		Return genrand_res53()*(max_value - min_value) + min_value
	Else
		Return genrand_res53()*(min_value - max_value) + max_value
	Endif
EndFunction

/*

	Generates random long.

*/
Function RandLong:Long(min_value:Long, max_value:Long) // max_value:Long = 1
	If SFMT_Initialized = False Then
		SeedRand(0)
	EndIf
	
	Local range:Long = max_value - min_value
	If range > 0 Then
		Return Long(genrand_res53()*(1L + range)) + min_value
	Else
		Return Long(genrand_res53()*(1L - range)) + max_value
	Endif
EndFunction

