// Module: Bismuth.Time
//
// Author: Andreas 'BladeRunner' Brennecke
// V1.0 Initial Release
//
//about:
//Bismuth.Time offers different functions for time management. 
//
Import "Time.cpp"
Import "Time.h"

Extern
	Function MilliSecs:Int()
	// returns the amount of millisecs passed since boot up.
	
	Function MicroSecs:Long()
	// returns the amount of microsecs passed since boot up.
	
	Class tm
		Field tm_sec:Int
		Field tm_min:Int
		Field tm_hour:Int
		Field tm_mday:Int
		Field tm_mon:Int
		Field tm_year:Int
		Field tm_wday:Int
		Field tm_yday:Int
		Field tm_isdst:Int
	EndClass
	
	Function time:Int(Timer:Int)
	Function WrappedLocaltime:tm(Timer:Int)
EndExtern

Function UnixTimeStamp:Int()
	return time(0)
EndFunction

Class BiTime
	Field _tm:tm
	Field _time:Int
	//Field _daynames:String[]=["Montag","Dienstag","Mittwoch","Donnerstag","Freitag","Samstag","Sonntag"]
	//Field _monnames:String[]=["Januar","Februar","Mrz","April","Mai","Juni","Juli","August","September","Oktober","November","Dezember"]
	
	Method SetTime:Int(timestamp:Int)
		If timestamp = -1 Then
			_time = time(0)
		Else
			_time = timestamp
		Endif
		_tm = WrappedLocaltime(_time)
	EndMethod
	
	Method SetNames:Int(names:String)
		//Names sollte folgendes Format haben:
		//"Montag,Dienstag,Mittwoch,Donnerstag,Freitag,Samstag,Sonntag,Januar,Februar,Mrz,
		//April,Mai,Juni,Juli,August,September,Oktober,November,Dezember"
		//in der entsprechenden Sprache.
	/*	
		For local i:Int = 0 to 6
			_daynames[i] = Part(names,0,Find(names,",",0)-1)
			names = Part(names,Find(names,",",0)+1,Length(names))
		Next
		For local i:Int = 0 to 11
			_monnames[i] = Part(names,0,Find(names,",",0)-1)
			names = Part(names,Find(names,",",0)+1,Length(names))
		Next
	*/	
	EndMethod
	
	Method Seconds:Int()
		Return _tm.tm_sec
	EndMethod

	Method Minutes:Int()
		Return _tm.tm_min
	EndMethod

	Method Hour:Int()
		Return _tm.tm_hour
	EndMethod
	
	Method Year:Int()
		Return _tm.tm_year+1900
	EndMethod
	
	Method MonthDay:Int()
		Return _tm.tm_mday
	EndMethod
	
	Method WeekDay:Int()
		Return _tm.tm_wday
	EndMethod
	
	Method YearDay:Int()
		Return _tm.tm_yday
	EndMethod
	
	Method Month:Int()
		Return _tm.tm_mon+1
	EndMethod
		
	Method isDaylightSavingTime:Int()
		Return _tm.tm_isdst
	EndMethod
	/*
	Method MonthName:String(mon:Int)
		Return _monnames[mon-1]
	EndMethod
	
	Method DayName:String(day:Int)
		Return _daynames[day]
	EndMethod
	*/
	Method TimeString:String(format:String)
		
		/*Format:
		% leitet einen Bezeichner ein. Groschreibung bedeutet Wiedergabe als String, Kleinschreibung als Zahl.
		Bei Groschreibung erfolgt auerdem optional die Angabe der auszuschreibenden Buchstaben als Ziffer(1-9). Folgt keine Ziffer werden die Namen ausgeschrieben.
		Beispiele:
		"Es ist %W, der %md. %MON %y." -> "Es ist Montag, der 16. August 1982"
		"%md.%MON3.%y" -> "16.Aug.1982"
		
		Bezeichner:
		h	- Stunde. Keine Groschreibung.
		m	- Minute. Keine Groschreibung.
		s	- Sekunde. Keine Groschreibung.
		md	- Tag des Monats. Keine Groschreibung.
		mon	- Monatsname.
		y	- Jahreszahl. Keine Groschreibung.
		w	- Wochentag.
		yd	- Tag des Jahres. Keine Groschreibung.
		dst	- Sommer-Winterzeit. (noch nicht implementiert).
		
		*/
		
		/*
			Vereinheitlichen?
			
			Seconds & Minutes <-> Hour & Year
			=>
			Seconds, Minutes, Hours, Years
			
			---
			
			Tag des Monats, Monatsname...
			mn - Monatszahl
		*/
		
		Local tmp:String
		
		// Mehrere Buchstaben zuerst, da "%m" in "%md" enthalten ist.
		
		If format.Contains("%md") Then 
			format = format.replace("%md", Self.MonthDay())
		EndIf
		
		If format.Contains("%mn") Then 
			format = format.replace("%mn", Self.Month())
		EndIf
		
		If format.Contains("%h") Then 
			tmp = Self.Hour()
			If tmp.length = 1 Then
				tmp = "0"+tmp
			Endif
			format = format.replace("%h", tmp)
		EndIf
		
		If format.Contains("%m") Then 
			tmp = Self.Minutes()
			If tmp.length = 1 Then
				tmp = "0"+tmp
			Endif
			format = format.replace("%m", tmp)
		EndIf
		
		If format.Contains("%s") Then 
			tmp = Self.Seconds()
			If tmp.length = 1 Then
				tmp = "0"+tmp
			Endif
			format = format.replace("%s", tmp)
		EndIf
		

		
		If format.Contains("%y") Then 
			format = format.replace("%y", Self.year())
		EndIf

		
		return format
		
		
	/*	Local h:String = Hour()
		If Hour() <10 Then
			h = "0"+h
		EndIf
		Local m:String = Minutes()
		If Minutes() <10 Then
			m = "0"+m
		EndIf
		Local s:String = Seconds()
		If seconds() <10 Then
			s = "0"+s
		EndIf
		Local md:String = MonthDay()
		If MonthDay() <10 Then
			md = "0"+md
		EndIf		
		Local mon:String = Month()
		If MOnth() <10 Then
			mon = "0"+mon
		EndIf
		Local y:String = Year()
		
		Local output:String = ""
	
		For Local i:Int = 0 to Length(format)-1
			If Part(format,i,i) = "%" Then
				Select Part(format,i+1,i+1)
					Case "%"
						output = output+"%"
						i = i + 1
					Case "s"
						output = output + s
					Case "m"
						If Part(format,i+1,i+2) = "md" Then
							output = output + md
							i = i + 2
						ElseIf Part(format,i+1,i+3) = "mon" Then
							output = output + mon
							i = i + 3
						Else						
							output = output + m
							i = i +1
						EndIf
					Case "h"
						output = output + h
					Case "w"
						output = output + WeekDay()
					Case "y"
						If Part(format,i+1,i+2) = "yd" Then
							output = output + yd
							i = i + 2
						Else
							output = output + y
							i = i + 1
						Endif
					Case "M"
						If To_Int(Part(format,i+4,i+4)) >0 Then
							output = output + Part(MonthName(Month()),0,To_Int(Part(format,i+4,i+4))-1)
							i = i + 4
						Else
							output = output + MonthName(Month())
							i = i + 3
						Endif
					Case "W"
						If To_Int(Part(format,i+2,i+2)) >0 Then
							output = output + Part(DayName(WeekDay()),0,To_Int(Part(format,i+2,i+2))-1)
							i = i +2
						Else
							output = output + DayName(WeekDay())
							i = i + 1
						Endif					
					Default
						//Parsing Error!
				EndSelect
			Else
				output = output + Part(format,i,i)
			EndIf
		Next
		

		
		Return output
	*/
	EndMethod
EndClass


Class BiMicroCounter
// Little Counter which counts microsecs. Perfect for profiling runtimes.	
	Field starttime:Long
	Field endtime:Long
	Field running:Byte
	
	Method Start:Int()
		If running = True Then
			starttime = MicroSecs()
			running = True
		EndIf
	EndMethod
	
	Method Stop:Long()
		endtime = MicroSecs()
		running = False
		Return endtime-starttime
	EndMethod
	
	Method Query:Long()
		If running = True Then
			Return MicroSecs()- starttime
		Else
			Return endtime - starttime
		EndIf
	EndMethod
	
	Method Reset:int()
		running = False
		starttime = False
		endtime = False
	EndMethod
	
EndClass

Class BiMicroTimer

	Field interval:Long
	Field starttime:Long
	Field lastquery:Long
	Field lasttick:Long
	Field running:Byte
	
	Function Create:BiMicroTimer(_interval:Long)
		local mt:BiMicroTimer = New BiMicroTimer
		mt.interval = _interval
		mt.running = True
		mt.starttime = MicroSecs()
		mt.lastquery = mt.starttime
		mt.lasttick = mt.starttime		
		return mt
	EndFunction
	
	
	Method Query:Int()
		Local time:Long = MicroSecs()
		If running = True Then
			If (lasttick + interval) < time Then //Firin' min. one time.
				Local times:Int = (time - lasttick) / interval
				
				lasttick = lasttick + (times*interval)
				lastquery = time
				return times
			EndIf
		EndIf
		lastquery = time
		return False
	EndMethod
	
	Method Wait:Int()
		Repeat
		Until Query()
		return True
	EndMethod
	
	Method Pause:Int()
		running = False
	EndMethod
	
	Method Resume:Int()
		if running = False Then
			running = True
			lasttick = MicroSecs()
			lastquery = lasttick
		EndIf
	EndMethod
	
	Method lastqueried:Long()
		Return lastquery
	EndMethod
EndClass
